(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** CPS-based intermediate language. *)

Require Import Arith Bool Image List NPeano Program Omega ClassicalChoice.
Require Export Sorting.Permutation.
Require Import vgtac basis countable lang_common lang_spec.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".



(** * Syntax. *)

Definition location := nat.

(** External references (labels) are treated as a special kind of
    variable. *)
Definition contvar := nat.
Definition itermvar := (label + termvar)%type.
Definition var := ((label + termvar) + contvar)%type.

Inductive exp : Set :=
| IEunit
| IEnat  (n:nat)
| IEpair (x:itermvar) (y:itermvar)
| IEinl  (x:itermvar)
| IEinr  (x:itermvar)
| IEbop  (op: binop) (x:itermvar + nat) (y:itermvar + nat)
| IEfst  (x:itermvar)
| IEsnd  (x:itermvar)
| IEfix  (bf:termvar) (bx:termvar) (bk:contvar) (e:exp)
| IEgen  (bk:contvar) (e:exp)
| IEcont (bx:termvar) (e:exp)
| IEnew    (x:itermvar) (bx:termvar) (e:exp)
| IEread   (x:itermvar) (bx:termvar) (e:exp)
| IEwrite  (x:itermvar) (y:itermvar) (e:exp)
| IEinput  (bx:termvar) (e:exp)
| IEoutput (x:itermvar) (e:exp)
| IEifnz   (x:itermvar) (ea:exp) (eb:exp)
| IEcase   (x:itermvar) (bx:termvar) (ea:exp) (eb:exp)
| IEret    (xk:contvar) (x:itermvar)
| IEapp    (xf:itermvar) (x:itermvar) (xk:contvar)
| IEinst   (xf:itermvar) (xk:contvar)
| IElet (ae:exp) (bx:termvar + contvar) (e:exp)
| IErefeq (x:itermvar) (y:itermvar) 
.

Fixpoint eqdec_exp (e1: exp) (e2: exp) : bool :=
  match e1, e2 with
  | IEunit, IEunit =>
    true
  | IEnat n1, IEnat n2 =>
    eqdec n1 n2
  | IEpair x1 y1, IEpair x2 y2 =>
    eqdec x1 x2 && eqdec y1 y2
  | IEinl x1, IEinl x2 => eqdec x1 x2
  | IEinr x1, IEinr x2 => eqdec x1 x2
  | IEbop op1 a1 b1, IEbop op2 a2 b2 =>
    eqdec op1 op2 && eqdec a1 a2 && eqdec b1 b2
  | IEfst x1, IEfst x2 => eqdec x1 x2
  | IEsnd x1, IEsnd x2 => eqdec x1 x2
  | IEfix tx1 ty1 cx1 e1', IEfix tx2 ty2 cx2 e2' =>
    eqdec tx1 tx2 && eqdec ty1 ty2 &&
    eqdec cx1 cx2 && eqdec_exp e1' e2'
  | IEgen cx1 e1', IEgen cx2 e2' =>
    eqdec cx1 cx2 && eqdec_exp e1' e2'
  | IEcont tx1 e1', IEcont tx2 e2' =>
    eqdec tx1 tx2 && eqdec_exp e1' e2'
  | IEnew x1 tx1 e1', IEnew x2 tx2 e2' =>
    eqdec x1 x2 && eqdec tx1 tx2 && eqdec_exp e1' e2'
  | IEread x1 tx1 e1', IEread x2 tx2 e2' =>
    eqdec x1 x2 && eqdec tx1 tx2 && eqdec_exp e1' e2'
  | IEwrite x1 y1 e1', IEwrite x2 y2 e2' =>
    eqdec x1 x2 && eqdec y1 y2 && eqdec_exp e1' e2'
  | IEinput tx1 e1', IEinput tx2 e2' =>
    eqdec tx1 tx2 && eqdec_exp e1' e2'
  | IEoutput x1 e1', IEoutput x2 e2' =>
    eqdec x1 x2 && eqdec_exp e1' e2'
  | IEifnz x1 e1' e1'', IEifnz x2 e2' e2'' =>
    eqdec x1 x2 && eqdec_exp e1' e2' && eqdec_exp e1'' e2''
  | IEcase x1 tx1 e1' e1'', IEcase x2 tx2 e2' e2'' =>
    eqdec x1 x2 && eqdec tx1 tx2 &&
    eqdec_exp e1' e2' && eqdec_exp e1'' e2''
  | IEret cx1 x1, IEret cx2 x2 =>
    eqdec cx1 cx2 && eqdec x1 x2
  | IEapp x1 y1 cx1, IEapp x2 y2 cx2 =>
    eqdec x1 x2 && eqdec y1 y2 && eqdec cx1 cx2
  | IEinst x1 cx1, IEinst x2 cx2 =>
    eqdec x1 x2 && eqdec cx1 cx2
  | IElet e1' x1 e1'', IElet e2' x2 e2'' =>
    eqdec_exp e1' e2' && eqdec x1 x2 && eqdec_exp e1'' e2''
  | IErefeq x1 y1, IErefeq x2 y2 =>
    eqdec x1 x2 && eqdec y1 y2
  | _, _ =>
    false
  end.

Lemma eqdec_exp_sound: forall e1 e2 (EQ: eqdec_exp e1 e2),
  e1 = e2.
Proof.
  induction e1; unfold eqdec_exp; i; destruct e2; try (by inversion EQ);
  fold eqdec_exp in *;
  repeat match goal with
    | H: sumbool_to_bool (eqdec ?a ?b) = _ |- _ =>
      destruct (eqdec a b)
    | H: is_true (sumbool_to_bool (eqdec ?a ?b)) |- _ =>
      destruct (eqdec a b)
    | H: andb ?x ?y = _ |- _ =>
      apply andb_true_iff in H; des
    | H: is_true (andb ?x ?y) |- _ =>
      apply andb_true_iff in H; des
    | H: eqdec_exp ?a ?b = _ |- _ =>
      first [apply IHe1 in H | apply IHe2 in H |
             apply IHe1_1 in H | apply IHe1_2 in H]; subst
  end; vauto.
Qed.

Lemma eqdec_exp_complete: forall e1 e2 (EQ: e1 = e2),
  eqdec_exp e1 e2.
Proof.
  i; subst; induction e2; unfold eqdec_exp; fold eqdec_exp
  ; try repeat (apply andb_true_iff; split)
  ; try destruct (eqdec _ _); vauto.
Qed.  

Instance Eqdec_exp: EqDec exp.
  refine {| eqdec := _ |}; i.
  assert (X := eqdec_exp_sound a b).
  assert (Y := @eqdec_exp_complete a b).
  destruct eqdec_exp; eauto.
Defined.



(** * Operational semantics.

    It uses an evironment to look up variables and is thus
    substitution-free.
*)

(** Runtime values.

    Functions are closures.
*)
Inductive val := 
| IVunit
| IVnat  (n:nat)
| IVpair (vx: val) (vy: val)
| IVinl  (v: val)
| IVinr  (v: val)
| IVclo  (xvs: list (var * val)) (e: exp)
| IVloc  (loc: location)
.

(** Environments. *)
Definition env :=
  list (var * val).

Definition ilookup (x: itermvar + nat) (xvs: env) : option val :=
  match x with
  | inl y => lookup (inl y) xvs
  | inr i => Some (IVnat i)
  end.

(** Evaluating an expression in an environment. *)
Definition eval (xvs: env) (ae: exp) : option val :=
  match ae with
  | IEunit => Some IVunit
  | IEnat n => Some (IVnat n)
  | IEpair x y => option_map2 IVpair (lookup (inl x) xvs) (lookup (inl y) xvs)
  | IEinl x => option_map IVinl (lookup (inl x) xvs)
  | IEinr x => option_map IVinr (lookup (inl x) xvs)
  | IEfix _ _ _ _
  | IEgen _ _ => Some (IVclo xvs ae)
  | IEbop op x y => 
    match ilookup x xvs, ilookup y xvs with
    | Some (IVnat n), Some (IVnat m) => Some (IVnat (bosem op n m))
    | _,_ => None
    end
  | IEfst x =>
    match lookup (inl x) xvs with
    | Some (IVpair vx vy) => Some vx
    | _ => None
    end
  | IEsnd x =>
    match lookup (inl x) xvs with
    | Some (IVpair vx vy) => Some vy
    | _ => None
    end
  | IErefeq x y =>
    match lookup (inl x) xvs, lookup (inl y) xvs with
    | Some (IVloc lx), Some (IVloc ly) => 
        Some (if eqdec lx ly then IVnat 1 else IVnat 0)
    | _, _ => None
    end
  | _ => None
  end.

(** Heaps are partial maps (which may be infinite). *)
Definition heap := @pmap location _ val.

Inductive reduce_ok : event -> heap -> env -> exp -> heap -> env -> exp -> Prop :=
| IRlet: forall h xvs ae x e v
    (EVAL: eval xvs ae = Some v),
  reduce_ok etau h xvs (IElet ae (inl x) e) h ((inl (inr x),v) :: xvs) e
| IRletcont: forall h xvs bx e' x e,
  reduce_ok etau h xvs (IElet (IEcont bx e') (inr x) e)
                 h ((inr x, IVclo xvs (IEcont bx e')) :: xvs) e
| IRinput: forall h xvs bx e n,
  reduce_ok (ein n) h xvs (IEinput bx e) h ((inl (inr bx),IVnat n) :: xvs) e
| IRoutput: forall h xvs x e n
    (GET: lookup (inl x) xvs = Some (IVnat n)),
  reduce_ok (eout n) h xvs (IEoutput x e) h xvs e
| IRifz: forall h xvs x ea eb
    (GET: lookup (inl x) xvs = Some (IVnat 0)),
  reduce_ok etau h xvs (IEifnz x ea eb) h xvs eb
| IRifnz: forall h xvs x ea eb n
    (GET: lookup (inl x) xvs = Some (IVnat (S n))),
  reduce_ok etau h xvs (IEifnz x ea eb) h xvs ea
| IRcasel: forall h xvs x bx ea eb v
    (GET: lookup (inl x) xvs = Some (IVinl v)),
  reduce_ok etau h xvs (IEcase x bx ea eb) h ((inl (inr bx),v) :: xvs) ea
| IRcaser: forall h xvs x bx ea eb v
    (GET: lookup (inl x) xvs = Some (IVinr v)),
  reduce_ok etau h xvs (IEcase x bx ea eb) h ((inl (inr bx),v) :: xvs) eb
| IRret: forall h xvs xk x xvs' bx e v
    (GET: lookup (inr xk) xvs = Some (IVclo xvs' (IEcont bx e)))
    (GETV: lookup (inl x) xvs = Some v),
  reduce_ok etau h xvs (IEret xk x) h ((inl (inr bx),v) :: xvs') e
| IRapp: forall h xvs xf x xk xvs' bf bx bk e v vk
    (GET: lookup (inl xf) xvs = Some (IVclo xvs' (IEfix bf bx bk e)))
    (GETV: lookup (inl x) xvs = Some v)
    (GETK: lookup (inr xk) xvs = Some vk),
  reduce_ok etau h xvs (IEapp xf x xk) 
    h ((inr bk,vk) :: (inl (inr bx),v) :: (inl (inr bf), IVclo xvs' (IEfix bf bx bk e)) :: xvs') e
| IRinst: forall h xvs xf xk xvs' bk e vk
    (GET: lookup (inl xf) xvs = Some (IVclo xvs' (IEgen bk e)))
    (GETK: lookup (inr xk) xvs = Some vk),
  reduce_ok etau h xvs (IEinst xf xk) h ((inr bk,vk) :: xvs') e
| IRnew: forall h xvs x bx e loc v
    (PV: pmvalid (pmplus (pmcell loc v) h))
    (GET: lookup (inl x) xvs = Some v),
  reduce_ok etau h xvs (IEnew x bx e) 
            (pmplus (pmcell loc v) h) ((inl (inr bx), IVloc loc) :: xvs) e
| IRread: forall h xvs x bx e loc v
    (GET: lookup (inl x) xvs = Some (IVloc loc))
    (LOAD: pmread h loc v),
  reduce_ok etau h xvs (IEread x bx e) h ((inl (inr bx),v) :: xvs) e
| IRwrite: forall h xvs x y e loc u v
    (GET: lookup (inl x) xvs = Some (IVloc loc))
    (GET: lookup (inl y) xvs = Some v),
  reduce_ok etau (pmplus (pmcell loc u) h) xvs (IEwrite x y e) 
            (pmplus (pmcell loc v) h) xvs e
.
Hint Constructors reduce_ok.

Definition exp_halted (e: exp) :=
  match e with
  | IEnat _ => True
  | _ => False
  end.

(** Small-step reduction. *)
Inductive mid_step: event -> heap * (env * exp) -> heap * (env * exp) -> Prop :=
| IRok: forall evt h xvs e h' xvs' e'
    (STEP: reduce_ok evt h xvs e h' xvs' e'),
  mid_step evt (h,(xvs,e)) (h',(xvs',e'))
| IRerr: forall h xvs e
    (NT: ~ exp_halted e)
    (STK: forall evt h' xvs' e', ~ reduce_ok evt h xvs e h' xvs' e'),
  mid_step etau (h,(xvs,e)) (None,(xvs,e))
.
Hint Constructors mid_step.



(** * Well-formedness. *)

Definition Ini (x: itermvar + nat) (vs: list var) :=
  match x with
  | inl x => In (inl x) vs
  | _ => True
  end.

Fixpoint purge (xs: list var) : list var :=
  match xs with
    | nil => nil
    | inr _ :: xs' => purge xs'
    | inl x :: xs' => inl x :: purge xs'
  end.

Definition wfatom_ wfexp (xs: list var) ae : Prop :=
  match ae with
  | IEbop op x y => Ini x xs /\ Ini y xs
  | IEfst x => In (inl x) xs
  | IEsnd x => In (inl x) xs
  | IEunit
  | IEnat _ => True
  | IEpair x y => In (inl x) xs /\ In (inl y) xs
  | IEinl x
  | IEinr x => In (inl x) xs
  | IEfix bf bx bk e => wfexp (inr bk :: inl (inr bx) :: inl (inr bf) :: purge xs) e
  | IEgen bk e => wfexp (inr bk :: purge xs) e
  | IEcont bx e => wfexp (inl (inr bx) :: xs) e
  | IErefeq x y => In (inl x) xs /\ In (inl y) xs
  | _ => False
  end.

Definition wfexp_ wfexp (xs: list var) e : Prop :=
  match e with
  | IElet ae (inl bx) e =>
      wfatom_ wfexp xs ae /\ wfexp (inl (inr bx) :: xs) e
  | IElet ae (inr bk) e =>
      wfatom_ wfexp xs ae /\ wfexp (inr bk :: xs) e
  | IEinput bx e =>
      wfexp (inl (inr bx) :: xs) e
  | IEoutput x e =>
      In (inl x) xs /\ wfexp xs e
  | IEifnz x ea eb =>
      In (inl x) xs /\ wfexp xs ea /\ wfexp xs eb
  | IEcase x bx ea eb =>
      In (inl x) xs /\ wfexp (inl(inr bx)::xs) ea /\ wfexp (inl(inr bx)::xs) eb
  | IEret xk x =>
      In (inr xk) xs /\ In (inl x) xs
  | IEapp xf x xk =>
      In (inl xf) xs /\ In (inl x) xs /\ In (inr xk) xs
  | IEinst xf xk =>
      In (inl xf) xs /\ In (inr xk) xs
  | IEnew x bx e =>
      In (inl x) xs /\ wfexp (inl (inr bx) :: xs) e
  | IEread x bx e =>
      In (inl x) xs /\ wfexp (inl (inr bx) :: xs) e
  | IEwrite x y e =>
      In (inl x) xs /\ In (inl y) xs /\ wfexp xs e
  | _ => False
  end.

Fixpoint wfexp xs e := wfexp_ wfexp xs e.

Definition wfatom := wfatom_ wfexp.

Inductive bad_fun : val -> Prop :=
| _bad_fun: forall n, bad_fun (IVclo [] (IEfix 0 0 0 (IEnat n))).
Hint Constructors bad_fun.

Inductive bad_gen : val -> Prop :=
| _bad_gen: forall n, bad_gen (IVclo [] (IEgen 0 (IEnat n))).
Hint Constructors bad_gen.

Fixpoint is_good_thunk e : Prop :=
  match e with
  | IEfix _ _ _ _ => ~ bad_fun (IVclo [] e)
  | IEgen _ _ => ~ bad_gen (IVclo [] e)
  | _ => False
  end.

Definition module := list (label * exp).

Inductive mid_wfmod: list label -> module -> list label -> Prop :=
| MHTnil: forall FI,
    mid_wfmod FI nil nil
| MHTcons: forall FI FO f e m
      (FLX: is_good_thunk e)
      (WFE: wfatom (map (fun l => inl (inl l)) FI) e)
      (UNQ: ~ In f FI)
      (WFM: mid_wfmod (FI ++ [f]) m FO),
    mid_wfmod FI ((f,e)::m) (f::FO)
.
Hint Constructors mid_wfmod.

Lemma mid_wfmod_uniq: forall FI m FO f
    (INI: In f FI)
    (INO: In f FO)
    (WFM: mid_wfmod FI m FO),
  False.
Proof.
  i; ginduction WFM; vauto; i.
  s in INO; des; [vauto|]; eapply IHWFM; [apply in_or_app; left|]; eauto.
Qed.


(** ** Structural properties *)


Fixpoint size (e: exp) : nat :=
  match e with
    | IEfix _ _ _ e
    | IEgen _ e
    | IEcont _ e
    | IEnew _ _ e
    | IEread _ _ e
    | IEwrite _ _ e
    | IEinput _ e
    | IEoutput _ e
      => 1 + size e
    | IEifnz _ ea eb
    | IEcase _ _ ea eb
    | IElet ea _ eb
      => 1 + size ea + size eb
    | _
      => 0
  end.


Lemma size_lt_well_founded:
  well_founded (fun e e' => size e < size e').
Proof.
  ii; remember (size a) as n; depgen a.
  induction n using (well_founded_ind lt_wf).
  econs; i; subst; eauto.
Qed.



(** *** Weakening *)


Lemma in_weaken: forall X (x': X) xsa x xsb
    (IN: In x' (xsa++xsb)),
  In x' (xsa++x::xsb).
Proof.
  i; apply in_app_or in IN; apply in_or_app.
  des; [left|do 2 right]; auto.
Qed.

Lemma ini_weaken: forall x' xsa x xsb
    (IN: Ini x' (xsa++xsb)),
  Ini x' (xsa++x::xsb).
Proof.
  destruct x'; s; auto using in_weaken.
Qed.

Lemma purge_purge: forall G,
  purge (purge G) = purge G.
Proof.
  induction G; try destruct a; s; try rewrite IHG; eauto.
Qed.

Lemma purge_app: forall xsa xsb,
  purge (xsa ++ xsb) = purge xsa ++ purge xsb.
Proof.
  induction xsa; [by auto|]; i.
  destruct a; s; vauto.
  rewrite IHxsa; eauto.
Qed.

Lemma purge_map_inl: forall xs,
  purge (map inl xs) = map inl xs.
Proof.
  induction xs; s; try rewrite IHxs; eauto.
Qed.

Lemma purge_inD: forall x G
    (IN: In x (purge G)),
  In x G.
Proof.
  i; induction G; ss.
  destruct a; ss; des; eauto.
Qed.

Lemma purge_inDr: forall x G
    (IN: In (inr x) (purge G)),
  False.
Proof.
  i; induction G; ss.
  destruct a; ss; des; vauto.
Qed.

Lemma purge_sublist: forall xs,
  sublist (purge xs) xs.
Proof.
  induction xs; try destruct a; ss; eauto.
Qed.
Hint Immediate purge_sublist.

Lemma wfatom_weaken_: forall e xsa x xsb
    (WF: wfatom (xsa++xsb) e)
    (IH: forall e', size e' < size e ->
           forall xsa x xsb, wfexp (xsa++xsb) e' -> wfexp (xsa++x::xsb) e'),
  wfatom (xsa++x::xsb) e.
Proof.
  destruct e; vauto; i; ss; des; try split
  ; try (by rrdes WF; s; try split; auto using ini_weaken, in_weaken)
  ; try (by eauto using in_weaken, IH).
  - rewrite purge_app, cons_app3, app_assoc in WF.
    rewrite purge_app, cons_app3, app_assoc.
    destruct x; [|by eauto]; simpl purge; eapply IH; auto; rr; s; omega.
  - rewrite purge_app in WF.
    rewrite purge_app, cons_app, app_assoc.
    destruct x; [|by eauto]; simpl purge; eapply IH; auto; rr; s; omega.
  - rewrite cons_app, app_assoc.
    destruct x; eapply IH; auto; rr; s; omega.
Qed.


Theorem wfexp_weaken: forall e xsa x xsb
    (WF: wfexp (xsa++xsb) e),
  wfexp (xsa++x::xsb) e.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i; ss; des; subst; try repeat split
  ; try (by eauto using in_weaken, IH)
  ; try (by rewrite cons_app, app_assoc; eauto; eapply IH; vauto)
  ; try (by eapply IH; rr; eauto).
  - eapply (IH e1); auto; rr; s; omega.
  - eapply (IH e2); auto; rr; s; omega.
  - rewrite cons_app, app_assoc; eapply (IH e1); auto; rr; s; omega.
  - rewrite cons_app, app_assoc; eapply (IH e2); auto; rr; s; omega.
  - destruct bx; split; des.
    + apply wfatom_weaken_; auto; i; eapply IH; auto; rr; rrdes H; s; omega.
    + rewrite cons_app, app_assoc; eapply (IH e2); auto; rr; s; omega.
    + apply wfatom_weaken_; auto; i; eapply IH; auto; rr; rrdes H; s; omega.
    + rewrite cons_app, app_assoc; eapply (IH e2); auto; rr; s; omega.
Qed.

Theorem wfatom_weaken: forall e xsa x xsb
    (WF: wfatom (xsa++xsb) e),
  wfatom (xsa++x::xsb) e.
Proof.
  i; eapply wfatom_weaken_; eauto.
  i; eapply wfexp_weaken; eauto.
Qed.


(** *** Contraction *)


Lemma in_contract: forall X (x': X) xsa x xsb xsc
    (IN: In x' (xsa++x::xsb++x::xsc)),
  In x' (xsa++xsb++x::xsc).
Proof.
  i; apply in_app_or in IN; apply in_or_app.
  des; [left|right]; auto; s in IN; des; auto.
  subst; apply in_or_app; right; s; auto.
Qed.

Lemma ini_contract: forall x' xsa x xsb xsc
    (IN: Ini x' (xsa++x::xsb++x::xsc)),
  Ini x' (xsa++xsb++x::xsc).
Proof.
  destruct x'; s; auto using in_contract.
Qed.

Lemma wfatom_contract_: forall e xsa x xsb xsc
    (WF: wfatom (xsa++x::xsb++x::xsc) e)
    (IH: forall e', size e' < size e ->
           forall xsa x xsb xsc, wfexp (xsa++x::xsb++x::xsc) e' ->
             wfexp (xsa++xsb++x::xsc) e'),
  wfatom (xsa++xsb++x::xsc) e.
Proof.
  destruct e; vauto; i; ss; des; try split
  ; try (by rrdes WF; s; try split; auto using ini_contract, in_contract)
  ; try (by eauto using in_contract, IH).
  - rewrite (cons_app x xsc), (cons_app x) in WF.
    rewrite !purge_app, cons_app3, app_assoc in WF.
    rewrite 2!purge_app, cons_app3, app_assoc.
    destruct x; [|by eauto]; simpl purge; eapply IH; auto; rr; s; omega.
  - rewrite (cons_app x xsc), (cons_app x) in WF.
    rewrite !purge_app in WF.
    rewrite !purge_app, cons_app, app_assoc.
    destruct x; [|by eauto]; simpl purge; eapply IH; auto; rr; s; omega.
  - rewrite (cons_app x xsc), (cons_app x) in WF.
    rewrite cons_app, app_assoc.
    destruct x; apply IH; auto; rr; s; omega.
Qed.


Theorem wfexp_contract: forall e xsa x xsb xsc
    (WF: wfexp (xsa++x::xsb++x::xsc) e),
  wfexp (xsa++xsb++x::xsc) e.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i; ss; des; subst; try repeat split
  ; try (by eauto using in_contract, IH)
  ; try (by rewrite cons_app, app_assoc; eapply IH; vauto)
  ; try (by eapply IH; rr; eauto).
  - eapply (IH e1); auto; rr; s; omega.
  - eapply (IH e2); auto; rr; s; omega.
  - rewrite cons_app, app_assoc; eapply (IH e1); auto; rr; s; omega.
  - rewrite cons_app, app_assoc; eapply (IH e2); auto; rr; s; omega.
  - destruct bx; split; des.
    + apply wfatom_contract_; auto; i; eapply IH; auto; rr; rrdes H; s; omega.
    + rewrite cons_app, app_assoc; eapply (IH e2); auto; rr; s; omega.
    + apply wfatom_contract_; auto; i; eapply IH; auto; rr; rrdes H; s; omega.
    + rewrite cons_app, app_assoc; eapply (IH e2); auto; rr; s; omega.
Qed.

Lemma wfatom_contract: forall e xsa x xsb xsc
    (WF: wfatom (xsa++x::xsb++x::xsc) e),
  wfatom (xsa++xsb++x::xsc) e.
Proof.
  i; eapply wfatom_contract_; eauto.
  i; eapply wfexp_contract; eauto.
Qed.


(** *** Exchange *)


Lemma ini_exchange: forall xs xs' x
    (EX: Permutation xs xs')
    (IN: Ini x xs),
  Ini x xs'.
Proof.
  destruct x; s; eauto using Permutation_in.
Qed.

Lemma purge_permute: forall xs xs'
    (EX: Permutation xs xs'),
  Permutation (purge xs) (purge xs').
Proof.
  i; induction EX using Permutation_ind_bis; auto.
  - destruct x; s; eauto.
  - destruct x; destruct y; s; eauto.
    eapply perm_trans; [by apply perm_swap|]; eauto.
  - eapply perm_trans; eauto.
Qed.
Hint Resolve purge_permute.

Lemma wfatom_exchange_: forall e xs xs'
    (WF: wfatom xs e)
    (EX: Permutation xs xs')
    (IH: forall e', size e' < size e ->
           forall xs xs', Permutation xs xs' -> wfexp xs e' -> wfexp xs' e'),
  wfatom xs' e.
Proof.
  destruct e; vauto; i; ss; des; try split
  ; try (by rrdes WF; s; try split; eauto using ini_exchange, Permutation_in)
  ; (eapply IH; last (by eauto); auto; rr; s; omega).
Qed.


Theorem wfexp_exchange: forall e xs xs'
    (WF: wfexp xs e)
    (EX: Permutation xs xs'),
  wfexp xs' e.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i; ss; des; subst; try repeat split
  ; try (by eauto using Permutation_in, IH)
  ; try (by eapply IH; rr; eauto).
  - eapply (IH e1); eauto; rr; s; omega.
  - eapply (IH e2); eauto; rr; s; omega.
  - eapply (IH e1); eauto; rr; s; omega.
  - eapply (IH e2); eauto; rr; s; omega.
  - destruct bx; split; des.
    + eapply wfatom_exchange_; eauto; i; eapply IH; eauto; rr; rrdes H; s; omega.
    + eapply (IH e2); eauto; rr; s; omega.
    + eapply wfatom_exchange_; eauto; i; eapply IH; eauto; rr; rrdes H; s; omega.
    + eapply (IH e2); eauto; rr; s; omega.
Qed.

Theorem wfatom_exchange: forall e xs xs'
    (WF: wfatom xs e)
    (EX: Permutation xs xs'),
  wfatom xs' e.
Proof.
  i; eapply wfatom_exchange_; eauto.
  i; eapply wfexp_exchange; eauto.
Qed.



(** * Unique variable assumption.

    Optimizations at the intermediate level may assume that in the
    whole module each variable is bound only once.  This makes the
    implementation of some optimizations much simpler.  CPS will
    establish this property; here we just define it formally.
*)

Fixpoint bvars (e: exp) : list var :=
  match e with
    | IEcase _ bx ea eb
    | IElet ea (inl bx) eb
      => [inl (inr bx)] ++ bvars ea ++ bvars eb

    | IElet ea (inr bk) eb
      => [inr bk] ++ bvars ea ++ bvars eb

    | IEfix bf bx bk e
      => [inl (inr bf); inl (inr bx); inr bk] ++ bvars e

    | IEgen bk e
      => [inr bk] ++ bvars e

    | IEcont bx e
    | IEnew _ bx e
    | IEread _ bx e
    | IEinput bx e
      => [inl (inr bx)] ++ bvars e

    | IEifnz _ ea eb
      => bvars ea ++ bvars eb

    | IEwrite _ _ e
    | IEoutput _ e
      => bvars e

    | _
      => nil
  end.

Fixpoint bvars_mod (m: module) : list var :=
  match m with
    | nil => nil
    | (_, e) :: m' => bvars e ++ bvars_mod m'
  end.

Fixpoint uniq X {M: EqDec X} (l: list X) :=
  match l with
    | nil => True
    | x :: l' => ~ In x l' /\ uniq l'
  end.


Definition uniqvar (m: module) : Prop :=
  uniq (bvars_mod m).

Definition within m n l :=
  m <= n /\
  (forall x (IN: In x l), m <= x) /\
  (forall x (IN: In x l), x < n).


Definition subuniq A {M: EqDec A} xs ys : Prop :=
  forall zs (UN: uniq (zs ++ ys)), uniq (zs ++ xs).


(** ** Properties about [bvars]. *)

Lemma bvars_no_labels: forall e l,
  ~ In (inl (inl l)) (bvars e).
Proof.
  induction e; auto; ss; i; des; eauto; vauto
  ; [| |destruct bx; s in FH; des; vauto]
  ; (apply in_app_or in FH; des; eauto).
Qed.


(** ** Properties about [uniq]. *)

Lemma uniq_sublist: forall X (M: EqDec X) (l' l: list X)
    (UNQ: uniq l')
    (LE: sublist l l'),
  uniq l.
Proof.
  i; ginduction LE; [by vauto|by i; ss; des; auto|].
  i; ss; des; split; eauto using sublist_in.
Qed.

Lemma uniq_perm: forall X {M: EqDec X} (l l': list X)
    (UNQ: uniq l')
    (PER: Permutation l' l),
  uniq l.
Proof.
  i; ginduction PER; auto.
  - i; ss; des; split; i; eauto using Permutation_in, Permutation_sym.
  - i; ss; des; split; i; des; eauto.
Qed.

Lemma uniq_app_le: forall n la lb
    (UNQa: uniq la)
    (UNQb: uniq lb)
    (LEa: forall x (IN: In x la), x < n)
    (LEb: forall x (IN: In x lb), n <= x),
  uniq (la ++ lb).
Proof.
  induction la; vauto; i; ss; des; split; auto.
  i; apply in_app_or in FH; des; auto.
  exploit LEa; eauto; i; exploit LEb; eauto; i.
  omega.
Qed.

Lemma uniq_app_within: forall m ma mb n (la lb: list nat)
    (UNa: uniq la)
    (UNb: uniq lb)
    (BWa: within m ma la)
    (BWb: within mb n lb)
    (LE: ma <= mb),
  uniq (la ++ lb).
Proof.
  induction la; vauto; i; ss; des; rrdes BWa BWb; split.
  - i; apply in_app_or in FH; des; auto.
    hexploit BWa1; [by left; auto|]; inst; rewrite plus_0_r; i.
    hexploit BWb0; eauto; i; inst; omega.
  - eapply uniq_app_le; eauto.
    i; unfold lt in *; rewrite <-LE; ss; auto.
Qed.

Lemma uniq_app: forall X {M: EqDec X} (la lb: list X)
    (UNa: uniq la)
    (UNb: uniq lb)
    (DJ: forall x (INa: In x la) (INb: In x lb), False),
  uniq (la ++ lb).
Proof.
  induction la; vauto.
  s; i; des; split; eauto.
  i; apply in_app_or in FH; des; eauto.
Qed.

Lemma uniq_appD: forall X {M: EqDec X} (la lb: list X)
    (UN: uniq (la ++ lb)),
  uniq la /\
  uniq lb /\
  forall x (INa: In x la) (INb: In x lb), False.
Proof.
  induction la; [by vauto|].
  i; ss; des; exploit IHla; eauto; i; des.
  splits; i; des; subst; eauto using in_or_app.
Qed.

Lemma uniq_sng: forall X {M: EqDec X} (x: X),
  uniq [x].
Proof.
  i; r; auto.
Qed.
Hint Resolve uniq_sng.

Lemma uniq_three: forall n,
  uniq [n; S n; S (S n)].
Proof.
  s; splits; i; des; vauto; omega.
Qed.
Hint Resolve uniq_three.

Lemma uniq_mapD: forall X {M: EqDec X} (xs: list X) Y {N: EqDec Y} (f: X -> Y)
    (UNQ: uniq (map f xs)),
  uniq xs.
Proof.
  i; ginduction xs; auto; i; ss; des; split; eauto.
  i; apply UNQ; apply in_split in FH; des; subst. 
  rewrite map_app; apply in_or_app; s; eauto.
Qed.


(** ** Properties about [subuniq]. *)

Lemma subuniq_refl: forall A {M: EqDec A} xs,
  subuniq xs xs.
Proof.
  i; vauto.
Qed.
Hint Resolve subuniq_refl.

Lemma subuniq_trans: forall A {M: EqDec A} ys xs zs
    (SP: subuniq xs ys)
    (SP': subuniq ys zs),
  subuniq xs zs.
Proof.
  ii; eauto.
Qed.

Lemma subuniq_uniq: forall A {M: EqDec A} (xs xs': list A)
    (SP: subuniq xs' xs)
    (UN: uniq xs),
  uniq xs'.
Proof.
  i; eauto using (SP nil).
Qed.

Lemma subuniq_uniq_in: forall A {M: EqDec A} (xs xs': list A) x
    (SP: subuniq xs' xs)
    (UN: uniq xs)
    (IN: In x xs'),
  In x xs.
Proof.
  i; dup SP; specialize (SP [x]).
  destruct (LEM (In x xs)); vauto.
  hexploit SP; [by split; auto|].
  i; ss; des; vauto.
Qed.

Lemma subuniq_uniq_app: forall A {M: EqDec A} (xs xs' ys ys': list A)
    (UN: uniq (xs ++ ys))
    (SP: subuniq xs' xs)
    (SP: subuniq ys' ys),
  uniq (xs' ++ ys').
Proof.
  i; apply uniq_appD in UN; des.
  eapply uniq_app; eauto using subuniq_uniq, subuniq_uniq_in.
Qed.

Lemma subuniq_app: forall A {M: EqDec A} (xs xs' ys ys': list A)
    (SPx: subuniq xs' xs)
    (SPy: subuniq ys' ys),
  subuniq (xs' ++ ys') (xs ++ ys).
Proof.
  ii.
  apply uniq_appD in UN; des.
  apply uniq_appD in UN0; des.
  eapply uniq_app; eauto.
  - eapply uniq_app; eauto using subuniq_uniq, subuniq_uniq_in.
  - i; eapply UN1; eauto.
    apply in_app_or in INb; des; eauto using in_or_app, subuniq_uniq_in.
Qed.

Lemma subuniq_perm: forall A {M: EqDec A} xs xs'
    (PE: Permutation xs xs'),
  subuniq xs' xs.
Proof.
  ii; eauto using uniq_perm, Permutation_app.
Qed.

Lemma subuniq_sublist: forall A {M: EqDec A} xs xs'
    (PE: sublist xs' xs),
  subuniq xs' xs.
Proof.
  ii; eauto using uniq_sublist, sublist_app.
Qed.


(** ** Properties about [within]. *)

Lemma within_nil: forall m,
  within m m [].
Proof.
  rr; splits; vauto.
Qed.
Hint Resolve within_nil.

Lemma within_sng: forall m,
  within m (S m) [m].
Proof.
  rr; splits; auto; s; i; des; vauto.
Qed.
Hint Resolve within_sng.








Lemma within_three: forall m,
  within m (S (S (S m))) [m; S m; S (S m)].
Proof.
  rr; splits; auto; s; i; des; vauto.
Qed.
Hint Resolve within_three.







Lemma within_app: forall m ma ma' mb mb' n la lb
    (BWa: within ma ma' la)
    (BWb: within mb mb' lb)
    (LEa: m <= ma)
    (LE: ma' <= mb)
    (LEb: mb' <= n),
  within m n (la ++ lb).
Proof.
  i; rrdes BWa BWb.
  rr; splits; eauto using le_trans.
  - i; apply in_app_or in IN; des.
    + rewrite LEa; auto.
    + rewrite LEa, BWa, LE; auto.
  - i; apply in_app_or in IN; des.
    + unfold lt in *; rewrite <-LEb, <-BWb, <-LE; auto.
    + unfold lt in *; rewrite <-LEb; auto.
Qed.

Lemma within_perm: forall xs xs' m n
    (EX: Permutation xs xs')
    (IN: within m n xs'),
  within m n xs.
Proof.
  unfold within; i; des; splits; eauto using Permutation_in.
Qed.

Lemma within_weaken: forall xs m n m' n'
    (WI: within m' n' xs)
    (LE: m <= m')
    (GE: n' <= n),
  within m n xs.
Proof.
  unfold within; i; des; splits
  ; i; [|hexploit WI0|hexploit WI1]; eauto; omega.
Qed.



(** * Implementation of language specification. *)

Section LangSpecImpl.

  Definition mid_val := val.
  Definition mid_cont := val.
  Definition mid_conf := (heap * eoption (env * exp)) % type.
  Definition mid_mach := (heap * (env * exp)) % type.
  Definition mid_module := module.
  Definition mid_anchor := unit.

  Fixpoint enum_exp (e: exp) : nat :=
    match e with
      | IEunit => enum (0, 0)
      | IEnat i => enum (1, i)
      | IEpair x1 x2 => enum (2, enum (x1, x2))
      | IEinl x => enum (3, enum x)
      | IEinr x => enum (4, enum x)
      | IEfst x => enum (6, enum x)
      | IEsnd x => enum (7, enum x)
      | IEfix f x k e => enum (8, enum (f, x, k, enum_exp e))
      | IEgen k e => enum (9, enum (k, enum_exp e))
      | IEcont bx e => enum (10, enum (bx, enum_exp e))
      | IEnew x bx e => enum (11, enum (x, bx, enum_exp e))
      | IEread x bx e => enum (12, enum (x, bx, enum_exp e))
      | IEwrite x y e => enum (13, enum (x, y, enum_exp e))
      | IEinput bx e => enum (14, enum (bx, enum_exp e))
      | IEoutput x e => enum (15, enum (x, enum_exp e))
      | IEifnz x e1 e2 => enum (16, enum (x, enum_exp e1, enum_exp e2))
      | IEcase x bx e1 e2 => enum (17, enum (x, bx, enum_exp e1, enum_exp e2))
      | IEret k x => enum (18, enum (k, x))
      | IEapp x y k => enum (19, enum (x, y, k))
      | IEinst x k => enum (20, enum (x, k))
      | IElet e1 bxk e2 => enum (21, enum (enum_exp e1, bxk, enum_exp e2))
      | IErefeq x y => enum (22, enum (x, y))
      | IEbop Plus a1 a2 => enum (23, enum (a1, a2))
      | IEbop Minus a1 a2 => enum (24, enum (a1, a2))
      | IEbop Mult a1 a2 => enum (25, enum (a1, a2))
      | IEbop Equal a1 a2 => enum (26, enum (a1, a2))
    end.

  Global Instance Enumerable_mid_exp: Enumerable exp :=
    {| enum := enum_exp |}.
  Proof.
    intro e; induction e; intro e'
      ; destruct e'; i; try destruct op; try destruct op0
      ; apply enum_inj in H; inversion H; subst
      ; try (apply enum_inj in H0; inversion H0; subst)
      ; try (apply enum_inj in H1; inversion H1; subst)
      ; try (apply enum_inj in H2; inversion H2; subst)
      ; try (apply enum_inj in H3; inversion H3; subst)
      ; try (apply enum_inj in H4; inversion H4; subst)
      ; try (apply enum_inj in H5; inversion H5; subst)
      ; try (apply enum_inj in H6; inversion H6; subst)
      ; try (erewrite IHe; eauto); try (erewrite IHe1; eauto)
      ; try (erewrite IHe2; eauto); try (erewrite IHe3; eauto)
      ; auto.
  Qed.

  Fixpoint size_val (v: val) : nat :=
    match v with
      | IVunit | IVnat _ | IVloc _ =>
        0
      | IVpair v1 v2 =>
        1 + size_val v1 + size_val v2
      | IVinl v' | IVinr v' =>
        1 + size_val v'
      | IVclo xs e =>
        1 + fold_right plus 0 (map (fun p => size_val (snd p)) xs) + size e
    end.

  Lemma size_val_lt_well_founded:
    well_founded (fun x y => size_val x < size_val y).
  Proof.
    intro v; remember (size_val v) as n; generalize dependent v.
    induction n using (well_founded_ind lt_wf); i; subst.
    econs; i; eauto.
  Qed.

  Program Fixpoint enum_list'' A (l: list A) (enumA: forall a, In a l -> nat) : nat :=
    match l with
      | nil => 0
      | a :: l' => enum (enumA a _, enum_list'' l' (fun x PF => enumA x _))
    end.
  (** The [Prop] argument to [enumA] is needed to get the proper proof
      obligations for [enum_val_] below.  Kudos to Viktor! *)
  Next Obligation. left; auto. Qed.
  Next Obligation. right; auto. Qed.

  Program Definition enum_val_
          (v: val)
          (enum_val: forall v', size_val v' < size_val v -> nat)
    : nat :=
    match v with
      | IVunit => enum (0, 0)
      | IVnat n => enum (1, enum n)
      | IVpair v1 v2 => enum (2, enum (enum_val v1 _, enum_val v2 _))
      | IVinl v => enum (3, enum_val v _)
      | IVinr v => enum (4, enum_val v _)
      | IVclo xs e =>
        enum (5, enum ((length xs,
                        enum_list'' xs (fun p PF => enum (enum (fst p), enum_val (snd p) _))),
                       enum e))
      | IVloc l => enum (6, enum l)
    end.
  Next Obligation. ss; omega. Qed.
  Next Obligation. ss; omega. Qed.
  Next Obligation.
    s; apply in_split in PF; destruct PF as (? & ? & ->).
    rewrite map_app, fold_right_app; s.
    clear enum_val; induction x; i; s; omega.
  Qed.

  
  Lemma enum_val_ext: forall x
      (f g: forall y : val, size_val y < size_val x -> nat)
      (H: forall y (p : size_val y < size_val x), f y p = g y p),
   enum_val_ x f = enum_val_ x g.
  Proof.
    i; induction x; eauto
    ; unfold enum_val_; repeat (f_equal; extensionalities).
  Qed.

  Definition enum_val: val -> nat :=
    Fix size_val_lt_well_founded _ enum_val_.

  Global Instance Enumerable_mid_val: Enumerable val :=
    {| enum := enum_val |}.
  Proof.
    intro v; induction v using (well_founded_ind size_val_lt_well_founded).
    rename H into IH; destruct v; intros v' ?; destruct v'; i
    ; inversion H; subst; auto
    ; repeat match goal with | H : enum_val _ = enum_val _ |- _ =>
        unfold enum_val in H; rewrite ?Init.Wf.Fix_eq in H; auto using enum_val_ext; s in H
      end
    ; repeat match goal with | H : enum _ = enum _ |- _ =>
        apply enum_inj in H; clarify
      end
    ; f_equal; try (apply IH; s; eauto; omega).
    depgen xvs0; induction xvs; destruct xvs0; vauto; i.
    apply enum_inj in H; fold enum_list' in *; clarify.
    unfold id in H2; subst.            
    destruct a, p; apply enum_inj in H4; clarify.
    apply enum_inj in H5; subst; repeat f_equal.
    - apply IH; [by s; omega|]; eauto.
    - apply IHxvs; eauto.
      i; apply IH; eauto.
      eapply lt_le_trans; eauto; inst; s; omega.
  Qed.

  Fixpoint mid_vload' (m: mid_module)
                      (xvs: list (label * val)) lbl : option mid_val :=
    match m with
      | nil => None
      | (l, e) :: m' =>
        let v := IVclo (map_pair (inl<*>inl) id xvs) e in
        if eqdec l lbl
        then Some v
        else mid_vload' m' (xvs ++ [(l, v)]) lbl
    end.

  Lemma mid_vload'_some: forall m xvs f
      (IN: In f (map fst m)),
    exists vm, mid_vload' m xvs f = Some vm.
  Proof.
    induction m; [by vauto|]; i.
    s in IN; des; subst; destruct a; s; des_if; vauto; eauto.
  Qed.

  Definition mid_vload (m: mid_module) (_: mid_anchor)
                       (xvs: list (label * val)) lbl (v: mid_val) :=
    mid_vload' m xvs lbl = Some v.
      
  Definition mid_cload (m: mid_module) (_: mid_anchor)
                       (imports: list (label * val)) (cg cl: mid_conf) :=
    cg = (pmemp, Nothing) /\
    cl = (pmemp, Nothing).

  Definition mid_cplus (c c': mid_conf) :=
    match c, c' with
    | (h, ee), (h', ee') => (pmplus h h', eoplus ee ee')
    end.

  Definition mid_cemp: mid_conf :=
    (pmemp, Nothing).

  Definition mid_halted (m: mid_mach) := 
    exp_halted (snd(snd m)).

  Definition mid_core (c: mid_conf) := 
    exists ee, c = (pmemp, Just ee).

  Definition mid_extra (c: mid_conf) := 
    exists h, c = (h, Nothing).

  Definition mid_real (c: mid_conf) (m: mid_mach) :=
    exists h ee,
      <<Ec: c = (h, Just ee)>> /\ <<Em: m = (h, ee)>> /\
      <<PV: pmvalid h>> /\ <<FIN: pmfinite h>>.

  Definition mid_lsi: LangSpec.
    refine
      {| lang_spec.val := mid_val
       ; cont := mid_cont
       ; conf := mid_conf
       ; mach := mid_mach
       ; lang_spec.module := mid_module
       ; anchor := mid_anchor
       ; vload := mid_vload
       ; cload := mid_cload
       ; cplus := mid_cplus
       ; cemp := mid_cemp
       ; halted := mid_halted
       ; core := mid_core
       ; extra := mid_extra
       ; step := mid_step
       ; real := mid_real
      |}.
    - split.
      + i; destruct m as [? []]; unfold mid_halted in *; ss.
        inversion FH; vauto; induction STEP; vauto.
      + i; destruct m as (h,(xvs,e)).
        r; s; apply NNPP; intro NV.
        destruct (classic (exists evt h' xvs' e', reduce_ok evt h xvs e h' xvs' e')); des; eauto using IRok.
    - destruct a, b as [].
      s; f_equal; eauto using pmplus_comm, eoplus_comm.
    - destruct a, b, c as [].
      s; f_equal; eauto using pmplus_assoc, eoplus_assoc.
    - destruct a; s; rewrite pmplus_emp; destruct e; eauto.
    - i; apply RL; i; rrdes FH0; clarify; inv ST; [|by rrdes PV; vauto].
      inv STEP; eapply FH; rr; esplits; try (by split_pmfinite FIN PV; eauto).
      eapply pmvalid_plus_cell; eauto.
    - i; apply NNPP; ii; apply NE; i; rrdes FH; clarify.
      inv ST; inv STEP; eapply E; rr; esplits; eauto.
    - i; invc COa COb RLa RLb; des; subst.
      destruct ca, cb; ss; noemp; clarify.
      destruct e, e0; vauto; eauto.
    - unfold mid_cemp; r; eauto.
    - i; rrdes XTR XTR'; subst; unfold mid_cplus; r; eauto.
    - set (dummy := (None, (nil, IEunit)) : mid_mach).
      destruct c.
      destruct e; [|M|M]; Mdo (by exists dummy; i; inv XTR; rrdes RL; vauto).
      econs; i; inv XTR; rrdes RL; inv Ec.
      split_pmfinite FIN PV; rr; esplits; eauto.
      
      
      
    - i; destruct m as [? []]; unfold mid_halted in *; ss; des.
      destruct cf; rrdes XTR; clarify.
      destruct mm; rrdes RL; unfold mid_cplus in *; destruct c'; clarify; ss.
      destruct e1; ss; clarify; rrdes RLm; vauto.
  Defined.

End LangSpecImpl.

Hint Unfold mid_core.



(** * Weakly deterministic reduction.

    All configurations whose expression is neither [IEnew] nor
    [IEinput] are weakly deterministic (as defined in [lang_spec]).
*)

(** Syntactic criterion. *)
Fixpoint mid_exp_wdet e :=
  match e with
    | IEnew _ _ _ | IEinput _ _ => False
    | _ => True
  end.

(** Preservation of [reduce_ok] under configuration-extension. *)
Lemma step_ok_frame: forall evt h env e h' env' e' cf h0 env0 e0
    (DE: mid_exp_wdet e)
    (ST: reduce_ok evt h env e h' env' e')
    (PL: mcplus mid_lsi (h, (env, e)) cf (h0, (env0, e0))),
  reduce_ok evt h0 env0 e0 (h' ## fst cf) env' e'.
Proof.
  i; rrdes PL RL RLm; destruct c', cf.
  destruct e1; vauto; destruct e2; vauto; ss; clarify.
  invc ST; vauto; try rewrite <-!pmplus_assoc in *; eauto using pmread_plus.
Qed.

(** Syntactic criterion implies semantic one. *)
Lemma mid_wdet: forall h env e
    (DE: mid_exp_wdet e),
  @wdet mid_lsi (h, (env, e)).
Proof.
  ii; invc STa; [|by exfalso; rrdes RL; unfold pmvalid in *; vauto].
  invc STb; [|by exfalso; eapply STK, step_ok_frame; eauto].
  rrdes PL RL0 RLm RL; destruct c', cf.
  destruct e1; vauto; destruct e2; vauto; ss; clarify.
  invc STEP; invc STEP0; split; auto; try (by rr; esplits; eauto).
  - rr; esplits; eauto.
    eapply pmread_plus in LOAD; eauto.
    hexploit (pmread_uniq LOAD LOAD0); intro E; inst; subst; auto.
  - rewrite <-!pmplus_assoc in *; repeat f_equal.
    apply pmcell_uniq in H1; des; subst.
    + rr; esplits; eauto; [by eapply pmvalid_plus_cell; eauto|].
      apply pmfinite_plusD in FIN; auto; des.
      apply pmfinite_plus; auto; apply pmfinite_cell.
    + rewrite <-H1 in PV; eapply pmvalid_plus_cell; eauto.
Qed.



(** * Misc. *)

Definition tkvar2var (x: termvar + contvar) : var :=
  match x with
    | inl xt => inl (inr xt)
    | inr xk => inr xk
  end.

Fixpoint purge_lbl (xs: list var) : list var :=
  match xs with
    | nil => nil
    | inl (inl _) :: xs' => purge_lbl xs'
    | x :: xs' => x :: purge_lbl xs'
  end.

Lemma purge_lbl_sublist: forall xs ys
    (SL: sublist xs ys),
  sublist (purge_lbl xs) (purge_lbl ys).
Proof.
  i; ginduction ys; s; i; [by inv SL|].
  destruct a; [destruct s|]; invc SL; s; eauto.
Qed.

Lemma purge_lbl_tkvar2var: forall x xs
    (IN: In (tkvar2var x) xs),
  In (tkvar2var x) (purge_lbl xs).
Proof.
  induction xs; vauto; i.
  apply in_consD in IN; des; subst.
  - destruct x; vauto.
  - destruct a; [destruct s|]; s; eauto.
Qed.

Lemma purge_lbl_sub G : sublist (purge_lbl G) G.
Proof. induction G; eauto; destruct a as [[]|]; s; eauto. Qed.

Lemma purge_lbl_cons x G : purge_lbl (tkvar2var x::G) = tkvar2var x :: purge_lbl G.
Proof. by destruct x. Qed.

Lemma purge_lbl_c : forall G x
  (INx : In (inr x) G), In (inr x) (purge_lbl G).
Proof. ii; fold (tkvar2var (inr x)). by eapply purge_lbl_tkvar2var. Qed.
Hint Resolve purge_lbl_c.

Lemma purge_lbl_t : forall G x
  (INx : In (inl (inr x)) G), In (inl (inr x)) (purge_lbl G).
Proof. ii; fold (tkvar2var (inl x)). by eapply purge_lbl_tkvar2var. Qed.
Hint Resolve purge_lbl_t.

Lemma purge_subinDl x G : 
  In (inl x) G -> In (inl x) (purge G).
Proof. induction G; ii; eauto. apply in_consD in H; des; clarify; s; eauto. destruct a; s; eauto. Qed.


Lemma purge_purge_lbl G :
  purge_lbl (purge G) = purge (purge_lbl G).
Proof. induction G; s; eauto. destruct a as [[a|a]|a]; s; by rewrite IHG. Qed.


  

Lemma purge_lbl_map_inl_inl: forall xs,
  purge_lbl (map (inl <*> inl) xs) = nil.
Proof.
  induction xs; eauto.
Qed.

Lemma mid_real_inj: forall ea eb m
    (RLa: real mid_lsi ea m)
    (RLb: real mid_lsi eb m),
  ea = eb.
Proof.
  i; rrdes RLa RLb; clarify.
Qed.

(** Useful when about to apply pesim_release. *)
Lemma mid_csplit: forall h ee,
  (h, Just ee) =
  @cplus mid_lsi (pmemp, Just ee) (h, Nothing).
Proof.
  i; s; noemp; auto.
Qed.


Lemma mid_error: forall ee,
   error mid_lsi (None, ee).
Proof.
  ii; rrdes FH; clarify; rrdes PV; auto.
Qed.
Hint Resolve mid_error.


Lemma mid_wfmod_export: forall FI m FO f
    (WF: mid_wfmod FI m FO)
    (IN: In f FO),
  In f (map fst m).
Proof.
  i; ginduction WF; [by auto|]; i; ss; des; eauto.
Qed.

Lemma wfexp_good_fix: forall G bf bx bk e
    (WF: wfexp G e),
  is_good_thunk (IEfix bf bx bk e).
Proof.
  ii; inv FH; destruct e.
Qed.

Lemma wfexp_good_gen: forall G bk e
    (WF: wfexp G e),
  is_good_thunk (IEgen bk e).
Proof.
  ii; inv FH; destruct e.
Qed.

Definition inl_sng X Y (a: X + Y) :=
  match a with
    | inl x => [x]
    | _ => []
  end.

Definition argvars (e: exp) : list var :=
  match e with
  | IEgen bk _
  | IElet _ (inr bk) _ => [inr bk]

  | IEfix bf bx bk _ => [inl (inr bf); inl (inr bx); inr bk]

  | IElet _ (inl bx) _
  | IEcont bx _ 
  | IEnew _ bx _ 
  | IEread _ bx _ 
  | IEinput bx _ 
  | IEcase _ bx _ _ => [inl (inr bx)]

  | _  => []
  end.


Fixpoint vars (e: exp) : list var :=
  match e with
    | IElet ea _ eb
      => vars ea ++ vars eb

    | IEpair x y
      => [inl x; inl y]

    | IEret k x
    | IEinst x k
      => [inl x; inr k]

    | IEinl x
    | IEinr x
    | IEfst x
    | IEsnd x
      => [inl x]

    | IEbop _ xn yn
      => map inl (inl_sng xn) ++ map inl (inl_sng yn)

    | IEapp x y k
      => [inl x; inl y; inr k]

    | IEnew x _ ea
    | IEread x _ ea
    | IEoutput x ea
      => [inl x] ++ vars ea

    | IEwrite x y ea
      => [inl x; inl y] ++ vars ea

    | IEgen _ ea
    | IEcont _ ea
    | IEinput _ ea
    | IEfix _ _ _ ea
      => vars ea

    | IEifnz x ea eb
    | IEcase x _ ea eb
      => [inl x] ++ vars ea ++ vars eb
                 
    | IErefeq x y
      => [inl x; inl y]

    | _
      => []
  end.

Lemma wfatom_strengthen_: forall e Gl Gr x
    (WF: wfatom (Gl++x::Gr) e)
    (NF: ~ In x (vars e))
    (IH: forall e', size e' < size e ->
           forall Gl Gr x, wfexp (Gl++x::Gr) e' -> ~ In x (vars e') ->
             wfexp (Gl++Gr) e'),
  wfatom (Gl++Gr) e.
Proof.
  destruct e; vauto; i; s in WF; des;
  try (by apply in_app_or in WF; try apply in_app_or in WF0
       ; ss; des; try (by exfalso; eauto)
       ; s; splits; eauto using in_or_app).
  - ss; des; destruct x, y; ss
    ; try apply in_app_or in WF; try apply in_app_or in WF0
    ; ss; des; try (by exfalso; eauto)
    ; s; splits; eauto using in_or_app.
  - ss; des; rewrite purge_app in WF.
    destruct x; ss; rewrite purge_app; [|by eauto].
    rewrite cons_app3, app_assoc; eapply IH; s; eauto.
  - ss; des; rewrite purge_app in WF.
    destruct x; ss; rewrite purge_app; [|by eauto].
    rewrite cons_app, app_assoc; eapply IH; s; eauto.
  - ss; des.
    destruct x; ss; rewrite cons_app, app_assoc; eapply IH; s; eauto.
Qed.

Lemma wfexp_strengthen: forall e Gl Gr x
    (WF: wfexp (Gl++x::Gr) e)
    (NF: ~ In x (vars e)),
  wfexp (Gl++Gr) e.
Proof.
  intro; remember (size e) as n; generalize dependent e.
  induction n using (well_founded_ind lt_wf).
  destruct e; vauto; i; ss; des
  ; try (by apply in_app_or in WF
         ; rewrite cons_app, app_assoc in WF0
         ; hexploit (H (size e)); eauto; [omega|]
         ; ss; des; try (by exfalso; eauto); eauto using in_or_app)
  ; try (by apply in_app_or in WF; apply in_app_or in WF0
         ; try (apply in_app_or in WF1)
         ; try (hexploit (H (size e)); eauto; [omega|])
         ; ss; des; try (by exfalso; eauto); eauto 10 using in_or_app).
  - rewrite app_comm_cons in WF.
    hexploit (H (size e)); eauto; omega.
  - apply in_app_or in WF.
    hexploit (H (size e)); eauto; [omega|].
    ss; des; try (by exfalso; eauto); eauto using in_or_app.
  - apply in_app_or in WF.
    hexploit (H (size e1)); eauto using in_or_app; [omega|].
    hexploit (H (size e2)); eauto using in_or_app; [omega|].
    ss; des; try (by exfalso; eauto); eauto using in_or_app.
  - apply in_app_or in WF.
    rewrite cons_app, app_assoc in WF0.
    rewrite cons_app, app_assoc in WF1.
    hexploit (H (size e1)); eauto using in_or_app; [omega|].
    hexploit (H (size e2)); eauto using in_or_app; [omega|].
    ss; des; try (by exfalso; eauto); eauto using in_or_app.
  - destruct bx; des; split
    ; try (by hexploit wfatom_strengthen_; eauto using in_or_app
           ; i; eapply H; eauto; omega)
    ; rewrite cons_app, app_assoc in WF0; hexploit (H (size e2))
    ; eauto using in_or_app; omega.
Qed.

Lemma wfatom_strengthen: forall e Gl Gr x
    (WF: wfatom (Gl++x::Gr) e)
    (NF: ~ In x (vars e)),
  wfatom (Gl++Gr) e.
Proof.
  i; eapply wfatom_strengthen_; eauto.
  i; eapply wfexp_strengthen; eauto.
Qed.

Lemma wfexp_strengthen': forall e G x
    (WF: wfexp G e)
    (NF: ~ In x (vars e)),
  wfexp (list_minus eqdec G [x]) e.
Proof.
  i; remember (length G) as n; generalize dependent G.
  induction n using (well_founded_ind lt_wf); i; subst.
  destruct (LEM (In x G)).
  - apply in_split in i; des; subst.
    rewrite cons_app, !list_minus_app, list_minus_nil, app_nil_l, <-list_minus_app.
    eapply H; eauto using wfexp_strengthen.
    rewrite !app_length; s; omega.
  - rewrite list_minus_notin; eauto.
Qed.






Definition list_disjb (l1 l2 : list var) : bool := 
  forallb (fun v => negb (list_mem eqdec v l2)) l1.

Lemma list_disjb_mem (l1 l2 : list var) : 
  list_disjb l1 l2 -> 
  forall x, In x l1 -> In x l2 -> False.
Proof.
  unfold list_disjb. destruct (forallb_forall (fun v => negb (list_mem eqdec v l2)) l1) as [F1 F2].
  i. assert (F:=F1 H). generalize (F x H0). i. eapply list_mem_false; try eassumption. eapply negb_true_iff.
  exact H2.
Qed.


Lemma wfatom_STRENGTHEN (e : exp) (Gl G Gr : list var):
   wfatom (Gl ++ G ++ Gr) e -> (forall x, In x G -> In x (vars e) -> False) -> wfatom (Gl ++ Gr) e.
Proof.
  i; revert e Gl Gr H H0. induction G; i; first by auto. s in H0. 
  apply IHG; eauto. eapply (wfatom_strengthen _ _ _ a); eauto.
Qed.

Lemma purge_sublist_ G G' :
  sublist G G' -> sublist (purge G) (purge G').
Proof. i; induction H; eauto; destruct x; s; eauto. Qed.

Lemma wfexp_sublist_ Gr Gr' e:
  (wfexp Gr e -> sublist Gr Gr' -> wfexp Gr' e)
  /\ (wfatom Gr e -> sublist Gr Gr' -> wfatom Gr' e).
Proof.
  revert Gr Gr'; induction e; s; splits; i; try (destruct bx); des; splits;
  try destruct x; try destruct y; simpl in *; try tauto; eauto using sublist_in;
  (eapply IHe||eapply IHe1||eapply IHe2); eauto 10 using sublist_cons_keep, purge_sublist_.
Qed. 

Lemma wfexp_sublist Gr Gr' e:
  wfexp Gr e -> sublist Gr Gr' -> wfexp Gr' e.
Proof. i. eapply wfexp_sublist_; eauto. Qed.

Lemma wfatom_sublist Gr Gr' e:
  wfatom Gr e -> sublist Gr Gr' -> wfatom Gr' e.
Proof. i. eapply wfexp_sublist_; eauto. Qed.

Lemma wfexp_vars e :
  (forall G, wfexp G e -> forall z (INxV : In z (vars e)) (NINxBV : In z (bvars e) -> False), In z G) /\
  (forall G, wfatom G e -> forall z (INxV : In z (vars e)) (NINxBV : In z (bvars e) -> False), In z G).
Proof.
  induction e; s; splits; ii; des; vauto; 
  try (apply in_app_or in INxV; des; eauto);
  try (destruct (bx:termvar+contvar) as [bx1|bx1]); des; s in NINxBV;
  try (by eapply IHe; eauto using in_or_app);
  try (by eapply IHe1; eauto using in_or_app);
  try (by eapply IHe2; eauto using in_or_app);
  try (by eapply IHe3; eauto using in_or_app);
  try (by try destruct x; try destruct y; simpl in *; des; clarify);
  try (by 
         (cut (In z (inr bk :: (inl (inr bx)) :: (inl (inr bf)) :: G)); 
          first ( s; ii; des; clarify; eauto;try by (exfalso; eauto 10));
          by (eapply IHe; eauto using in_or_app; 
              eapply wfexp_sublist; eauto using sublist_cons_keep, purge_sublist))
    );
  try (by 
         (cut (In z (inr bk :: G)); 
          first ( s; ii; des; clarify; eauto; try by (exfalso; eauto 10));
          by (eapply IHe; eauto using in_or_app; 
              eapply wfexp_sublist; eauto using sublist_cons_keep, purge_sublist))
    );
  try (by 
         (cut (In z (inl (inr bx) :: G)); 
          first ( s; ii; des; clarify; eauto; by (exfalso; eauto 10));
          eauto 7 using in_or_app, wfexp_sublist, sublist_cons_keep)
    );
  try (by 
         (cut (In z (inl (inr bx1) :: G)); 
          first ( s; ii; des; clarify; eauto; by (exfalso; eauto 10));
          eauto 8 using in_or_app, wfexp_sublist, sublist_cons_keep)
    );
  try (by 
         (cut (In z (inr bx1 :: G)); 
          first ( s; ii; des; clarify; eauto; by (exfalso; eauto 10));
          eauto 8 using in_or_app, wfexp_sublist, sublist_cons_keep)
    ).
Qed.

Lemma wfexp_dup : forall x G e (IN: In x G),
  (wfexp (x::G) e -> wfexp G e) /\
  (wfatom (x::G) e -> wfatom G e).
Proof.
  ii. apply in_split in IN; des; clarify.
  rewrite <- (app_nil_l (x::_)); splits.
  - eapply wfexp_contract.
  - eapply wfatom_contract.
Qed. 
