(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool Image List Program Omega ClassicalChoice.
Require Import vgtac basis countable.
Require Export lang_common types lang_spec lang_src.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".


(** Language Spec Implementation *)



(** closedness is checked separately *)
Definition src_val := val.

Definition src_cont := cont.
Definition src_conf := (heap * eoption env * eoption exp) % type.
Definition src_mach := (heap * env * exp) % type.
Definition src_module := module.
Definition src_anchor := unit.


Fixpoint enum_exp (e: exp) : nat :=
  match e with
    | Einput => enum (0, 0)
    | Eoutput e => enum (1, enum_exp e)
    | Evar x => enum (2, enum x)
    | Elbl l => enum (3, enum l)
    | Eunit => enum (4, 0)
    | Enat i => enum (5, enum i)
    | Eifnz e1 e2 e3 => enum (7, enum (enum_exp e1, enum_exp e2, enum_exp e3))
    | Epair e1 e2 => enum (8, enum (enum_exp e1, enum_exp e2))
    | Efst e => enum (9, enum_exp e)
    | Esnd e => enum (10, enum_exp e)
    | Einl e => enum (11, enum_exp e)
    | Einr e => enum (12, enum_exp e)
    | Ecase e1 x e2 e3 => enum (13, enum (x, enum_exp e1, enum_exp e2, enum_exp e3))
    | Eroll e => enum (14, enum_exp e)
    | Eunroll e => enum (15, enum_exp e)
    | Efix f x e => enum (16, enum (f, x, enum_exp e))
    | Eapp e1 e2 => enum (17, enum (enum_exp e1, enum_exp e2))
    | Egen e => enum (18, enum_exp e)
    | Einst e => enum (19, enum_exp e)
    | Epack e => enum (20, enum_exp e)
    | Eunpack x e1 e2 => enum (21, enum (x, enum_exp e1, enum_exp e2))
    | Eloc l => enum (22, l)
    | Enew e => enum (23, enum_exp e)
    | Eread e => enum (24, enum_exp e)
    | Ewrite e1 e2 => enum (25, enum (enum_exp e1, enum_exp e2))
    | Erefeq e1 e2 => enum (26, enum (enum_exp e1, enum_exp e2))
    | Ebop Plus e1 e2 => enum (27, enum (enum_exp e1, enum_exp e2))
    | Ebop Minus e1 e2 => enum (28, enum (enum_exp e1, enum_exp e2))
    | Ebop Mult e1 e2 => enum (29, enum (enum_exp e1, enum_exp e2))
    | Ebop Equal e1 e2 => enum (30, enum (enum_exp e1, enum_exp e2))
  end.

Instance Enumerable_src_exp: Enumerable exp :=
  {| enum := enum_exp |}.
Proof.
  intro e; induction e; intro e'
  ; destruct e'; i; try destruct op; try destruct op0
  ; apply enum_inj in H; fold enum_exp in H; inversion H; subst
  ; try (apply enum_inj in H0; fold enum_exp in H0; inversion H0; subst)
  ; try (apply enum_inj in H1; fold enum_exp in H1; inversion H1; subst)
  ; try (apply enum_inj in H2; fold enum_exp in H2; inversion H2; subst)
  ; try (apply enum_inj in H3; fold enum_exp in H3; inversion H3; subst)
  ; try (apply enum_inj in H4; fold enum_exp in H4; inversion H4; subst)
  ; try (erewrite IHe; eauto); try (erewrite IHe1; eauto)
  ; try (erewrite IHe2; eauto); try (erewrite IHe3; eauto)
  ; auto.
Qed.

Instance Enumerable_src_val: Enumerable val :=
  {| enum := enum <*> e_of |}.
Proof.
  ii; ss; apply enum_inj in H; destruct x, y; ss; subst.
  f_equal; auto using proof_irrelevance.
Qed.


Definition src_link (ma mb: src_module) : src_module := ma ++ mb.

Definition src_vload (m: src_module) (_: src_anchor) (_: list (label * val)) lbl (v: val) :=
  lookup lbl m = Some v.

Definition src_cload (m: src_module) (_: src_anchor) (imports: list (label * val)) (cg cl: src_conf) :=
  exists envf,
    <<Ecg: cg = (pmemp, Just (imports ++ m ++ envf), Nothing)>> /\
    <<Ecl: cl = (pmemp, Nothing, Nothing)>>.

Definition src_cplus (c c': src_conf) :=
  match c, c' with
    | (h, env, e), (h', env', e') => (h ## h', eoplus env env', eoplus e e')
  end.

Definition src_cemp : src_conf :=
  (pmemp, Nothing, Nothing).

Definition src_halted (m: src_mach) :=
  is_value (snd m).

Definition src_core (c: src_conf) :=
  exists e, c = (pmemp, Nothing, Just e).
Hint Unfold src_core.

Definition src_extra (c: src_conf) :=
  exists h, c = (h, Nothing, Nothing).

Definition src_real (c: src_conf) (m: src_mach) :=
  exists h env e,
    <<Ec: c = (h, Just env, Just e)>> /\ <<Em: m = (h, env, e)>> /\
    <<PV: pmvalid h>> /\ <<FIN: pmfinite h>>.

Definition src_lsi: LangSpec.
  refine
    {| lang_spec.val := src_val
     ; lang_spec.cont := src_cont
     ; conf := src_conf
     ; mach := src_mach
     ; lang_spec.module := src_module
     ; anchor := src_anchor
     ; vload := src_vload
     ; cload := src_cload
     ; cplus := src_cplus
     ; cemp := src_cemp
     ; halted := src_halted
     ; core := src_core
     ; extra := src_extra
     ; step := src_step
     ; real := src_real
    |}.
  - split.
    + i; destruct m as [[]]; unfold src_halted in *; ss.
      inversion FH; vauto; induction RED; vauto; [by inv RED; vauto|].
      apply IHRED; i; destruct SK; vauto; ss; des; vauto.
    + i; destruct m as [[h env] e].
      r; s; apply NNPP; intro NV.
      destruct (classic(exists evt h' e', reduce_ok evt env h e h' e' )); eauto using Rok.
  - destruct a as [[]]; destruct b as [[]]; s.
    f_equal; [f_equal|]; eauto using pmplus_comm, eoplus_comm.
  - destruct a as [[]]; destruct b as [[]]; destruct c as [[]]; s.
    f_equal; [f_equal|]; eauto using pmplus_assoc, eoplus_assoc.
  - destruct a as [[]]; s.
    rewrite pmplus_emp; destruct e, e0; eauto.
  - i; inv ST; [|by apply RL; i; rrdes FH0 PV; vauto].
    clear ST; induction RED.
    + inv RED; apply RL; i; rrdes FH0; clarify; eapply FH
      ; rr; esplits; try (by split_pmfinite FIN PV; eauto).
      { eapply pmvalid_plus_cell; by eauto. }
    + apply RL; i; rrdes FH0; clarify; eapply FH; rr; esplits; eauto.
    + apply IHRED.
      i; apply RL; i; rrdes FH1; clarify; eapply FH0; rr; esplits; eauto.
      i; rrdes FH0; clarify; eapply FH; rr; esplits; eauto.
  - i; inv ST; vauto; clear ST; induction RED; auto.
    + apply NNPP; ii.
      inv RED; apply NE; i; rrdes FH; clarify; eapply E; rr; esplits; eauto.
    + apply IHRED; auto.
      * i; rrdes FH; clarify; eapply E; rr; esplits; eauto.
      * i; apply NE; i; rrdes FH0; clarify; eapply FH; rr; esplits; eauto.
  - i; invc COa COb RLa RLb; des; subst.
    destruct ca as [[]], cb as [[]]; ss.
    destruct e2, e4; noemp; clarify; vauto.
  - unfold src_cemp; r; eauto.
  - i; rrdes XTR XTR'; subst.
    unfold src_extra, src_cplus; s; eauto.
  - i; set (dummy := (None, nil, Eunit) : src_mach).
    destruct c as [[]].
    destruct e; [|M|M]; Mdo (by exists dummy; i; inv XTR; rrdes RL; vauto).
    destruct e0; [|M|M]; Mdo (by exists dummy; i; inv XTR; rrdes RL; vauto).
    econs; i; inv XTR; rrdes RL; inv Ec.
    split_pmfinite FIN PV; rr; esplits; eauto.
    
    
    
  - i; destruct m as [[]]; unfold src_halted in *; ss; des.
    destruct cf as [[]]; rrdes XTR; clarify.
    destruct mm; rrdes RL; unfold src_cplus in *; destruct c' as [[]].
    destruct e3, e4; vauto; rrdes RLm; vauto.
Defined.

Definition src_wfmod (m: src_module) exports : Prop :=
  <<OUT: exports = map fst m>>.



(** * Misc. *)

(** Useful when about to apply pesim_release. *)
Lemma src_csplit: forall h en ex,
  (h, Just en, Just ex) =
  @cplus src_lsi (pmemp, Nothing, Just ex) (h, Just en, Nothing).
Proof.
  i; s; noemp; auto.
Qed.
