(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List NPeano Omega Program ClassicalChoice.
Require Import ssr vgtac basis myclassical lang_common lang_spec countable.
Require Import Wellfounded Relation_Operators.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".



(** * Auxiliaries. *)

Definition lt_natpair :=
  tc (symprod _ _ lt lt).

Lemma wf_natpair: well_founded lt_natpair.
  exact (tc_well_founded (wf_symprod _ _ _ _ lt_wf lt_wf)).
Qed.


Lemma pow_pos: forall x y,
  0 < S x ^ y.
Proof.
  i; assert (X := Nat.pow_eq_0_iff (S x) y).
  destruct (S x ^ y); [|omega].
  des; destruct (X eq_refl); vauto.
Qed.


Lemma shiftr_lt: forall x n
    (NZ: x <> 0),
  shiftr x (S n) < x.
Proof.
  i; rewrite Nat.shiftr_div_pow2; apply Nat.div_lt; s; [omega|].
  assert (X:= pow_pos 1 n); omega.
Qed.


Notation "'mdo' X <- A ; B" :=
  (match A with | Some X => B | None => None end)
(at level 200, X ident, A at level 100, B at level 200).


Lemma land_ones: forall m n,
  land (Nat.ones m) (Nat.ones n) = Nat.ones (min m n).
Proof.
  i; Nat.bitwise.
  destruct (LEM (m0 < min m n)).
  - do 3 (rewrite Nat.ones_spec_low; [|by eauto with arith]); auto.
  - destruct (LEM (m0 < m)).
    + rewrite Nat.ones_spec_low; [|by eauto with arith].
      destruct (LEM (m0 < n)).
      * exfalso; auto using Nat.min_glb_lt.
      * do 2 (rewrite Nat.ones_spec_high; [|by apply not_lt]); auto.
    + rewrite Nat.ones_spec_high; [|by apply not_lt].
      destruct (LEM (m0 < n)).
      * rewrite Nat.ones_spec_low; [|by eauto with arith].
        rewrite Nat.ones_spec_high; [|by apply not_lt]; auto.
      * do 2 (rewrite Nat.ones_spec_high; [|by apply not_lt]; auto).
Qed.


Notation word := nat.  

Notation impossible := None.


(** ** Last [n] bits of [x]. *)

Definition last_bits (n: nat) (x: nat) : nat :=
  land x (Nat.ones n).

Lemma testbit_last_bits_low: forall n x m
    (LT: m < n),
  testbit (last_bits n x) m = testbit x m.
Proof.
  unfold last_bits; i.
  rewrite Nat.land_spec Nat.ones_spec_low; auto.
  destruct testbit; auto.
Qed.

Lemma testbit_last_bits_high: forall n x m
    (NLT: ~ m < n),
  testbit (last_bits n x) m = false.
Proof.
  unfold last_bits; i.
  rewrite Nat.land_spec Nat.ones_spec_high; [|by apply not_lt].
  destruct testbit; auto.
Qed.

Lemma last_bits_lor: forall n x y,
  last_bits n (lor x y) = lor (last_bits n x) (last_bits n y).
Proof.
  unfold last_bits; i; auto using Nat.land_lor_distr_l.
Qed.

Lemma last_bits_shiftl: forall n m x
    (LE: n <= m),
  last_bits n (shiftl x m) = 0.
Proof.
  i; Nat.bitwise; destruct (LEM (m0 < n)).
  - rewrite testbit_last_bits_low; auto.
    rewrite shiftl_spec_low; eauto with arith.
  - rewrite testbit_last_bits_high; auto.
Qed.

Lemma last_bits_zero: forall n x
    (E: last_bits n x = 0),
  x = 0 \/ n <= log2 x.
Proof.
  unfold last_bits; i.
  destruct (LEM (n <= log2 x)); [by auto|].
  rewrite Nat.land_ones_low in E; [by auto|].
  apply not_le in f; auto.
Qed.

Lemma last_bits_last_bits: forall n m x
    (LE: n <= m),
  last_bits n (last_bits m x) = last_bits n x.
Proof.
  unfold last_bits; i.
  rewrite -Nat.land_assoc land_ones min_r; auto.
Qed.

Lemma shiftr_last_bits: forall n x m
    (LE: n <= m),
  shiftr (last_bits n x) m = 0.
Proof.
  i; Nat.bitwise.
  rewrite shiftr_spec testbit_last_bits_high; auto; i; omega.
Qed.

Lemma split_last_bits: forall n x,
  x = lor (shiftl (shiftr x n) n) (last_bits n x).
Proof.
  i; Nat.bitwise; destruct (LEM (m < n)).
  - rewrite shiftl_spec_low; auto.
    rewrite testbit_last_bits_low; auto.
  - rewrite testbit_last_bits_high; auto.
    apply not_lt in f; rewrite shiftl_spec_high; auto.
    rewrite shiftr_spec; replace (_ + _) with m; [|omega].
    destruct testbit; vauto.
Qed.


(** ** Encoding/decoding a pair [p] whose first component has at most [len1] bits. *)

Definition concat len1 p : nat :=
  let '(n1, n2) := p in
  lor n1 (shiftl n2 len1).

Arguments concat _ _ : simpl never.

Definition uncat len1 n : nat * nat :=
  (last_bits len1 n, shiftr n len1).

Lemma uncat_concat: forall len1 n1 n2
    (E: n1 = last_bits len1 n1),
  uncat len1 (concat len1 (n1, n2)) = (n1, n2).
Proof.
  i; rewrite E; unfold concat, uncat; f_equal.
  - rewrite last_bits_lor last_bits_shiftl; auto.
    by rewrite Nat.lor_0_r last_bits_last_bits.
  - rewrite Nat.shiftr_lor Nat.shiftr_shiftl_r; [|by auto].
    rewrite shiftr_last_bits; auto.
    by rewrite minus_diag Nat.shiftr_0_r Nat.lor_0_l.
Qed.



Lemma concat_uncat: forall len1 n n1 n2
    (E: uncat len1 n = (n1, n2)),
  n = concat len1 (n1, n2).
Proof.
  i; unfold concat, uncat in *; clarify.
  rewrite Nat.lor_comm; auto using split_last_bits.
Qed.

Lemma concat_inj: forall len1 n1 n2 n1' n2'
    (E1: n1 = last_bits len1 n1)
    (E1': n1' = last_bits len1 n1')
    (E: concat len1 (n1, n2) = concat len1 (n1', n2')),
  (n1, n2) = (n1', n2').
Proof.
  i; rewrite -(uncat_concat _ _ E1) -(uncat_concat _ _ E1'); auto.
Qed.

Lemma uncat_zero: forall len1 n
    (E: uncat len1 n = (0, 0)),
  n = 0.
Proof.
  unfold uncat; i; clarify.
  apply Nat.shiftr_eq_0_iff in H0; des; [by auto|].
  apply last_bits_zero in H1; des; omega.
Qed.

Lemma split_concat: forall x n,
  concat n (last_bits n x, shiftr x n) = x.
Proof.
  i; Nat.bitwise; destruct (LEM (m < n)); unfold concat.
  - rewrite lor_spec shiftl_spec_low; auto.
    rewrite testbit_last_bits_low; auto.
    destruct testbit; vauto.
  - rewrite lor_spec testbit_last_bits_high; auto.
    apply not_lt in f; rewrite shiftl_spec_high; auto.
    rewrite shiftr_spec; replace (_ + _) with m; [|omega].
    destruct testbit; vauto.
Qed.


(** ** Encoding/decodig a pair of unbounded numbers *)

Program Definition encode_pair_ p
    (encode_pair: forall p', lt_natpair p' p -> nat) :=
  match eqdec (fst p) 0, eqdec (snd p) 0 with
    | left _, left _ => 0
    | _, _ =>
      concat 2 (concat 1 (last_bits 1 (fst p), last_bits 1 (snd p)),
                encode_pair (shiftr (fst p) 1, shiftr (snd p) 1) _)
  end.
Next Obligation.
  ss; destruct (eq_nat_dec n 0); destruct (eq_nat_dec n0 0); subst.
  - exfalso; apply (H eq_refl eq_refl); auto.
  - assert (shiftr n0 1 < n0) by (apply shiftr_lt; auto).
    econs 1; econs 2; eauto.
  - assert (shiftr n 1 < n) by (apply shiftr_lt; auto).
    econs 1; econs 1; eauto.
  - assert (shiftr n0 1 < n0) by (apply shiftr_lt; auto).
    assert (shiftr n 1 < n) by (apply shiftr_lt; auto).
    econs 2; econs 1; [econs 1|econs 2]
    ; rewrite Nat.shiftr_div_pow2; apply Nat.div_lt; s; omega.
Qed.

Definition encode_pair: (nat * nat) -> nat :=
  Fix wf_natpair _ encode_pair_.

(** Needed for [Fix_eq]. *)
Lemma encode_pair_ext: forall x
    (f g: forall y, lt_natpair y x -> word)
    (E: forall y (LT: lt_natpair y x), f y LT = g y LT),
  encode_pair_ f = encode_pair_ g.
Proof.
  i; f_equal; extensionalities; auto.
Qed.

Ltac unfold_encode_pair :=
  unfold encode_pair in *;
  rewrite ->Init.Wf.Fix_eq in *; [|by auto using encode_pair_ext];
  fold encode_pair in *; unfold encode_pair_ in *.

Lemma concat_zero: forall len1 n1 n2
    (E: concat len1 (n1, n2) = 0),
  n1 = 0 /\ n2 = 0.
Proof.
  unfold concat; i.
  apply Nat.lor_eq_0_iff in E; des; subst.
  apply Nat.shiftl_eq_0_iff in E0; auto.
Qed.                            

Lemma encode_pair_zero: forall p
    (E: encode_pair p = 0),
  fst p = 0 /\ snd p = 0.
Proof.
  induction p using (well_founded_ind wf_natpair); i; unfold_encode_pair.
  destruct p; ss; clarify; des_ifs.
  - do 2 (apply concat_zero in E; des).
    apply H in E0.
    + des; ss.
      rewrite (split_last_bits 1 n0) E1 E2 Nat.shiftl_0_l Nat.lor_diag; auto.
    + rewrite Nat.shiftr_0_l; econs 1; econs 2; auto using shiftr_lt.
  - do 2 (apply concat_zero in E; des).
    destruct (LEM (n0 = 0)).
    + subst; apply H in E0.
      * des; ss.
        rewrite (split_last_bits 1 n) E0 E Nat.shiftl_0_l Nat.lor_diag; auto.
      * rewrite Nat.shiftr_0_l; econs 1; econs 1; auto using shiftr_lt.
    + apply H in E0.
      * des; ss.
        rewrite (split_last_bits 1 n0) E1 E2 Nat.shiftl_0_l Nat.lor_diag; auto.
        rewrite (split_last_bits 1 n) E E0 Nat.shiftl_0_l Nat.lor_diag; auto.
      * econs 2; econs 1; [econs 1|econs 2]; auto using shiftr_lt.
Qed.


Program Definition decode_pair_ n
    (decode_pair: forall n', n' < n -> nat * nat) :=
  match eqdec n 0 with
    | left _ => (0, 0)
    | right _ =>
      let p := uncat 2 n in
      let (n1, n2) := decode_pair (snd p) _ in
      let (b1, b2) := uncat 1 (fst p) in
      (lor b1 (shiftl n1 1), lor b2 (shiftl n2 1))
  end.
Next Obligation.
  rewrite Nat.shiftr_div_pow2; apply Nat.div_lt; s; omega.
Qed.

Definition decode_pair: nat -> (nat * nat) :=
  Fix lt_wf _ decode_pair_.

(** Needed for [Fix_eq]. *)
Lemma decode_pair_ext: forall x
    (f g: forall y, y < x -> word * word)
    (E: forall y (LT: y < x), f y LT = g y LT),
  decode_pair_ f = decode_pair_ g.
Proof.
  i; f_equal; extensionalities; auto.
Qed.

Ltac unfold_decode_pair :=
  unfold decode_pair in *;
  rewrite ->Init.Wf.Fix_eq in *; [|by auto using decode_pair_ext];
  fold decode_pair in *; unfold decode_pair_ in *.


Lemma decode_pair_zero: forall n
    (E: decode_pair n = (0, 0)),
  n = 0.
Proof.
  induction n using (well_founded_ind lt_wf); i; unfold_decode_pair.
  des_ifs; apply Nat.lor_eq_0_iff in H1; apply Nat.lor_eq_0_iff in H2; des.
  apply Nat.shiftl_eq_0_iff in H0; apply Nat.shiftl_eq_0_iff in H3; subst.
  apply H in Heq; [|by apply shiftr_lt].
  apply uncat_zero in Heq0.
  hexploit (@uncat_zero 2 n); auto; destruct uncat; auto.
Qed.


Lemma encode_pair_decode: forall n,
  encode_pair (decode_pair n) = n.
Proof.
  induction n using (well_founded_ind lt_wf); i; unfold_decode_pair.
  des_ifs; unfold_encode_pair; s; [by auto|].
  des_ifs.
  - apply Nat.lor_eq_0_iff in e; apply Nat.lor_eq_0_iff in e0; des.
    apply Nat.shiftl_eq_0_iff in e1; apply Nat.shiftl_eq_0_iff in e2; subst.
    apply decode_pair_zero in Heq.
    apply uncat_zero in Heq0.
    hexploit (@uncat_zero 2 n); auto; destruct uncat; auto.
  - unfold uncat in Heq0; rewrite last_bits_last_bits in Heq0; [|by auto].
    clarify.
    assert (E1: last_bits 1 (lor (last_bits 1 n) (shiftl n1 1)) =
                last_bits 1 n).
    { rewrite last_bits_lor last_bits_last_bits; [|by auto].
      rewrite last_bits_shiftl; auto; rewrite Nat.lor_0_r; auto.
    }
    assert (E2: last_bits 1 (lor (shiftr (last_bits 2 n) 1) (shiftl n2 1)) =
                shiftr (last_bits 2 n) 1).
    { rewrite last_bits_lor last_bits_shiftl; auto; rewrite Nat.lor_0_r.
      rewrite {2}(split_last_bits 1 (shiftr _ _)) Nat.shiftr_shiftr; s.
      rewrite (@shiftr_last_bits 2 _ 2); auto.
      rewrite Nat.shiftl_0_l Nat.lor_0_l; auto.
    }
    rewrite E1 E2 -{1}(@last_bits_last_bits 1 2); [|by auto].
    rewrite split_concat Nat.shiftr_lor shiftr_last_bits; auto.
    rewrite Nat.shiftr_shiftl_r; s; [|by auto].
    rewrite Nat.lor_0_l Nat.shiftr_0_r Nat.shiftr_lor Nat.shiftr_shiftr; s.
    rewrite shiftr_last_bits; auto; rewrite Nat.shiftr_shiftl_r; s; [|by auto].
    rewrite Nat.lor_0_l Nat.shiftr_0_r -Heq H; [by apply split_concat|].
    unfold uncat in *; ss; auto using shiftr_lt.
  - 
    unfold uncat in Heq0; rewrite last_bits_last_bits in Heq0; [|by auto].
    clarify.
    assert (E1: last_bits 1 (lor (last_bits 1 n) (shiftl n1 1)) =
                last_bits 1 n).
    { rewrite last_bits_lor last_bits_last_bits; [|by auto].
      rewrite last_bits_shiftl; auto; rewrite Nat.lor_0_r; auto.
    }
    assert (E2: last_bits 1 (lor (shiftr (last_bits 2 n) 1) (shiftl n2 1)) =
                shiftr (last_bits 2 n) 1).
    { rewrite last_bits_lor last_bits_shiftl; auto; rewrite Nat.lor_0_r.
      rewrite {2}(split_last_bits 1 (shiftr _ _)) Nat.shiftr_shiftr; s.
      rewrite (@shiftr_last_bits 2 _ 2); auto.
      rewrite Nat.shiftl_0_l Nat.lor_0_l; auto.
    }
    rewrite E1 E2 -{1}(@last_bits_last_bits 1 2); [|by auto].
    rewrite split_concat Nat.shiftr_lor shiftr_last_bits; auto.
    rewrite Nat.shiftr_shiftl_r; s; [|by auto].
    rewrite Nat.lor_0_l Nat.shiftr_0_r Nat.shiftr_lor Nat.shiftr_shiftr; s.
    rewrite shiftr_last_bits; auto; rewrite Nat.shiftr_shiftl_r; s; [|by auto].
    rewrite Nat.lor_0_l Nat.shiftr_0_r -Heq H; [by apply split_concat|].
    unfold uncat in *; ss; auto using shiftr_lt.
Qed.

Lemma junk: forall n n0,
  concat 1 (last_bits 1 n, last_bits 1 n0) =
  last_bits 2 (concat 1 (last_bits 1 n, last_bits 1 n0)).
Proof.
 i; Nat.bitwise; unfold concat.
 rewrite last_bits_lor !lor_spec.
 destruct (LEM (m < 2)).
 - rewrite !(@testbit_last_bits_low 2); auto.
 - rewrite shiftl_spec_high; auto; [|by apply not_le; ii; apply f; omega].
   rewrite !testbit_last_bits_high; auto.
   i; apply f; omega.
Qed.

Lemma decode_pair_encode: forall p,
  decode_pair (encode_pair p) = p.
Proof.
  induction p using (well_founded_ind wf_natpair); unfold_encode_pair.
  destruct p; des_ifs; unfold_decode_pair; subst; first by ss; auto.
  - des_ifs.
    + ss; clarify; do 2 (apply concat_zero in e0; des).
      apply encode_pair_zero in e1; des; ss.
      rewrite (split_last_bits 1 n0) e2 e3 Nat.shiftl_0_l Nat.lor_diag; auto.
    + rewrite uncat_concat in Heq0; last by apply junk.
      rewrite uncat_concat in Heq; last by apply junk.
      rewrite uncat_concat in Heq0; [|by rewrite last_bits_last_bits].
      clarify; ss; rewrite H in Heq; last first.
      { subst; rewrite Nat.shiftr_0_l; econs 1; econs 2; auto using shiftr_lt. }
      clarify; f_equal; rewrite Nat.lor_comm; rewrite -split_last_bits; auto.
  - des_ifs.
    + ss; clarify; do 2 (apply concat_zero in e; des).
      apply encode_pair_zero in e0; des; ss.
      rewrite (split_last_bits 1 n) e e0 Nat.shiftl_0_l Nat.lor_diag; auto.
      rewrite (split_last_bits 1 n0) e1 e2 Nat.shiftl_0_l Nat.lor_diag; auto.
    + rewrite uncat_concat in Heq0; last by apply junk.
      rewrite uncat_concat in Heq; last by apply junk.
      rewrite uncat_concat in Heq0; [|by rewrite last_bits_last_bits].
      rewrite H in Heq; last first.
      { s; destruct (LEM (n0 = 0)).
        - subst; rewrite Nat.shiftr_0_l; econs 1; econs 1; auto using shiftr_lt.
        - econs 2; econs 1; [econs 1|econs 2]; auto using shiftr_lt.
      }
      clarify; f_equal; rewrite Nat.lor_comm; rewrite -split_last_bits; auto.
Qed.

Lemma encode_pair_inj:
  Image.injective _ _ encode_pair.
Proof.
  ii; rewrite -(decode_pair_encode x) H; auto using decode_pair_encode.
Qed.



(** * Machine language syntax. *)


(** ** Registers. *)

Inductive reg := r_sp | r_env | r_arg | r_clo | r_ret | r_aux | r_i.

Instance EqDec_reg: EqDec reg := mkEqDec _.
Proof.
  destruct a, b; first [by left; eauto|by right; eauto].
Qed.

Definition enum_reg (r: reg) : nat :=
  match r with
    | r_sp => 0
    | r_env => 1
    | r_arg => 2
    | r_clo => 3
    | r_ret => 4
    | r_aux => 5
    | r_i => 6
  end.

Definition decode_reg n : option reg :=
  match n with
    | 0 => Some r_sp
    | 1 => Some r_env
    | 2 => Some r_arg
    | 3 => Some r_clo
    | 4 => Some r_ret
    | 5 => Some r_aux
    | 6 => Some r_i
    | _ => None
  end.

Instance Enumerable_reg: Enumerable reg :=
  {| enum := enum_reg |}.
  ii; destruct x, y; vauto.
Defined.

Lemma enum_reg_last_bits: forall (r: reg),
  enum r = last_bits 3 (enum r).
Proof.
  i; destruct r; auto.
Qed.

Lemma decode_reg_enum: forall r,
  decode_reg (enum r) = Some r.
Proof.
  destruct r; auto.
Qed.

Lemma decode_reg_some: forall n r
    (E: decode_reg n = Some r),
  n = enum_reg r.
Proof.
  i; repeat (destruct n; ss; vauto).
Qed.


(** ** Signed offsets. *)

Inductive signed_offset :=
| o_pos (off : nat)
| o_neg (off : nat).

Fixpoint enum_signed_offset (s: signed_offset) : nat :=
  match s with
    | o_pos n => concat 1 (0, n)
    | o_neg n => concat 1 (1, n)
  end.

Instance Enumerable_signed_offset: Enumerable signed_offset :=
  {| enum := enum_signed_offset |}.
  ii; destruct x, y; unfold enum_signed_offset in *; apply concat_inj in H
  ; vauto; apply enum_inj in H; vauto.
Defined.


Definition decode_signed_offset n : option signed_offset :=
  let (n1, n2) := uncat 1 n in
  match n1 with
    | 0 => Some (o_pos n2)
    | 1 => Some (o_neg n2)
    | _ => impossible
  end.

Lemma decode_signed_offset_enum: forall so,
  decode_signed_offset (enum so) = Some so.
Proof.
  unfold decode_signed_offset; unfold enum, Enumerable_signed_offset; i.
  destruct (uncat _ _) eqn:E, so; rewrite uncat_concat in E; auto; clarify.
Qed.

Lemma decode_signed_offset_some: forall n so
    (E: decode_signed_offset n = Some so),
  n = enum_signed_offset so.
Proof.
  i; unfold decode_signed_offset in E; destruct (uncat _ _) eqn:U.
  invc U; des_ifs; s; unfold concat
  ; rewrite -{1}Heq Nat.lor_comm; auto using split_last_bits.
Qed.


(** ** Operands. *)

Inductive operand :=
| o_imm (w: word)
| o_reg (r: reg)
| o_sind (r: reg) (soff: signed_offset)
| o_hind (r: reg) (soff: signed_offset)
.

Fixpoint enum_operand (o: operand) : nat :=
  match o with
    | o_imm w => concat 2 (0, w)
    | o_reg r => concat 2 (1, enum r)
    | o_sind r s => concat 2 (2, concat 3 (enum r, enum s))
    | o_hind r s => concat 2 (3, concat 3 (enum r, enum s))
  end.

Instance Enumerable_operand: Enumerable operand :=
  {| enum := enum_operand |}.
  ii; destruct x, y; unfold enum_operand in *; apply concat_inj in H; clarify.
  - apply enum_inj in H; vauto.
  - apply concat_inj in H; clarify; auto using enum_reg_last_bits.
    apply enum_inj in H; apply enum_inj in H1; vauto.
  - apply concat_inj in H; clarify; auto using enum_reg_last_bits.
    apply enum_inj in H; apply enum_inj in H1; vauto.
Defined.


Definition decode_operand n : option operand :=
  let (n1, n2) := uncat 2 n in
  match n1 with
    | 0 => Some (o_imm n2)
    | 1 => mdo r <- decode_reg n2;
           Some (o_reg r)
    | 2 => let (xr, xso) := uncat 3 n2 in
           mdo r <- decode_reg xr;
           mdo so <- decode_signed_offset xso;
           Some (o_sind r so)
    | 3 => let (xr, xso) := uncat 3 n2 in
           mdo r <- decode_reg xr;
           mdo so <- decode_signed_offset xso;
           Some (o_hind r so)
    | _ => impossible
  end.

Lemma decode_operand_enum: forall o,
  decode_operand (enum o) = Some o.
Proof.
  unfold decode_operand; unfold enum, Enumerable_operand; i.
  destruct (uncat _ _) eqn:E, o; rewrite uncat_concat in E; auto; clarify.
  - rewrite decode_reg_enum; auto.
  - rewrite uncat_concat; auto using enum_reg_last_bits.
    rewrite decode_reg_enum decode_signed_offset_enum; auto.
  - rewrite uncat_concat; auto using enum_reg_last_bits.
    rewrite decode_reg_enum decode_signed_offset_enum; auto.
Qed.

Lemma decode_operand_some: forall n o
    (E: decode_operand n = Some o),
  n = enum_operand o.
Proof.
  i; unfold decode_operand in E; destruct (uncat _ _) eqn:U.
  invc U; des_ifs; s; unfold concat
  ; rewrite -{1}Heq Nat.lor_comm; auto using split_last_bits.
  - apply decode_reg_some in Heq0.
    simpl; unfold enum, Enumerable_reg; rewrite -Heq0.
    apply split_last_bits.
  - apply decode_reg_some in Heq1.
    apply decode_signed_offset_some in Heq2.
    simpl; unfold enum, Enumerable_reg, Enumerable_signed_offset.
    invc Heq0; rewrite -H0 -H1.
    rewrite (Nat.lor_comm (last_bits 3 _)).
    rewrite -split_last_bits -split_last_bits; auto.
  - apply decode_reg_some in Heq1.
    apply decode_signed_offset_some in Heq2.
    simpl; unfold enum, Enumerable_reg, Enumerable_signed_offset.
    invc Heq0; rewrite -H0 -H1.
    rewrite (Nat.lor_comm (last_bits 3 _)).
    rewrite -split_last_bits -split_last_bits; auto.
Qed.


(** ** Instructions. *)

Inductive instr :=
| i_jmp (addr: operand)
| i_jnz (r: reg) (addr: operand)
| i_ld  (r: reg) (addr: operand)
| i_sto (addr: operand) (src: reg)
| i_lpc (r: reg)
| i_opr (op: binop) (tgt: reg) (arga argb: operand)
| i_inp (r: reg)
| i_out (r: reg)
| i_alc (r: reg) (rsize: reg)
.

Fixpoint enum_instr (i: instr) : nat :=
  match i with
    | i_jmp o => concat 4 (0, enum o)
    | i_jnz r o => concat 4 (1, concat 3 (enum r, enum o))
    | i_ld r o => concat 4 (2, concat 3 (enum r, enum o))
    | i_sto o r =>  concat 4 (3, concat 3 (enum r, enum o))
    | i_lpc r => concat 4 (4, enum r)
    | i_opr b r o o' =>
      let operands := encode_pair (enum o, enum o') in
      concat 4 (5, concat 2 (enum b, concat 3 (enum r, operands)))
    | i_inp r => concat 4 (6, enum r)
    | i_out r => concat 4 (7, enum r)
    | i_alc r r' => concat 4 (8, concat 3 (enum r, enum r'))
  end.

Lemma enum_binop_last_bits: forall (b: binop),
  enum b = last_bits 2 (enum b).
Proof.
  i; destruct b; auto.
Qed.

Instance Enumerable_instr: Enumerable instr :=
  {| enum := enum_instr |}.
  ii; destruct x, y; unfold enum_instr in *; apply concat_inj in H
  ; clarify; try apply enum_inj in H; vauto.
  - apply concat_inj in H; clarify; auto using enum_reg_last_bits.
    apply enum_inj in H; apply enum_inj in H1; vauto.
  - apply concat_inj in H; clarify; auto using enum_reg_last_bits.
    apply enum_inj in H; apply enum_inj in H1; vauto.
  - apply concat_inj in H; clarify; auto using enum_reg_last_bits.
    apply enum_inj in H; apply enum_inj in H1; vauto.
  - apply concat_inj in H; clarify; auto using enum_binop_last_bits.
    apply concat_inj in H; clarify; auto using enum_reg_last_bits.
    apply encode_pair_inj in H; vauto.
    apply enum_inj in H; apply enum_inj in H1; vauto.
    apply enum_inj in H2; apply enum_inj in H3; vauto. 
  - apply concat_inj in H; clarify; auto using enum_reg_last_bits.
    apply enum_inj in H; apply enum_inj in H1; vauto.
Defined.

Definition decode_instr n : option instr :=
  let (n1, n2) := uncat 4 n in
  match n1 with
    | 0 => mdo o <- decode_operand n2;
           Some (i_jmp o)
    | 1 => let (xr, xo) := uncat 3 n2 in
           mdo r <- decode_reg xr;
           mdo o <- decode_operand xo;
           Some (i_jnz r o)
    | 2 => let (xr, xo) := uncat 3 n2 in
           mdo r <- decode_reg xr;
           mdo o <- decode_operand xo;
           Some (i_ld r o)
    | 3 => let (xr, xo) := uncat 3 n2 in
           mdo r <- decode_reg xr;
           mdo o <- decode_operand xo;
           Some (i_sto o r)
    | 4 => mdo r <- decode_reg n2;
           Some (i_lpc r)
    | 5 => let (xb, n3) := uncat 2 n2 in
           mdo b <- decode_binop xb;
           let (xr, xos) := uncat 3 n3 in
           mdo r <- decode_reg xr;
           let (xo, xo') := decode_pair xos in
           mdo o <- decode_operand xo;
           mdo o' <- decode_operand xo';
           Some (i_opr b r o o')
    | 6 => mdo r <- decode_reg n2;
           Some (i_inp r)
    | 7 => mdo r <- decode_reg n2;
           Some (i_out r)
    | 8 => let (xr, xr') := uncat 3 n2 in
           mdo r <- decode_reg xr;
           mdo r' <- decode_reg xr';
           Some (i_alc r r')
    | _ => impossible
  end.

Lemma decode_instr_enum: forall o,
  decode_instr (enum o) = Some o.
Proof.
  unfold decode_instr; unfold enum, Enumerable_instr; i.
  destruct (uncat _ _) eqn:E, o; rewrite uncat_concat in E; auto; clarify.
  - rewrite decode_operand_enum; auto.
  - rewrite uncat_concat; auto using enum_reg_last_bits.
    rewrite decode_reg_enum decode_operand_enum; auto.
  - rewrite uncat_concat; auto using enum_reg_last_bits.
    rewrite decode_reg_enum decode_operand_enum; auto.
  - rewrite uncat_concat; auto using enum_reg_last_bits.
    rewrite decode_reg_enum decode_operand_enum; auto.
  - rewrite decode_reg_enum; auto.
  - rewrite uncat_concat; auto using enum_binop_last_bits.
    rewrite decode_binop_enum uncat_concat; auto using enum_reg_last_bits.
    rewrite decode_reg_enum decode_pair_encode.
    rewrite 2!decode_operand_enum; auto.
  - rewrite decode_reg_enum; auto.
  - rewrite decode_reg_enum; auto.
  - rewrite uncat_concat; auto using enum_reg_last_bits.
    rewrite 2!decode_reg_enum; auto.
Qed.

Lemma decode_instr_some: forall n i
    (E: decode_instr n = Some i),
  n = enum_instr i.
Proof.
  i; unfold decode_instr in E; destruct (uncat _ _) eqn:U.
  invc U; des_ifs; s; unfold concat
  ; rewrite -{1}Heq Nat.lor_comm; auto using split_last_bits.
  - apply decode_operand_some in Heq0.
    simpl; unfold enum, Enumerable_operand; rewrite -Heq0.
    apply split_last_bits.
  - apply decode_reg_some in Heq1.
    apply decode_operand_some in Heq2.
    simpl; unfold enum, Enumerable_reg, Enumerable_operand.
    invc Heq0; rewrite -H0 -H1.
    rewrite (Nat.lor_comm (last_bits 3 _)).
    rewrite -split_last_bits -split_last_bits; auto.
  - apply decode_reg_some in Heq1.
    apply decode_operand_some in Heq2.
    simpl; unfold enum, Enumerable_reg, Enumerable_operand.
    invc Heq0; rewrite -H0 -H1.
    rewrite (Nat.lor_comm (last_bits 3 _)).
    rewrite -split_last_bits -split_last_bits; auto.
  - apply decode_reg_some in Heq1.
    apply decode_operand_some in Heq2.
    simpl; unfold enum, Enumerable_reg, Enumerable_operand.
    invc Heq0; rewrite -H0 -H1.
    rewrite (Nat.lor_comm (last_bits 3 _)).
    rewrite -split_last_bits -split_last_bits; auto.
  - apply decode_reg_some in Heq0.
    simpl; unfold enum, Enumerable_reg.
    invc Heq0; rewrite -split_last_bits; auto.
  - apply decode_binop_some in Heq1.
    apply decode_reg_some in Heq3.
    apply decode_operand_some in Heq5.
    apply decode_operand_some in Heq6.
    simpl; unfold enum, Enumerable_binop, Enumerable_reg, Enumerable_operand.
    invc Heq0 Heq2 Heq4; rewrite encode_pair_decode.
    rewrite -H0 (Nat.lor_comm (last_bits 3 _)) (Nat.lor_comm (last_bits 2 _)).
    rewrite -3!split_last_bits; auto.
  - apply decode_reg_some in Heq0.
    simpl; unfold enum, Enumerable_reg.
    invc Heq0; rewrite -split_last_bits; auto.
  - apply decode_reg_some in Heq0.
    simpl; unfold enum, Enumerable_reg.
    invc Heq0; rewrite -split_last_bits; auto.
  - apply decode_reg_some in Heq1; apply decode_reg_some in Heq2.
    simpl; unfold enum, Enumerable_reg.
    invc Heq0; rewrite -H0 -H1.
    rewrite (Nat.lor_comm (last_bits 3 _)).
    rewrite -split_last_bits -split_last_bits; auto.
Qed.
