(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import countable vgtac basis lang_common lang_spec.
Require Import model_common model_untyped model_properties_untyped gw_untyped mm mm_test mm.
Require Import ssr util opt_transform lang_mid lang_mid_renaming.
Require Import compiler_common opt_freshen.


Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


(** Inlining of top-level functions. *)


Fixpoint inline (phi: clomem) (e: exp) : exp :=
  match e with
    | IEapp xf xv xk =>
        match lookup (inl xf) phi with
          | Some (IEfix f x k e') => rename ren_id[xf/f][xv/x][c xk/k] e'
          | _ => e
        end
    | IEinst xg xk =>
        match lookup (inl xg) phi with
          | Some (IEgen k e') => rename ren_id[c xk/k] e'
          | _ => e
        end
    | IEfix bf bx bk e => IEfix bf bx bk (inline phi e)
    | IEgen bk e => IEgen bk (inline phi e) 
    | IEcont bx e => IEcont bx (inline phi e) 
    | IEnew y bx e => IEnew y bx (inline phi e)
    | IEread y bx e => IEread y bx (inline phi e)
    | IEwrite y x e => IEwrite y x (inline phi e)
    | IEinput bx e => IEinput bx (inline phi e)
    | IEoutput x e => IEoutput x (inline phi e)
    | IEifnz x ea eb => IEifnz x (inline phi ea) (inline phi eb)
    | IEcase y bx ea eb => IEcase y bx (inline phi ea) (inline phi eb)
    | IElet ea bx eb => IElet (inline phi ea) bx (inline phi eb) 
    | _ => e
  end.


Definition analyze: module -> list label := map fst.
Lemma analyze_sane: forall m l
    (IN: In l (analyze m)),
  In l (map fst m).
Proof. auto. Qed.

Fixpoint inline_mod_ lbls phi (m: module) : module :=
  match m with
    | nil => nil
    | (l, e) :: m' =>
      let e' := inline phi e in
      let phi' := if list_mem eqdec l lbls then (inl (inl l), e) :: phi else phi in
      (l, e') :: inline_mod_ lbls phi' m'
  end.

        








Lemma respects_nil: forall G env,
  respects G env nil.
Proof.
  ii; vauto.
Qed.
Hint Immediate respects_nil.


Lemma sublist_rev {X} `{M: EqDec X} (xs ys : list X):
  sublist xs ys -> sublist (rev xs) (rev ys).
Proof.
  intros H. induction H; eauto using sublist_nil'.
  - s; by eapply sublist_r_app_l.
  - s; by eapply sublist_app.
Qed.

Lemma sublist_rev_sym {X} `{M: EqDec X} (xs ys : list X):
  sublist xs (rev ys) -> sublist (rev xs) ys.
Proof.
  revert ys. induction xs; intros.
  - eapply sublist_nil'.
  - invc H; rewrite <- (rev_involutive ys); [rewrite <- H2|rewrite <- H3];
    eapply sublist_rev.
    + rewrite (cons_app x). by eapply sublist_r_app_r.
    + eauto.
Qed.

Lemma sublist_rev_trans {X} `{M: EqDec X} (xs ys zs: list X):
  sublist xs (rev ys) -> sublist ys (rev zs) -> sublist xs zs.
Proof. intros H1 H2. eapply sublist_trans. exact H1. by eapply sublist_rev_sym. Qed.


Lemma sublist_weaken {X} `{M : EqDec X} (xs ys zs : list X) : 
  sublist xs zs -> sublist ys xs -> sublist ys zs.
Proof.
  intros H1 H2; revert ys H2; induction H1; eauto; ii. invc H2. 
  - eapply sublist_trans. exact SB. eauto.
  - eapply sublist_cons_keep. eauto.
Qed.


Lemma sublist_purge_lbl (xs : list var):
  sublist (purge_lbl xs) xs.
Proof. induction xs; eauto; destruct a as [[l|t]|k]; s; eauto using sublist_nil'. Qed.


Lemma uniq_quick (bs bsr xs ys : list var) : 
  uniq (xs ++ (rev bs) ++ bsr ++ ys) ->
  uniq (xs ++ bsr ++ (bs ++ ys)).
Proof.
  intros. eapply uniq_perm; eauto.
  eapply Permutation_app; auto.
  eapply Permutation_trans.
  eapply Permutation_app; auto.
  rewrite !app_assoc.
  eapply Permutation_app; auto.
  eapply Permutation_trans.
  eapply Permutation_app_comm.
  eapply Permutation_app; auto.
  symmetry. eapply Permutation_rev.
Qed.

Lemma uniq_quickl (bs bsr xs ys zs : list var) : 
  uniq (xs ++ (rev bs) ++ (bsr ++ zs) ++ ys) ->
  uniq (xs ++ bsr ++ (bs ++ ys)).
Proof.
  intros. eapply uniq_quick.
  eapply uniq_sublist; eauto.
  eapply sublist_app; auto.
  eapply sublist_app; auto.
  eapply sublist_app; auto.
Qed.

Lemma uniq_quickr (bs bsr xs ys zs : list var) : 
  uniq (xs ++ (rev bs) ++ (zs ++ bsr) ++ ys) ->
  uniq (xs ++ bsr ++ (bs ++ ys)).
Proof.
  intros. eapply uniq_quick.
  eapply uniq_sublist; eauto.
  eapply sublist_app; auto.
  eapply sublist_app; auto.
  eapply sublist_app; auto.
Qed.

Lemma uniq_quickm (bs bsr xs ys zs zs' : list var) : 
  uniq (xs ++ (rev bs) ++ (zs ++ bsr ++ zs') ++ ys) ->
  uniq (xs ++ bsr ++ (bs ++ ys)).
Proof.
  intros. eapply uniq_quick.
  eapply uniq_sublist; eauto.
  eapply sublist_app; auto.
  eapply sublist_app; auto.
  eapply sublist_app; auto.
Qed.

    
Lemma rename_good: forall e ren
    (GT: good_exp e),
  good_exp (rename ren e).
Proof.
  induction e; vauto.
  destruct bx; vauto.
Qed.

Lemma blaaaa: forall lw G (phia phib: clomem) bf bx bk e f x k
    (INf: In (inl f) G)
    (INx: In (inl x) G)
    (INk: In (inr k) G)
    (INL: lookup (inl f) phib = Some (IEfix bf bx bk e))
    
    (BVV: forall y (E: y = inl f \/ y = inl x \/ y = inr k), ~ In y (bvars e))
    (WF: wfatom (defbefore (inl f) G) (IEfix bf bx bk e))
    (STA: stable lw),
  mm_test.mm_psim lw G ren_id phia phib (inline phib (IEapp f x k)) (IEapp f x k).
Proof.
  ii; unfold inline; rewrite INL.
  rr; splits.
  { assert (good_exp e) by (ii; subst; vauto).
    eapply rename_good; eauto.
  }
  { rr; vauto. }
  set (ren := (((ren_id)[f/bf])[x/bx])[c k/bk]).

  assert (SIM: mm.mm_psim lw (inr bk :: inl (inr bx) ::
                                  inl (inr bf) :: purge (defbefore (inl f) G))
                              ren (rename ren e) e).
  { apply mm.mm_psim_rename; eauto.
    i;     revert BV. 
    apply in_consD in IN; des; subst.
    { s; tmext_simpl. i.
      exfalso; eauto.
    }
    rewrite ren_c_ext_neq; vauto.
    apply in_consD in IN; des; subst.
    { s; tmext_simpl. i.
      exfalso; eauto.
    }
    rewrite ren_ext_neq; vauto.
    apply in_consD in IN; des; subst.
    { s; tmext_simpl. i.
      exfalso; eauto.
    }
    rewrite ren_ext_neq; vauto.
    rewrite ren_var_id; auto.
  }
  rrdes SIM.

  econs; i.
  eapply pesim_acquire; i.
  eapply pesim_stepb; i.
  dup HR; destruct HR as (? & ? & ? & ? & WHX & WHY & ? & ?).
  apply mm_confrel_shape in WHX; des.
  subst.
  destruct x1, x3.
  dup WHY; eapply lwconf_extra in WHY; des.
  rrdes WHY WHY1 XTR. rrdes RL. s in Ec. clarify. noemp.

  dup INL; apply RPb in INL0; des; subst.
  eapply RV in INf; des.
  rewrite IN in INf0; clarify.
  eapply RV in INx; des.
  eapply RK in INk; des.

  esplits; [|right; split].
  - econs; inst_pairs; econs; vauto.
  - inst_pairs; s; noemp; rr; esplits; eauto.
  - s. noemp.
    rewrite 2!mid_csplit.
    eapply pesim_release; first M; Mskip by rr; eauto.
    eapply SIM1; eauto.
    + i; apply in_consD in IN0; des; vauto.
      rewrite ren_ext_it_neq_c list_assoc_cons_neq; vauto.
      apply in_consD in IN0; des; vauto.
      { simpl rename_itvar; rewrite list_assoc_cons_eq; tmext_simpl.
        rewrite ->ren_contvar_id, ren_itvar_id in *.
        esplits; eauto.
      }
      rewrite ren_ext_it_neq; [|by ii; vauto].
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      apply in_consD in IN0; des; clarify.
      { simpl rename_itvar; rewrite list_assoc_cons_eq; tmext_simpl.
        rewrite ->ren_contvar_id, ren_itvar_id in *.
        esplits; eauto.
      }
      rewrite ren_ext_it_neq; [|by ii; vauto].
      rewrite list_assoc_cons_neq; [|by ii; vauto].

      dup IN0.
      apply purge_inD, EN in IN0.
      rewrite IN0.
      eapply RV.
      eauto.
      eapply sublist_in; last first; eauto.
      eapply sublist_trans; eauto using purge_sublist.
      by eapply defbefore_sublist.
    + i; apply in_consD in IN0; des; clarify.
      { simpl rename_contvar; rewrite list_assoc_cons_eq; tmext_simpl.
        rewrite ->ren_contvar_id, ren_itvar_id in *.
        esplits; eauto.
      }
      rewrite ren_ext_c_neq'; [|by ii; vauto].
      unfold ren in *; rewrite list_assoc_cons_neq; [|by ii; vauto].
      repeat (apply in_consD in IN0; des; vauto).
      exfalso; eauto using purge_inDr.
Grab Existential Variables.
exact 0.
Qed.


Lemma blaaaa_inst: forall lw G (phia phib: clomem) bf e f k
    (INf: In (inl f) G)
    (INk: In (inr k) G)
    (INL: lookup (inl f) phib = Some (IEgen bf e))
    
    (BVV: forall y (E: y = inl f \/ y = inr k), ~ In y (bvars e))
    (WF: wfatom (defbefore (inl f) G) (IEgen bf e))
    (STA: stable lw),
  mm_test.mm_psim lw G ren_id phia phib (inline phib (IEinst f k)) (IEinst f k).
Proof.
  ii; unfold inline; rewrite INL.
  rr; splits.
  { assert (good_exp e) by (ii; subst; vauto).
    eapply rename_good; eauto.
  }
  { rr; vauto. }
  set (ren := (((ren_id)[c k/bf]))).

  assert (SIM: mm.mm_psim lw (
                                  inr bf :: purge (defbefore (inl f) G))
                              ren (rename ren e) e).
  { apply mm.mm_psim_rename; eauto. 
    i.     revert BV. 
    apply in_consD in IN; des; subst.
    { s; tmext_simpl. i.
      exfalso; eauto.
    }
    rewrite ren_c_ext_neq; vauto.
    rewrite ren_var_id; auto.
  }
  rrdes SIM.

  econs; i.
  eapply pesim_acquire; i.
  eapply pesim_stepb; i.
  dup HR; destruct HR as (? & ? & ? & ? & WHX & WHY & ? & ?).
  apply mm_confrel_shape in WHX; des.
  subst.
  destruct x0, x2.
  dup WHY; eapply lwconf_extra in WHY; des.
  rrdes WHY WHY1 XTR. rrdes RL. s in Ec. clarify. noemp.

  dup INL; apply RPb in INL0; des; subst.
  eapply RV in INf; des.
  rewrite IN in INf0; clarify.
  eapply RK in INk; des.

  esplits; [|right; split].
  - econs; inst_pairs; econs; vauto.
  - inst_pairs; s; noemp; rr; esplits; eauto.
  - s. noemp.
    rewrite 2!mid_csplit.
    eapply pesim_release; first M; Mskip by rr; eauto.
    eapply SIM1; eauto.
    + i; apply in_consD in IN0; des; vauto.
      rewrite ren_ext_it_neq_c list_assoc_cons_neq; vauto.
      dup IN0.
      apply purge_inD, EN in IN0.
      rewrite IN0.
      eapply RV.
      eauto.
      eapply sublist_in; last first; eauto.
      eapply sublist_trans; eauto using purge_sublist.
      by eapply defbefore_sublist.
    + i; apply in_consD in IN0; des; clarify.
      { simpl rename_contvar; rewrite list_assoc_cons_eq; tmext_simpl.
        rewrite ->ren_contvar_id, ren_itvar_id in *.
        esplits; eauto.
      }
      rewrite ren_ext_c_neq'; [|by ii; vauto].
      unfold ren in *; rewrite list_assoc_cons_neq; [|by ii; vauto].
      repeat (apply in_consD in IN0; des; vauto).
      exfalso; eauto using purge_inDr.
Grab Existential Variables.
exact 0.
Qed.






























Lemma extendable_t_id: forall G x,
  mm_test.extendable_t G ren_id x x.
Proof. 
  i; apply extendable_t_same; i; rewrite ren_var_id; auto.
Qed.
Hint Resolve extendable_t_id.

Definition purged' G phi :=
  purged phi /\ forall x e (IN: lookup x phi = Some e), defbefore x G = defbefore x (purge G).

Lemma inline_ksim: forall lw e G (phib: clomem)
    (WF: wfatom G e)
    (BV: forall x e' (IN: lookup x phib = Some e'), uniq (bvars e' ++ bvars e ++ purge_lbl G))
    (BV': forall x e' (IN: lookup x phib = Some e') z (IN': In z (bvars e' ++ bvars e)), ~ In z G)
    (WFA: forall x e' (IN: lookup x phib = Some e'), wfatom (defbefore x G) e')
    (DOM: forall x e' (IN: lookup x phib = Some e'), In x G)
    (FR: forall x (IN: In x (bvars e)), fresh G x phib)
    (PU: purged' G phib)
    (ST: stable lw)
    (IH: forall y : exp,
       size y < size e ->
       forall (G : list (label + termvar + contvar)) (phib : clomem),
       wfexp G y ->
       (forall (x : var) (e' : exp),
        lookup x phib = Some e' -> uniq (bvars e' ++ bvars y ++ purge_lbl G)) ->
       (forall x e' (IN: lookup x phib = Some e') z (IN': In z (bvars e' ++ bvars y)), ~ In z G) ->
       (forall (x : var) (e' : exp),
        lookup x phib = Some e' -> wfatom (defbefore x G) e') ->
       (forall (x : var) (e' : exp), lookup x phib = Some e' -> In x G) ->
       (forall x : var, In x (bvars y) -> fresh G x phib) -> purged' G phib ->
       stable lw -> mm_test.mm_psim lw G ren_id nil phib (inline phib y) y),
  mm_test.mm_ksim lw G ren_id nil phib (inline phib e) e.
Proof.
  ii; destruct e; vauto.
  s.
  hexploit IH; eauto.
  - s; eauto.
  - instantiate (1:=phib). ii; s. 
    hexploit BV; eauto; ii.
    rewrite cons_app.
    by eapply uniq_quick.
    
  - ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
    apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
    eapply uniq_appD in H3; des.
    repeat (apply in_consD in FH; des; clarify;
            first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
    eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
    eapply purge_lbl_nolbl; last eauto.
    eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
  - i; rewrite -defbefore_cons; eauto.
  - s; eauto.
  - ii. specialize (FR x (or_intror H) y e IN).
    erewrite defbefore_cons in FR; last eapply bvars_nolbl. exact FR.
    + eapply DOM; eauto. 
    + instantiate (1:=IEcont bxb eb'). s. auto.
  - split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
    erewrite <- !(defbefore_cons); try eauto. eauto using purge_subinDl.
  - i.
    esplits; [by auto|].
    rewrite -(ren_id_self_t bxb) in H.
    hexploit mm_test.mm_compat_cont; eauto.
    + ii; vauto.
    + eapply FR; s; eauto.
    + intro SIM. unfold mm_test.mm_ksim in SIM.
      hexploit SIM; last first; eauto; i; des; eauto.
Qed.

Lemma inline_asim: forall lw e G (phib: clomem)
    (WF: wfatom G e)
    (BV: forall x e' (IN: lookup x phib = Some e'), uniq (bvars e' ++ bvars e ++ purge_lbl G))
    (BV': forall x e' (IN: lookup x phib = Some e') z (IN': In z (bvars e' ++ bvars e)), ~ In z G)
    (WFA: forall x e' (IN: lookup x phib = Some e'), wfatom (defbefore x G) e')
    (DOM: forall x e' (IN: lookup x phib = Some e'), In x G)
    (FR: forall x (IN: In x (bvars e)), fresh G x phib)
    (PU: purged' G phib)
    (ST: stable lw)
    (IH: forall y : exp,
       size y < size e ->
       forall (G : list (label + termvar + contvar)) (phib : clomem),
       wfexp G y ->
       (forall (x : var) (e' : exp),
        lookup x phib = Some e' -> uniq (bvars e' ++ bvars y ++ purge_lbl G)) ->
       (forall x e' (IN: lookup x phib = Some e') z (IN': In z (bvars e' ++ bvars y)), ~ In z G) ->
       (forall (x : var) (e' : exp),
        lookup x phib = Some e' -> wfatom (defbefore x G) e') ->
       (forall (x : var) (e' : exp), lookup x phib = Some e' -> In x G) ->
       (forall x : var, In x (bvars y) -> fresh G x phib) -> purged' G phib ->
       stable lw -> mm_test.mm_psim lw G ren_id nil phib (inline phib y) y),
  mm_test.mm_asim lw G ren_id nil phib (inline phib e) e.
Proof.
  i; destruct e; vauto.
  - s; eapply mm_test.mm_compat_unit.
  - s; eapply mm_test.mm_compat_nat.
  - s in WF; des.
    eapply mm_test.mm_compat_pair; rewrite ?ren_itvar_id; auto.
  - eapply mm_test.mm_compat_inl; rewrite ?ren_itvar_id; auto.
  - eapply mm_test.mm_compat_inr; rewrite ?ren_itvar_id; auto.
  - s in WF; des.
    destruct x, y; eapply mm_test.mm_compat_bop; rewrite ?ren_itvar_id; auto.
  - eapply mm_test.mm_compat_fst; rewrite ?ren_itvar_id; auto.
  - eapply mm_test.mm_compat_snd; rewrite ?ren_itvar_id; auto.
  - s.
    eapply mm_test.mm_compat_fix; auto.
    + rr; vauto.
    + eapply FR; s; eauto.
    + rr; vauto.
    + eapply FR; s; eauto.
    + rr; vauto.
    + eapply FR; s; eauto.
    + vauto.
    + rrdes PU;  eauto.
    + rewrite ?ren_id_self_t ?ren_id_self_c; eapply IH; eauto.
      * ii; specialize (BV _ _ H). s purge_lbl. rewrite cons_app3.
        eapply uniq_quick. s.
        eapply (uniq_sublist BV).
        repeat eapply sublist_app; eauto. s.
        rewrite !app_comm_cons.
        eapply sublist_app; eauto.
        eapply sublist_trans. eapply purge_lbl_sublist. eapply purge_sublist.
        eauto.
      * ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app3. ii.
        apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
        eapply uniq_appD in H3; des.
        repeat (apply in_consD in FH; des; clarify;
                first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
        eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
        eapply purge_lbl_nolbl; last by eapply purge_inD.
        eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
      * i; rewrite -?defbefore_cons; auto.
        { dup H; apply WFA in H.
          apply PU in H0; rewrite H0 in H. auto.
        }
        { dup H; apply DOM in H.
          rrdes PU. apply PU in H0. des; subst. eauto using purge_subinDl.
        }
        { specialize (DOM _ _ H). right. rrdes PU. specialize (PU _ _ H); des; subst. 
          by eapply purge_subinDl.
        }
        { specialize (DOM _ _ H). do!right. rrdes PU. specialize (PU _ _ H); des; subst. 
          by eapply purge_subinDl.
        }
      * i; dup H; apply DOM in H.
          rrdes PU. apply PU in H0. des; subst. s; eauto using purge_subinDl. 
      * ii. specialize (FR x). edestruct FR; eauto; clear FR; s; eauto. 
        split; eauto.
        rrdes PU. specialize (PU0 _ _ IN). specialize (PU _ _ IN). des; clarify; i.
        rewrite PU0 in H1.
        repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
      * rrdes PU; split; eauto; ii. s. rewrite purge_purge; rewrite <- defbefore_cons; eauto.
        specialize (PU _ _ IN); des; clarify. s; eauto using purge_subinDl.
  - s.
    eapply mm_test.mm_compat_gen; auto.
    + rr; vauto.
    + eapply FR; s; eauto.
    + rr; vauto.
    + rrdes PU;  eauto.
    + rewrite ?ren_id_self_t ?ren_id_self_c; eapply IH; eauto.
      * ii; specialize (BV _ _ H). s purge_lbl. rewrite cons_app.
        eapply uniq_quick. s.
        eapply (uniq_sublist BV).
        repeat eapply sublist_app; eauto. s.
        rewrite !app_comm_cons.
        eapply sublist_app; eauto.
        eapply sublist_trans. eapply purge_lbl_sublist. eapply purge_sublist.
        eauto.
      * ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
        apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
        eapply uniq_appD in H3; des.
        repeat (apply in_consD in FH; des; clarify;
                first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
        eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
        eapply purge_lbl_nolbl; last by eapply purge_inD.
        eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
      * i; rewrite -?defbefore_cons; auto.
        { dup H; apply WFA in H.
          apply PU in H0; rewrite H0 in H. auto.
        }
        { dup H; apply DOM in H.
          rrdes PU. apply PU in H0. des; subst. eauto using purge_subinDl.
        }
      * i; dup H; apply DOM in H.
          rrdes PU. apply PU in H0. des; subst. s; eauto using purge_subinDl. 
      * ii. specialize (FR x). edestruct FR; eauto; clear FR; s; eauto. 
        split; eauto.
        rrdes PU. specialize (PU0 _ _ IN). specialize (PU _ _ IN). des; clarify; i.
        rewrite PU0 in H1.
        repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
      * rrdes PU; split; eauto; ii. s. rewrite purge_purge; rewrite <- defbefore_cons; eauto.
        specialize (PU _ _ IN); des; clarify. s; eauto using purge_subinDl.
  - s in WF; des.
    eapply mm_test.mm_compat_refeq; rewrite ?ren_itvar_id; auto.
Qed.

Lemma inline_psim: forall lw e G (phib: clomem)
    (WF: wfexp G e)
    (BV: forall x e' (IN: lookup x phib = Some e'), uniq (bvars e' ++ bvars e ++ purge_lbl G))
    (BV': forall x e' (IN: lookup x phib = Some e') z (IN': In z (bvars e' ++ bvars e)), ~ In z G)
    (WFA: forall x e' (IN: lookup x phib = Some e'), wfatom (defbefore x G) e')
    (DOM: forall x e' (IN: lookup x phib = Some e'), In x G)
    (FR: forall x (IN: In x (bvars e)), fresh G x phib)
    (PU: purged' G phib)
    (ST: stable lw),
  mm_test.mm_psim lw G ren_id nil phib (inline phib e) e.
Proof.
  intro; induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i.
  - s in WF;  des.
    s; eapply mm_test.mm_compat_new; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto).
    + ii. specialize (BV _ _ IN). s; rewrite cons_app. by eapply uniq_quick.
    + ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
      apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
      eapply uniq_appD in H3; des.
      repeat (apply in_consD in FH; des; clarify;
          first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
      eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
      eapply purge_lbl_nolbl; last eauto.
      eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
    + i; rewrite -defbefore_cons; eauto.
    + s; eauto.
    + ii. specialize (FR x0). edestruct FR; eauto; clear FR; s; eauto. 
      split; eauto.
      rrdes PU. specialize (PU0 _ _ IN0). specialize (PU _ _ IN0). des; clarify; i.
      repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
    + split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
      erewrite <- !(defbefore_cons); try eauto. eauto using purge_subinDl.
  - s in WF;  des.
    s; eapply mm_test.mm_compat_read; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto).
    + ii; s; rewrite cons_app; eapply uniq_quick; eauto. 
    + ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
      apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
      eapply uniq_appD in H3; des.
      repeat (apply in_consD in FH; des; clarify;
          first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
      eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
      eapply purge_lbl_nolbl; last eauto.
      eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
    + i; rewrite -defbefore_cons; eauto.
    + s; eauto.
    + ii. specialize (FR x0). edestruct FR; eauto; clear FR; s; eauto. 
      split; eauto.
      rrdes PU. specialize (PU0 _ _ IN0). specialize (PU _ _ IN0). des; clarify; i.
      repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
    + split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
      erewrite <- !(defbefore_cons); try eauto. eauto using purge_subinDl.
  - s in WF;  des.
    s; eapply mm_test.mm_compat_write; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto).
  - s in WF;  des.
    s; eapply mm_test.mm_compat_input; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto).
    + ii; s; rewrite cons_app; eapply uniq_quick; eauto. 
    + ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
      apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
      eapply uniq_appD in H3; des.
      repeat (apply in_consD in FH; des; clarify;
          first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
      eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
      eapply purge_lbl_nolbl; last eauto.
      eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
    + i; rewrite -defbefore_cons; eauto.
    + s; eauto.
    + ii. specialize (FR x). edestruct FR; eauto; clear FR; s; eauto. 
      split; eauto.
      rrdes PU. specialize (PU0 _ _ IN0). specialize (PU _ _ IN0). des; clarify; i.
      repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
    + split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
      erewrite <- !(defbefore_cons); try eauto. eauto using purge_subinDl.
  - s in WF;  des.
    s; eapply mm_test.mm_compat_output; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto).
  - s in WF;  try apply uniq_appD in UN; des.
    s; eapply mm_test.mm_compat_ifnz; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto using in_or_app)
    ; try (s; omega).
    + ii. rewrite <- (app_nil_l (purge_lbl _)). eapply uniq_quickl. eapply BV; eauto.
    + ii. hexploit BV; eauto. rewrite app_assoc. s bvars. ii.
      apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
      eapply uniq_appD in H3; des.
      repeat (apply in_consD in FH; des; clarify;
          first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
      eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
      eapply purge_lbl_nolbl; last eauto.
      eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
    + ii. rewrite <- (app_nil_l (purge_lbl _)). eapply uniq_quickr. eapply BV; eauto.
    + ii. hexploit BV; eauto. rewrite app_assoc. s bvars. ii.
      apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
      eapply uniq_appD in H3; des.
      repeat (apply in_consD in FH; des; clarify;
          first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto)).
      eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
      eapply purge_lbl_nolbl; last eauto.
      eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
  - s in WF;  des; try apply uniq_appD in UN0; des.
    s; eapply mm_test.mm_compat_case; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto using in_or_app)
    ; try (s; omega).
    + ii; s; rewrite cons_app; eapply uniq_quickl; eauto. eapply BV; eauto.
    + ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
      apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
      eapply uniq_appD in H3; des.
      repeat (apply in_consD in FH; des; clarify;
          first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto using in_or_app)).
      eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
      eapply purge_lbl_nolbl; last eauto.
      eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
    + i; rewrite -defbefore_cons; eauto.
    + s; eauto.
    + ii. specialize (FR x0). edestruct FR; eauto; clear FR; s; eauto using in_or_app.  
      split; eauto.
      rrdes PU. specialize (PU0 _ _ IN0). specialize (PU _ _ IN0). des; clarify; i.
      repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
    + split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
      erewrite <- !(defbefore_cons); try eauto. eauto using purge_subinDl.
    + ii; s; rewrite cons_app; eapply uniq_quickr; eapply BV; eauto.
    + ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
      apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
      eapply uniq_appD in H3; des.
      repeat (apply in_consD in FH; des; clarify;
          first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto using in_or_app)).
      eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
      eapply purge_lbl_nolbl; last eauto.
      eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
    + i; rewrite -defbefore_cons; eauto.
    + s; eauto.
    + ii. specialize (FR x0). edestruct FR; eauto; clear FR; s; eauto using in_or_app.  
      split; eauto.
      rrdes PU. specialize (PU0 _ _ IN0). specialize (PU _ _ IN0). des; clarify; i.
      repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
    + split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
      erewrite <- !(defbefore_cons); try eauto. eauto using purge_subinDl.
  - s in WF;  des.
    s; eapply mm_test.mm_compat_ret; rewrite ?ren_itvar_id; vauto
    ; try (by eapply FR; s; eauto)
    ; rewrite ?ren_id_self_t; eapply IH; eauto
    ; try (by i; apply FR; s; eauto).
  - s; destruct (list_assoc _ _ _) eqn:E.
    + destruct e
      ; try (by s in WF;  des
             ; s; eapply mm_test.mm_compat_app; rewrite ?ren_itvar_id; vauto).
      s in WF;  des.
      dup E.
      eapply blaaaa in E; eauto; first 1 last.
      * hexploit BV; eauto. ii.
        apply uniq_appD in H; des; apply uniq_appD in H0; des; clarify.
        { eapply H1. 
          - s; do 3!right; eauto. 
          - eapply in_or_app; right. eapply purge_lbl_nolbl; eauto using bvars_nolbl.
        }
        { eapply H1. 
          - s; do 3!right; eauto. 
          - eapply in_or_app; right. eapply purge_lbl_nolbl; eauto using bvars_nolbl.
        }
        { eapply H1. 
          - s; do 3!right; eauto. 
          - eapply in_or_app; right. eapply purge_lbl_nolbl; eauto using bvars_nolbl.
        }
      * s in E. rewrite E0 in E. eauto.
    + s in WF;  des.
      s; eapply mm_test.mm_compat_app; rewrite ?ren_itvar_id; vauto.
  - s; destruct (list_assoc _ _ _) eqn:E.
    + destruct e
      ; try (by s in WF;  des
             ; s; eapply mm_test.mm_compat_inst; rewrite ?ren_itvar_id; vauto).
      s in WF;  des.
      dup E.
      eapply blaaaa_inst in E; eauto; first 1 last.
      * hexploit BV; eauto. ii.
        apply uniq_appD in H; des; apply uniq_appD in H0; des; clarify.
        { eapply H1. 
          - s; do 1!right; eauto. 
          - eapply in_or_app; right. eapply purge_lbl_nolbl; eauto using bvars_nolbl.
        }
        { eapply H1. 
          - s; do 1!right; eauto. 
          - eapply in_or_app; right. eapply purge_lbl_nolbl; eauto using bvars_nolbl.
        }
      * s in E. rewrite E0 in E. eauto.
    + s in WF;  des.
      s; eapply mm_test.mm_compat_inst; rewrite ?ren_itvar_id; vauto
      ; try (by eapply FR; s; eauto).
  - destruct bx.
    + s in WF;  des.
      s; eapply mm_test.mm_compat_let; rewrite ?ren_itvar_id; vauto
      ; try (by eapply FR; s; eauto)
      ; rewrite ?ren_id_self_t.
      * eapply inline_asim; eauto.
        { ii. rewrite <- (app_nil_l (purge_lbl _)). eapply uniq_quickm. eapply BV; eauto. }
        { 

          ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
          apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
          eapply uniq_appD in H3; des.
          repeat (apply in_consD in FH; des; clarify;
                  first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto using in_or_app)).
          eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
          eapply purge_lbl_nolbl; last eauto.
          eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
        }
        { ii. specialize (FR x). edestruct FR; eauto; clear FR; s; eauto using in_or_app. }
        { i; eapply IH; eauto; s; omega. }
      * { eapply IH; eauto.
          - s; omega.
          - ii; s; rewrite cons_app. eapply uniq_quickr. eapply BV; eauto.
          - ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
            apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
            eapply uniq_appD in H3; des.
            repeat (apply in_consD in FH; des; clarify;
                  first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto using in_or_app)).
            eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
            eapply purge_lbl_nolbl; last eauto.
            eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
          - i; rewrite -defbefore_cons; eauto.
          - s; eauto.
          -  ii. specialize (FR x). edestruct FR; eauto; clear FR; s; eauto using in_or_app.  
             split; eauto.
             rrdes PU. specialize (PU0 _ _ IN0). specialize (PU _ _ IN0). des; clarify; i.
             repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
          - split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
            erewrite <- !(defbefore_cons); try eauto. eauto using purge_subinDl.
        }
    + s in WF;  des.
      s; eapply mm_test.mm_compat_letcont; rewrite ?ren_itvar_id; vauto
      ; try (by eapply FR; s; eauto)
      ; rewrite ?ren_id_self_c.
      * eapply inline_ksim; eauto.
        { ii. rewrite <- (app_nil_l (purge_lbl _)). eapply uniq_quickm. eapply BV; eauto. }
        { 
          ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
          apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
          eapply uniq_appD in H3; des.
          repeat (apply in_consD in FH; des; clarify;
                  first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto using in_or_app)).
          eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
          eapply purge_lbl_nolbl; last eauto.
          eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
          }
        { ii. specialize (FR x). edestruct FR; eauto; clear FR; s; eauto using in_or_app. }
        { i; eapply IH; eauto; s; try omega. }
      * { eapply IH; eauto.
          - s; omega.
          - ii; s; rewrite cons_app; eapply uniq_quickr; eapply BV; eauto.

          - ii. hexploit BV; eauto. rewrite app_assoc. s bvars. rewrite cons_app. ii.
            apply uniq_appD in H; des. dup H. apply uniq_appD in H2; des.
            eapply uniq_appD in H3; des.
            repeat (apply in_consD in FH; des; clarify;
                  first (eapply in_app_or in IN'; des; [eapply H4 | eapply H6]; last eauto; s; eauto using in_or_app)).
            eapply H1. eapply sublist_in; last exact IN'; eapply sublist_app; eauto.
            eapply purge_lbl_nolbl; last eauto.
            eapply in_app_or in IN'; des; eapply bvars_nolbl; eauto.
          - i; rewrite -defbefore_cons; eauto.
          - s; eauto.
          -  ii. specialize (FR x). edestruct FR; eauto; clear FR; s; eauto using in_or_app.  
             split; eauto.
             rrdes PU. specialize (PU0 _ _ IN0). specialize (PU _ _ IN0). des; clarify; i.
             repeat (rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl).
          - split; rrdes PU; eauto; ii; s; specialize (PU0 _ _ IN); specialize (PU _ _ IN); des; subst.
            erewrite <- !(defbefore_cons); try eauto. 
        }
Qed.

Lemma mm_msim_cons_l: forall FI FO (phia phib: clomem) x e ma mb
    (SIM: mm_msim FI FO phia phib ma mb)
    (FR: forall y e (IN: lookup y phia = Some e), y <> x),
  mm_msim FI FO ((x, e)::phia) phib ma mb.
Proof.
  i; rrdes SIM. econs; i.
  hexploit SIM; eauto.
  ii; hexploit FR; eauto; []; i.
  exploit PHIa; [by erewrite list_assoc_cons_neq; eauto|].
  i; des; esplits; eauto.
Qed.

Lemma mm_msim_cons_r: forall FI FO (phia phib: clomem) x e ma mb
    (SIM: mm_msim FI FO phia phib ma mb)
    (FR: forall y e (IN: lookup y phib = Some e), y <> x),
  mm_msim FI FO phia ((x, e)::phib) ma mb.
Proof.
  i; rrdes SIM. econs; i.
  hexploit SIM; eauto.
  ii; hexploit FR; eauto; []; i.
  exploit PHIb; [by erewrite list_assoc_cons_neq; eauto|].
  i; des; esplits; eauto.
Qed.


Definition inline_mod (m: module) : module :=
  inline_mod_ (analyze m) nil m.


Lemma lbls_upto_nolbl : forall x y xs
  (N: nolbl x),
  In x (lbls_upto y xs) -> False.
Proof.
  ii. induction xs; eauto.
  destruct a as [[la|ta]|ka]; s; eauto.
  revert H. s. destruct eq_nat_dec; clarify; s.
  ii; des; clarify. by eapply N.
Qed.

Lemma lbls_upto_notin_prefix y xs :
  lbls_upto y (xs ++ [inl (inl y)]) =  lbls_upto y xs.
Proof. 
  induction xs; eauto.
  - s; destruct eq_nat_dec; clarify.
  - destruct a as [[la|ta]|ka]; eauto.
    + s; destruct eq_nat_dec; clarify; eauto.
      by rewrite IHxs.
Qed.

Lemma lbls_upto_lbls x xs:
  (In x xs -> False) ->
  lbls_upto x (map (inl <*> inl) xs) = map (inl <*> inl) xs.
Proof.
  induction xs; eauto.
  ii. s; destruct eq_nat_dec; clarify.
  - exfalso; eapply H; s; auto.
  - rewrite IHxs; eauto.
    s in H. auto.
Qed.

Lemma lbls_upto_in_prefix x (xs ys : list var) :
  In (inl (inl x)) xs ->
  lbls_upto x (xs ++ ys) = lbls_upto x xs.
Proof.
  induction xs.
  - ii; clarify.
  - ii. apply in_consD in H; des; clarify.
    + s. destruct eq_nat_dec; clarify.
    + s. destruct a as [[la|ta]|ka]; clarify; eauto.
      destruct eq_nat_dec; clarify.
      by rewrite IHxs.
Qed.

Lemma lbls_upto_purge : forall x xs,
  lbls_upto x (purge xs) = lbls_upto x xs.
Proof.
  induction xs; eauto.
  destruct a as [[la|ta]|ka]; s; eauto.
  destruct eq_nat_dec; clarify; eauto.
  by rewrite IHxs.
Qed.
                       
Lemma purge_lbls_upto : forall x xs,
  purge (lbls_upto x  xs) = lbls_upto x xs.
Proof.
  induction xs; eauto.
  destruct a as [[la|ta]|ka]; s; eauto.
  destruct eq_nat_dec; clarify; eauto.
  s. by rewrite IHxs.
Qed.
  

Lemma inline_msim_: forall FI FO m lbls phi
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m)
    (UN: forall x e (IN: lookup x phi = Some e),
           uniq (bvars e ++ bvars_mod m))
    (WFA: forall x e (IN: lookup x phi = Some e),
            wfatom (defbefore x (map (inl<*>inl) FI)) e)
    (DOM: forall x e (IN: lookup x phi = Some e),
            In x (map (inl<*>inl) FI)),
  mm_msim FI FO nil phi (inline_mod_ lbls phi m) m.
Proof.
  i; ginduction WF; [by auto using mm_test.mm_msim_msim, mm_test.mm_msim_nil|].
  destruct e; vauto.
  {
  i; simpl inline_mod_.
  eapply mm_test.mm_msim_cons.
  - s; econs; eapply inline_psim; eauto using lwemp_stable.
    + s in WFE.
      rewrite map_comp in WFE.
      rewrite purge_map_inl in WFE. 
      rewrite map_comp.
      exact WFE. 
    + simpl bvars_mod in UN. 
      s purge_lbl. rewrite purge_lbl_map_inl_inl.
      ii. apply UN in IN. s in IN.
      apply uniq_appD in IN; des.
      rewrite cons_app3 in IN0.
      apply uniq_appD in IN0; des.
      apply uniq_appD in IN2; des.
      apply uniq_app; auto.
      apply uniq_app; auto.
      eapply uniq_perm; last (symmetry; eapply Permutation_rev); auto.
      ii. s in INb; des; clarify.
      eapply IN3; s; last (by eauto using in_or_app); auto. 
      eapply IN3; s; last (by eauto using in_or_app); auto.
      eapply IN3; s; last (by eauto using in_or_app); auto.
      ii. eapply IN1; eauto.
      rewrite cons_app3.
      rewrite app_assoc.
      apply in_or_app. left.
      eapply Permutation_in; last exact INb.
      eapply Permutation_trans.
      eapply Permutation_app_comm.
      eapply Permutation_app; eauto.
      eapply Permutation_rev.
    + i; apply UN in IN; s in IN. rewrite !app_comm_cons app_assoc in IN.
      apply uniq_appD in IN; des. rewrite cons_app3 in IN.
      eapply uniq_perm in IN; last first. 
      { eapply Permutation_app; first eauto. eapply Permutation_app_comm. }
      rewrite app_assoc in IN. apply uniq_appD in IN; des.
      specialize (IN3 _ IN').
      repeat (apply in_consD in FH; des; clarify; first eapply IN3; s; auto).
      apply in_map_iff in FH; des; clarify. apply in_app_or in IN'. 
      des; s in IN'; eapply bvars_nolbl; s; eauto.
    + ii. dup IN. apply WFA in IN.
      apply DOM in IN0.
      apply in_map_iff in IN0; des; clarify.
    + s; eauto.
    + ii. apply bvars_nolbl in IN. apply DOM in IN0.
      apply in_map_iff in IN0; des; clarify.
      destruct x as [[lx|tx]|kx]; try (by exfalso; eapply IN); vauto; splits; eauto;
      eapply lbls_upto_nolbl; eauto.
    + rr; splits; eauto. 
      * ii. apply DOM in IN; apply in_map_iff in IN; des; clarify; s; eauto.
      * ii. s. rewrite map_comp purge_map_inl. 
        apply DOM in IN; apply in_map_iff in IN; des; clarify; s; eauto.
  - apply mm_msim_cons_l; [|by vauto].
    des_ifs.
    + apply IHWF.
      * unfold uniqvar. unfold uniqvar in UV. s in UV. des.
        apply uniq_appD in UV2; des; eauto.
      * ii. apply list_assoc_cons_or in IN; des; clarify.
        unfold uniqvar. unfold uniqvar in UV. s in UV. des.
        apply uniq_appD in UV2; des; eauto.
        apply uniq_app; auto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        eapply X1; eauto. do 3!right; s; eauto using in_or_app.
      * ii. apply list_assoc_cons_or in IN; des; clarify.
        { s. 
          rewrite purge_lbls_upto.
          rewrite map_app.
          rewrite lbls_upto_notin_prefix; eauto.
          rewrite lbls_upto_lbls; eauto.
          s in WFE.
          rewrite map_comp in WFE.
          rewrite purge_map_inl in WFE. 
          rewrite map_comp. eauto.
        }
        { dup IN0; apply DOM in IN0; apply in_map_iff in IN0; des; clarify.
          apply WFA in IN1.
          rewrite map_app. s.
          rewrite lbls_upto_in_prefix; eauto.
          eauto. eapply in_map_iff; s; eauto.
        }
      * ii. apply list_assoc_cons_or in IN; des; clarify.
        { s. eapply in_map_iff; eexists; split; s; eauto. eapply in_or_app. s; auto. }
        { s. apply DOM in IN0. rewrite map_app. eapply in_or_app; eauto. }
    + apply mm_msim_cons_r; last first.
      { ii. dup IN. apply DOM in IN0; apply in_map_iff in IN0; des; ss; clarify. }
      apply IHWF.
      * unfold uniqvar. unfold uniqvar in UV. s in UV. des.
        apply uniq_appD in UV2; des; eauto.
      * ii. unfold uniqvar in UV. s in UV. des. apply uniq_appD in UV2; des; eauto. apply uniq_app; auto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        eapply X1; eauto. do 3!right; s; eauto using in_or_app.
      * ii. s. dup IN; apply DOM in IN0; apply in_map_iff in IN0; des; clarify.
          rewrite map_app. s.
          rewrite lbls_upto_in_prefix; eauto.
          apply WFA in IN; eauto.
      * i; rewrite map_app; apply in_or_app; s; eauto.
  - auto.
  - vauto.
  - ii. apply DOM in IN; apply in_map_iff in IN; des; clarify; s; eauto.
  - auto using lwemp_stable.
  - ii; rr; auto.
 }
  {
  i; simpl inline_mod_.
  eapply mm_test.mm_msim_cons.
  - s; econs; eapply inline_psim; eauto using lwemp_stable.
    + s in WFE.
      rewrite map_comp in WFE.
      rewrite purge_map_inl in WFE. 
      rewrite map_comp.
      exact WFE. 
    + simpl bvars_mod in UN. 
      s purge_lbl. rewrite purge_lbl_map_inl_inl.
      ii. apply UN in IN. s in IN.
      apply uniq_appD in IN; des.
      rewrite cons_app in IN0.
      apply uniq_appD in IN0; des.
      apply uniq_appD in IN2; des.
      apply uniq_app; auto.
      apply uniq_app; auto.
      ii. eapply IN3; eauto using in_or_app.
      ii. eapply IN1; eauto. rewrite app_comm_cons. eapply in_or_app. left.
      eapply Permutation_in.
      rewrite cons_app. eapply Permutation_trans.
      eapply Permutation_app_comm. 
      eapply Permutation_app; eauto.
      eauto.
    + i; apply UN in IN; s in IN. rewrite !app_comm_cons app_assoc in IN.
      apply uniq_appD in IN; des. rewrite cons_app in IN.
      eapply uniq_perm in IN; last first. 
      { eapply Permutation_app; first eauto. eapply Permutation_app_comm. }
      rewrite app_assoc in IN. apply uniq_appD in IN; des.
      specialize (IN3 _ IN').
      repeat (apply in_consD in FH; des; clarify; first eapply IN3; s; auto).
      apply in_map_iff in FH; des; clarify. apply in_app_or in IN'. 
      des; s in IN'; eapply bvars_nolbl; s; eauto.
    + ii. dup IN. apply WFA in IN.
      apply DOM in IN0.
      apply in_map_iff in IN0; des; clarify.
    + s; eauto.
    + ii. apply bvars_nolbl in IN. apply DOM in IN0.
      apply in_map_iff in IN0; des; clarify.
      destruct x as [[lx|tx]|kx]; try (by exfalso; eapply IN); vauto; splits; eauto;
      eapply lbls_upto_nolbl; eauto.
    + rr; splits; eauto. 
      * ii. apply DOM in IN; apply in_map_iff in IN; des; clarify; s; eauto.
      * ii. s. rewrite map_comp purge_map_inl. 
        apply DOM in IN; apply in_map_iff in IN; des; clarify; s; eauto.
  - apply mm_msim_cons_l; [|by vauto].
    des_ifs.
    + apply IHWF.
      * unfold uniqvar. unfold uniqvar in UV. s in UV. des.
        apply uniq_appD in UV0; des; eauto.
      * ii. apply list_assoc_cons_or in IN; des; clarify.
        unfold uniqvar. unfold uniqvar in UV. s in UV. des.
        apply uniq_appD in UV0; des; eauto.
        apply uniq_app; auto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        eapply X1; eauto. do 1!right; s; eauto using in_or_app.
      * ii. apply list_assoc_cons_or in IN; des; clarify.
        { s. 
          rewrite purge_lbls_upto.
          rewrite map_app.
          rewrite lbls_upto_notin_prefix; eauto.
          rewrite lbls_upto_lbls; eauto.
          s in WFE.
          rewrite map_comp in WFE.
          rewrite purge_map_inl in WFE. 
          rewrite map_comp. eauto.
        }
        { dup IN0; apply DOM in IN0; apply in_map_iff in IN0; des; clarify.
          apply WFA in IN1.
          rewrite map_app. s.
          rewrite lbls_upto_in_prefix; eauto.
          eauto. eapply in_map_iff; s; eauto.
        }
      * ii. apply list_assoc_cons_or in IN; des; clarify.
        { s. eapply in_map_iff; eexists; split; s; eauto. eapply in_or_app. s; auto. }
        { s. apply DOM in IN0. rewrite map_app. eapply in_or_app; eauto. }
    + apply mm_msim_cons_r; last first.
      { ii. dup IN. apply DOM in IN0; apply in_map_iff in IN0; des; ss; clarify. }
      apply IHWF.
      * unfold uniqvar. unfold uniqvar in UV. s in UV. des.
        apply uniq_appD in UV0; des; eauto.
      * ii. unfold uniqvar in UV. s in UV. des. apply uniq_appD in UV0; des; eauto. apply uniq_app; auto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        hexploit UN; eauto; ii. eapply uniq_appD in X; des; eauto.
        eapply X1; eauto. do 1!right; s; eauto using in_or_app.
      * ii. s. dup IN; apply DOM in IN0; apply in_map_iff in IN0; des; clarify.
          rewrite map_app. s.
          rewrite lbls_upto_in_prefix; eauto.
          apply WFA in IN; eauto.
      * i; rewrite map_app; apply in_or_app; s; eauto.
  - auto.
  - vauto.
  - ii. apply DOM in IN; apply in_map_iff in IN; des; clarify; s; eauto.
  - auto using lwemp_stable.
  - ii; rr; auto.
 }
Grab Existential Variables.
  exact tt.
  exact tt.
Qed.

Theorem inline_msim: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  msim mm_gw FI FO (inline_mod m) m.
Proof.
  i; unfold inline_mod.
  apply mm_test.mm_msim_msim.
  apply inline_msim_; vauto.
Qed.

Lemma inline_good_thunk : forall e (GT : is_good_thunk e) phi
    (GEfix : forall x bf bx bk e (IN : lookup x phi = Some (IEfix bf bx bk e)),
               good_exp e)
    (GEgen : forall x bf e (IN : lookup x phi = Some (IEgen bf e)),
               good_exp e),
  is_good_thunk e ->
  is_good_thunk (inline phi e).
Proof.
  ii. 
  induction e; i; eauto;
  try (by invc H);
  (
  s; ii; invc FH; s in GT; (revert H4 || revert H2);
  destruct e; try discriminate; ss; eauto;
  generalize (GEfix (inl xf)); generalize (GEgen (inl xf)); s;
  destruct (lookup (inl xf) phi) eqn:E; s in E; rewrite E; try discriminate;
     destruct e; try discriminate; intros g f N;
     (specialize (f _ _ _ _ eq_refl) || specialize (g _ _ eq_refl));
     destruct e; try discriminate;
     try (by (destruct (_ : termvar + contvar) in N));
     try (by ((eapply f; eauto) || (eapply g; eauto)));
     by (eapply g; eauto) 
   
  ).
Qed. 

Lemma in2_split {X} (x y : X) xs : 
  In x xs ->
  In y xs ->
  exists xs',
    x <> y /\ Permutation (x::y::xs') xs 
    \/ x = y /\ Permutation (x::xs') xs.
Proof.
  revert x y. induction xs; ii; vauto.
  apply in_consD in H; apply in_consD in H0; des; clarify.
  - exists (xs). right. eauto.
  - apply in_split in H; des; clarify.
    exists (l1 ++ l2).
    left; splits; auto.
    eapply Permutation_trans.
    { eapply perm_swap. }
    eapply Permutation_cons.
    eapply Permutation_middle.
  - apply in_split in H0; des; clarify.
    exists (l1 ++ l2).
    left; splits; auto.
    eapply Permutation_cons.
    eapply Permutation_middle.
  - hexploit IHxs; [exact H | exact H0 | ii; des; clarify ].
    + exists (a::xs').
      left; splits; auto.
      rewrite cons_app2. eapply Permutation_trans.
      symmetry.
      eapply Permutation_middle.
      eapply Permutation_cons. eauto.
    + exists (a::xs').
      right; splits; auto.
      rewrite cons_app. eapply Permutation_trans.
      symmetry.
      eapply Permutation_middle.
      eapply Permutation_cons. eauto.
Qed.

 
Lemma inline_wf : forall FI xs phi e
    (DOM: forall x e (IN: lookup x phi = Some e),
            In x (map (inl<*>inl) FI))
    (WFA: forall x e (IN: lookup (inl (inl x)) phi = Some e),
            wfatom (defbefore (inl (inl x)) (map (inl<*>inl) FI)) e)
    (WFfix : forall x bf bx bk e (IN : lookup (inl (inl x)) phi = Some (IEfix bf bx bk e)),
               wfexp (inr bk :: inl (inr bx) :: inl (inr bf) :: purge (defbefore (inl (inl x)) (map (inl<*>inl) FI))) e)
    (WFgen : forall x bf e (IN : lookup (inl (inl x)) phi = Some (IEgen bf e)),
               wfexp (inr bf :: purge (defbefore (inl (inl x)) (map (inl<*>inl) FI))) e)
                  ,
  (forall (WF: wfatom (xs ++ map (inl<*>inl) FI) e), wfatom (xs ++ map (inl<*>inl) FI) (inline phi e))
  /\ (forall (WF: wfexp (xs ++ map (inl<*>inl) FI) e), wfexp (xs ++ map (inl<*>inl) FI) (inline phi e)).
Proof.
  ii. revert  FI xs phi DOM WFA WFfix WFgen; induction e; ii; eauto;
  try (splits; eauto; s; by eapply IHe);
  try (splits; eauto; s; splits; des; eauto; by eapply IHe);
  try (splits; eauto; s; splits; des; eauto; [by eapply IHe1|by eapply IHe2]).
  - splits; eauto; s.
    rewrite cons_app3. rewrite purge_app app_assoc.
    rewrite map_comp purge_map_inl. rewrite <- map_comp.
    eapply IHe; eauto.
  - splits; eauto; s.
    rewrite cons_app. rewrite purge_app app_assoc.
    rewrite map_comp purge_map_inl. rewrite <- map_comp.
    eapply IHe; eauto.
  - splits; eauto; s.
    rewrite cons_app. rewrite app_assoc.
    eapply IHe; eauto.
  - splits; eauto; s.
    rewrite cons_app. rewrite app_assoc; ii.
    des; splits; eauto.
    eapply IHe; eauto.
  - splits; eauto; s.
    rewrite cons_app. rewrite app_assoc; ii.
    des; splits; eauto.
    eapply IHe; eauto.
  - splits; eauto; s.
    rewrite cons_app. rewrite app_assoc; ii.
    des; splits; eauto.
    eapply IHe; eauto.
  - splits; eauto; s.
    rewrite cons_app. rewrite app_assoc; ii.
    des; splits; eauto.
    eapply IHe1; eauto.
    eapply IHe2; eauto.
  - splits; eauto; s; clarify.
    ii; des.
    destruct (lookup (inl xf) phi) eqn:H; rewrite H.
    + hexploit DOM; eauto; ii. apply in_map_iff in X; s in X; des; clarify.
      hexploit WFA; eauto; ii.
      destruct e; s; eauto.
      hexploit WFfix; eauto; ii.
      s in X. 
      eapply wfexp_dup; [exact WF|].
      eapply wfexp_dup; [right; exact WF0|].
      eapply wfexp_dup; [right; right; exact WF1|].
      eapply rename_wfexp; [| |reflexivity]; last first.
      * do !econs. 
      * eapply wfexp_sublist; [exact X1|].
        do !eapply sublist_cons_keep. s. rewrite purge_lbls_upto.
        eapply sublist_r_app_r. eapply lbls_upto_sublist.
    + s; splits; eauto.
  - splits; eauto; s; clarify.
    ii; des.
    destruct (lookup (inl xf) phi) eqn:H; rewrite H.
    + hexploit DOM; eauto; ii. apply in_map_iff in X; s in X; des; clarify.
      hexploit WFA; eauto; ii.
      destruct e; s; eauto.
      hexploit WFgen; eauto; ii.
      des.
      destruct (in2_split _ _ _ WF0 WF) as [xs' Hxs']; des; clarify.
      eapply wfexp_dup; [exact WF0|].
      eapply rename_wfexp; [| |reflexivity]; last first.
      * do 2!econs.
      * eapply wfexp_sublist; [exact X1|].
        eapply sublist_cons_keep. s. rewrite purge_lbls_upto.
        eapply sublist_r_app_r. eapply lbls_upto_sublist.
    + s; splits; eauto.
  - splits; eauto; ii; eauto.
    destruct bx; s; splits; eauto; s in WF; des; eauto.
    + eapply IHe1; eauto.
    + rewrite app_comm_cons; eapply IHe2; eauto.
    + eapply IHe1; eauto.
    + rewrite app_comm_cons; eapply IHe2; eauto.
Qed.

Lemma inline_wfmod_: forall FI FO m lbls phi
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m)
    (GTA: forall x e (IN: lookup x phi = Some e),
            is_good_thunk e)
    (GEfix : forall x bf bx bk e (IN : lookup x phi = Some (IEfix bf bx bk e)),
               good_exp e)
    (GEgen : forall x bf e (IN : lookup x phi = Some (IEgen bf e)),
               good_exp e)
    (WFA: forall x e (IN: lookup x phi = Some e),
            wfatom (defbefore x (map (inl<*>inl) FI)) e)
    (DOM: forall x e (IN: lookup x phi = Some e),
            In x (map (inl<*>inl) FI)),
  mid_wfmod FI (inline_mod_ lbls phi m) FO.
Proof.
  i; ginduction WF; [by auto using mm_test.mm_msim_msim, mm_test.mm_msim_nil|].
  ii; econs; eauto using inline_good_thunk.
  - destruct e; vauto.
    + s. rewrite cons_app3. rewrite map_comp purge_map_inl. 
      rewrite <- map_comp.
      eapply inline_wf; eauto.
      * ii. s. hexploit WFA; eauto.
      * ii. s. hexploit WFA; eauto.
      * revert WFE. s. by rewrite map_comp purge_map_inl.
    + s. rewrite cons_app. rewrite map_comp purge_map_inl. 
      rewrite <- map_comp.
      eapply inline_wf; eauto.
      * ii. s. hexploit WFA; eauto.
      * ii. s. hexploit WFA; eauto.
      * revert WFE. s. by rewrite map_comp purge_map_inl.
  - eapply IHWF; eauto.
    + eapply uniq_appD in UV; des; eauto.
    + destruct (list_mem eq_nat_dec f lbls) eqn:H; rewrite H; eauto.
      ii. eapply list_assoc_cons_or in IN; des; clarify.
      eapply GTA; eauto.
    + destruct (list_mem eq_nat_dec f lbls) eqn:H; rewrite H; eauto.
      ii. eapply list_assoc_cons_or in IN; des; clarify.
      eapply GEfix; eauto.
    + destruct (list_mem eq_nat_dec f lbls) eqn:H; rewrite H; eauto.
      ii. eapply list_assoc_cons_or in IN; des; clarify.
      eapply GEgen; eauto.
    + destruct (list_mem eq_nat_dec f lbls) eqn:H; rewrite H; eauto.
      * ii. eapply list_assoc_cons_or in IN; des; clarify.
        { s. rewrite map_app. s.
          rewrite lbls_upto_notin_prefix. eauto.
          rewrite lbls_upto_lbls; eauto. }
        { hexploit DOM; eauto. ii. eapply in_map_iff in H0; des; clarify.
          s. rewrite map_app; s. rewrite lbls_upto_in_prefix; eauto.
          by eapply (WFA (inl (inl x0))).
        }
      * ii. hexploit DOM; eauto. ii. eapply in_map_iff in H0; des; clarify.
          s. rewrite map_app; s. rewrite lbls_upto_in_prefix; eauto.
          by eapply (WFA (inl (inl x0))).
    + destruct (list_mem eq_nat_dec f lbls) eqn:H; rewrite H; eauto.
      * ii. eapply list_assoc_cons_or in IN; des; clarify.
        { rewrite map_app. apply in_or_app; right; s; auto. }
        { rewrite map_app. apply in_or_app; left; eapply DOM; eauto. }
      * ii. rewrite map_app. apply in_or_app; left; eapply DOM; eauto.
Qed.

Theorem inline_wfmod: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  mid_wfmod FI (inline_mod m) FO.
Proof.
  i; unfold inline_mod.
  eapply inline_wfmod_; ii; ss.
Qed.

Theorem inline_freshen_correct:
  <<WF: modtf_wf (fun _ => freshen_mod <*> inline_mod)>> /\
  <<SIM: modtf_sim (fun _ => freshen_mod <*> inline_mod)>>.
Proof.
  split; ii.
  - hexploit inline_wfmod; eauto; []; i.
    split; [eapply freshen_wfmod|eapply freshen_uniq]; eauto.
  - hexploit inline_msim; eauto; []; i.
    econs 3; [eapply freshen_msim|]; eauto using inline_wfmod.
Qed.
