(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program Relations.
Require Import countable vgtac basis lang_common lang_spec lang_mid.
Require Import model_common model_untyped model_properties_untyped gw_untyped
        mm mm_test.
Require Import ssr util lang_mid_renaming opt_freshen.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


Lemma fold_within: forall (A : Type) (f : A -> A -> A)
   (ASSOC: forall x y z : A, f x (f y z) = f (f x y) z)
   (COMM: forall x y : A, f x y = f y x)
   (ID: forall x, f x x = x),
  forall a0 l0 l1,
    f (fold_right f a0 l0) (fold_right f a0 l1) =
    (fold_right f (fold_right f a0 l1) l0).
Proof.
  ii; rewrite <-fold_right_app; revert l0.
  induction l1 as [|l1']; ss; ii; try rewrite (app_nil_r _)
  ; induction l0 as [|l0']; ss.
  - rewrite <-ASSOC; rewrite ->IHl0; auto.
  - rewrite (COMM l1'); rewrite ASSOC; rewrite (IHl1 nil); auto.
  - rewrite <-IHl0; rewrite <-ASSOC; auto.
Qed.

(** * Expressions annotated with data of type [A]. *)

Inductive aexp (A: Type) :=
| IEunit
| IEnat  (n:nat)
| IEpair (x:itermvar) (y:itermvar)
| IEinl  (x:itermvar)
| IEinr  (x:itermvar)
| IEbop  (op: binop) (x:itermvar + nat) (y:itermvar + nat)
| IEfst  (x:itermvar)
| IEsnd  (x:itermvar)
| IEfix  (bf:termvar) (bx:termvar) (bk:contvar) (e:aexp A)
| IEgen  (bk:contvar) (e:aexp A)
| IEcont (bx:termvar) (e:aexp A)
| IEnew    (x:itermvar) (bx:termvar) (e:aexp A)
| IEread   (x:itermvar) (bx:termvar) (e:aexp A)
| IEwrite  (x:itermvar) (y:itermvar) (e:aexp A)
| IEinput  (bx:termvar) (e:aexp A)
| IEoutput (x:itermvar) (e:aexp A)
| IEifnz   (x:itermvar) (ea:aexp A) (eb:aexp A)
| IEcase   (x:itermvar) (bx:termvar) (ea:aexp A) (eb:aexp A)
| IEret    (xk:contvar) (x:itermvar)
| IEapp (a:A) (xf:itermvar) (x:itermvar) (xk:contvar)
| IEinst (a:A) (xf:itermvar) (xk:contvar)
| IElet (a:A) (ae:aexp A) (bx:termvar + contvar) (e:aexp A)
| IErefeq (x:itermvar) (y:itermvar) 
.


(** Erase annotations. *)

Fixpoint aerase A (e: aexp A) : exp :=
  match e with
    | IEunit
      => lang_mid.IEunit
    | IEnat n
      => lang_mid.IEnat n
    | IEpair x y
      => lang_mid.IEpair x y
    | IEret k x
      => lang_mid.IEret k x
    | IEinst _ x k
      => lang_mid.IEinst x k
    | IEinl x
      => lang_mid.IEinl x
    | IEinr x
      => lang_mid.IEinr x
    | IEfst x
      => lang_mid.IEfst x
    | IEsnd x
      => lang_mid.IEsnd x
    | IEbop b xn yn
      => lang_mid.IEbop b xn yn
    | IEapp _ x y k
      => lang_mid.IEapp x y k
    | IErefeq x y
      => lang_mid.IErefeq x y

    | IEfix f x k ea
      => lang_mid.IEfix f x k (aerase ea)
    | IEgen x ea
      => lang_mid.IEgen x (aerase ea)
    | IEcont x ea
      => lang_mid.IEcont x (aerase ea)
    | IEnew x y ea
      => lang_mid.IEnew x y (aerase ea)
    | IEread x y ea
      => lang_mid.IEread x y (aerase ea)
    | IEoutput x ea
      => lang_mid.IEoutput x (aerase ea)
    | IEwrite x y ea
      => lang_mid.IEwrite x y (aerase ea)
    | IEinput x ea
      => lang_mid.IEinput x (aerase ea)

    | IEifnz x ea eb
      => lang_mid.IEifnz x (aerase ea) (aerase eb)
    | IEcase x y ea eb
      => lang_mid.IEcase x y (aerase ea) (aerase eb)
    | IElet t ea x eb 
      => lang_mid.IElet (aerase ea) x (aerase eb)
  end.


Definition annotate_with A a : exp -> aexp A :=
  fix aw (e: exp) :=
  match e with
    | lang_mid.IElet e1 x e2
      => IElet a (aw e1) x (aw e2)

    | lang_mid.IEapp x y k
      => IEapp a x y k

    | lang_mid.IEinst x k
      => IEinst a x k

    | lang_mid.IEunit
      => IEunit _
    | lang_mid.IEnat n
      => IEnat _ n
    | lang_mid.IEpair x y
      => IEpair _ x y
    | lang_mid.IEret k x
      => IEret _ k x
    | lang_mid.IEinl x
      => IEinl _ x
    | lang_mid.IEinr x
      => IEinr _ x
    | lang_mid.IEfst x
      => IEfst _ x
    | lang_mid.IEsnd x
      => IEsnd _ x
    | lang_mid.IEbop b xn yn
      => IEbop _ b xn yn
    | lang_mid.IErefeq x y
      => IErefeq _ x y

    | lang_mid.IEfix f x k ea
      => IEfix f x k (aw ea)
    | lang_mid.IEgen x ea
      => IEgen x (aw ea)
    | lang_mid.IEcont x ea
      => IEcont x (aw ea)
    | lang_mid.IEnew x y ea
      => IEnew x y (aw ea)
    | lang_mid.IEread x y ea
      => IEread x y (aw ea)
    | lang_mid.IEoutput x ea
      => IEoutput x (aw ea)
    | lang_mid.IEwrite x y ea
      => IEwrite x y (aw ea)
    | lang_mid.IEinput x ea
      => IEinput x (aw ea)

    | lang_mid.IEifnz x ea eb
      => IEifnz x (aw ea) (aw eb)
    | lang_mid.IEcase x y ea eb
      => IEcase x y (aw ea) (aw eb)
  end.

Lemma annotate_with_erase: forall A (a: A) e,
  aerase (annotate_with a e) = e.
Proof.
  i; induction e; s; rewrite ?IHe ?IHe1 ?IHe2; eauto.
Qed.



(** * Expression transformations as annotations. *)

Definition exptf :=
  list var -> nat -> exp -> option (exp * nat).


Definition list_between n n' : list nat :=
  seq n (n'-n).

Definition vars2nat (l : list var) := map varid l.
Hint Unfold vars2nat.

Fixpoint nextbvar' (e : exp) : nat :=
  match e with
  | lang_mid.IEunit => 1
  | lang_mid.IEnat _ => 1
  | lang_mid.IEpair _ _ => 1
  | lang_mid.IEinl _ => 1
  | lang_mid.IEinr _ => 1
  | lang_mid.IEbop _ _ _ => 1
  | lang_mid.IEfst _ => 1
  | lang_mid.IEsnd _ => 1
  | lang_mid.IEfix bf bx bk e0 =>
      max
       (nextbvar' e0) 
       (S (fold_left max (vars2nat ([inl (inr bf); inl (inr bx); inr bk])) 0))
       
  | lang_mid.IEgen bk e0 => max (nextbvar' e0) (S (varid (inr bk)))
  | lang_mid.IEcont bx e0 => max (nextbvar' e0) (S (varid (inl (inr bx))))
  | lang_mid.IEnew _ bx e0 => max (nextbvar' e0) (S (varid (inl (inr bx))))
  | lang_mid.IEread _ bx e0 => max (nextbvar' e0) (S (varid (inl (inr bx))))
  | lang_mid.IEwrite _ _ e0 => nextbvar' e0
  | lang_mid.IEinput bx e0 => max (nextbvar' e0) (S (varid (inl (inr bx))))
  | lang_mid.IEoutput _ e0 => nextbvar' e0
  | lang_mid.IEifnz _ ea eb => max (nextbvar' ea) (nextbvar' eb)
  | lang_mid.IEcase _ bx ea eb =>
      max (max (nextbvar' ea) (nextbvar' eb)) (S (varid (inl (inr bx))))
  | lang_mid.IEret _ _ => 1
  | lang_mid.IEapp _ _ _ => 1
  | lang_mid.IEinst _ _ => 1
  | lang_mid.IElet ea (inl bx) eb =>
      max (max (nextbvar' ea) (nextbvar' eb)) (S (varid (inl (inr bx))))
  | lang_mid.IElet ea (inr bk) eb => 
      max (max (nextbvar' ea) (nextbvar' eb)) (S (varid (inr bk)))
  | lang_mid.IErefeq _ _ => 1
  end.

Lemma vars2nat_sublist: forall xs ys
    (SUB: sublist xs ys),
  sublist (vars2nat xs) (vars2nat ys).
Proof.
  i; ginduction ys; ii; try (by inversion SUB; auto)
  ; destruct a as [[?|?]|?]; inversion SUB; s; clarify; auto.
Qed.

Lemma vars2nat_purge_sublist: forall xs,
  sublist (vars2nat (purge xs)) (vars2nat xs).
Proof.
 ii; apply vars2nat_sublist; apply purge_sublist.
Qed.

Ltac svbv :=
  simpl bvars in *; simpl vars2nat in *; unfold vars2nat in *
  ; repeat first [ rewrite ->!map_app in *
                 | rewrite <-!app_assoc in *
                 | rewrite <-app_comm_cons in *].
  
  


Fixpoint boundvar (m : module) (f : nat -> nat -> nat) : nat :=
  match m with
  | nil => 0
  | (_,e) :: m' => f (fold_left f (vars2nat (bvars e)) 0) (boundvar m' f)
  end.
Hint Unfold boundvar.

Definition minvar : module -> nat :=
  fun m => boundvar m min.


Definition maxbvar (e : exp) : nat :=
  fold_left max (vars2nat (bvars e)) 0.

Lemma maxbvar_l_rewrite: forall e,
  maxbvar e = fold_left max (vars2nat (bvars e)) 0.
Proof. auto. Qed.

Lemma maxbvar_r_rewrite: forall e,
  maxbvar e = fold_right max 0 (vars2nat (bvars e)).
Proof. 
 ii; rewrite <-fold_symmetric
 ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm); auto.
Qed.

Definition nextbvar (e : exp) : nat :=
  S (maxbvar e).


Ltac simpl_nbvar :=
repeat
  (try (match goal with
       | |- context[nextbvar ?e] =>
         unfold nextbvar,maxbvar; rewrite ->fold_symmetric by
           (apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm)
         ; simpl bvars; simpl vars2nat; simpl fold_right
         ; try (rewrite !Max.succ_max_distr); rewrite <-fold_symmetric
         ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm)
         ; (match goal with
              | |- context[S (fold_left max (vars2nat (bvars ?e)) 0)] =>
                  replace (S (fold_left max (vars2nat (bvars e)) 0)) with
                  (nextbvar e) by vauto
              end)
       end)
   ; match goal with
     | [MX: context[nextbvar ?e] |- _] =>
       unfold nextbvar,maxbvar in MX
       ; rewrite ->fold_symmetric in MX by
                   (apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm)
       ; simpl bvars in MX; simpl vars2nat in MX; simpl fold_right in MX
       ; try (rewrite !Max.succ_max_distr in MX)
       ; rewrite <-fold_symmetric in MX
       ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm)
       ; (match goal with
          | [H: context[S (fold_left max (vars2nat (bvars ?e)) 0)] |- _]
            =>
              replace (S (fold_left max (vars2nat (bvars e)) 0)) with
              (nextbvar e) in H by vauto
          end)
     end).

Ltac ssbvar :=
  simpl nextbvar'
  ; repeat first [ rewrite ->!NPeano.Nat.succ_max_distr in *
                 | rewrite ->!NPeano.Nat.max_lub_iff in *; des].

Lemma nextbvar_equiv: forall e,
    nextbvar e = nextbvar' e.
Proof.
  induction e; try (by unfold nextbvar,maxbvar; s; vauto)
  ; try (s; unfold nextbvar, maxbvar; rewrite fold_symmetric
         ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm)
         ; s; rewrite !Max.succ_max_distr; rewrite <-fold_symmetric
         ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm)
         ; rewrite <-maxbvar_l_rewrite)
  ; try (by rewrite <-IHe; rhs3 rewrite NPeano.Nat.max_comm; auto).
  - rewrite <-IHe; rhs3 rewrite NPeano.Nat.max_comm; s.
    rewrite NPeano.Nat.max_assoc; rhs rewrite <-NPeano.Nat.max_assoc; auto.
  - s; unfold nextbvar, maxbvar; svbv
    ; rewrite fold_symmetric
    ; first (rewrite fold_right_app; rewrite <-fold_within)
    ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm
              || apply NPeano.Nat.max_idempotent).
    rewrite <-!maxbvar_r_rewrite; rewrite !Max.succ_max_distr
    ; rewrite <-IHe1; rewrite <-IHe2; unfold nextbvar; vauto.
  - s; unfold nextbvar, maxbvar; svbv
    ; rewrite fold_symmetric; s
    ; first (rewrite fold_right_app; rewrite <-fold_within)
    ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm
              || apply NPeano.Nat.max_idempotent)
    ; rewrite <-!maxbvar_r_rewrite; rewrite !Max.succ_max_distr
    ; rewrite <-IHe1; rewrite <-IHe2
    ; rhs3 rewrite NPeano.Nat.max_comm; s; auto.
  - destruct bx; s; unfold nextbvar, maxbvar; svbv
    ; rewrite fold_symmetric; s
    ; try (rewrite fold_right_app; rewrite <-fold_within)
    ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm
              || apply NPeano.Nat.max_idempotent)
    ; rewrite <-!maxbvar_r_rewrite; rewrite !Max.succ_max_distr
    ; rewrite <-IHe1; rewrite <-IHe2
    ; rhs3 rewrite NPeano.Nat.max_comm; s; auto.
Qed.


Fixpoint maxvar (m : module) : nat :=
  match m with
  | nil => 0
  | (_,e) :: m' => max (fold_left max (vars2nat (bvars e)) 0) (maxvar m')
  end.
Hint Unfold maxvar.

Definition nextvar (m : module) : nat :=
  S (maxvar m).
Hint Unfold nextvar.

Lemma largest_in_bvar: forall e x
    (IN: In x (vars2nat (bvars e))),
  x <= maxbvar e.
Proof.
  ii; unfold maxbvar; rewrite fold_symmetric
  ; try (by apply NPeano.Nat.max_assoc || apply NPeano.Nat.max_comm)
  ; revert IN; generalize (vars2nat (bvars e)).
  induction l as [|l']; vauto.
  ii; ss; des; clarify; firstorder using NPeano.Nat.max_le_iff.
Qed.

Lemma nextbvar_rename: forall e r,
    nextbvar (rename r e) = nextbvar e.
Proof.
  ii; unfold nextbvar,maxbvar; rewrite rename_bvars; auto.
Qed.

Lemma maxvar_largest: forall m x
    (IN: In x (vars2nat (bvars_mod m))),
  x <= maxvar m.
Proof.
  ii; revert IN; induction m as [|(f,e) m']; vauto.
  simpl maxvar; svbv; ii. 
  apply in_app_or in IN; des.
  - apply NPeano.Nat.max_le_iff; left; exploit largest_in_bvar; eauto.
  - firstorder; apply NPeano.Nat.max_le_iff; right; auto.
Qed.

Lemma le_nextvar: forall m f e n
    (LE: nextvar ((f,e) :: m) <= n),
  nextvar m <= n.
Proof.
  ii; unfold nextvar in LE; simpl maxvar in LE
  ; rewrite Max.succ_max_distr in LE.
  replace (S (maxvar m)) with (nextvar m) in LE by vauto.
  apply NPeano.Nat.max_lub_r in LE; auto.
Qed.
  
Lemma nextvar_nin_mod: forall m n
    (LE: nextvar m <= n),
  ~ In n (vars2nat (bvars_mod m)).
Proof.
  induction n as [|n']; ii; try (by unfold nextvar in LE; omega).
  apply maxvar_largest in FH; unfold nextvar in LE; omega.
Qed.

Lemma nextvar_nin_exp: forall e n
    (LE: (S (maxbvar e)) <= n),
  ~ In n (vars2nat (bvars e)).
Proof.
  induction n as [|n']; ii; try (by unfold nextvar in LE; omega).
  apply largest_in_bvar in FH; omega.
Qed.

Lemma nextvar_rewrite: forall f e m,
  nextvar ((f,e)::m) = max (nextbvar e) (nextvar m).
Proof.
  ii; lhs3 unfold nextvar; unfold nextbvar; simpl maxvar
  ; rewrite Max.succ_max_distr.
  replace (fold_left max (vars2nat (bvars e)) 0) with (maxbvar e) by vauto.
  replace (S (maxvar m)) with (nextvar m) by vauto.
  vauto.
Qed.

Lemma what: forall m n
    (LE: nextvar m <= n)
    (UN: uniq (vars2nat (bvars_mod m))),
  uniq (n :: vars2nat (bvars_mod m)).
Proof.
  induction m as [|(f,e) m']; vauto; []; ii; ss; splits; auto
  ; replace (bvars e ++ bvars_mod m') with (bvars_mod ((f,e) :: m')) by vauto
  ; apply nextvar_nin_mod; auto.
Qed.

Definition exptf_wf f :=
  forall e e' G n n'
      (E: f G n e = Some (e', n'))
      (MX: (nextbvar e) <= n)
      (WF: wfexp G e)
      (UN: forall m (BIG: n <= m),
             uniq (m :: vars2nat (bvars e) ++ vars2nat (purge_lbl G))),
    <<MAX: max (nextbvar e') n <= n' >> /\
    <<WF: wfexp G e'>> /\
    <<BV: subuniq (vars2nat (bvars e'))
                  (list_between n (nextbvar e') ++ vars2nat (bvars e))>> /\
    <<GT: is_good_thunk e -> is_good_thunk e'>> /\
    <<GE: good_exp e -> good_exp e'>>.

Definition exptf_sim f :=
  forall e e' n (n':nat) G lw phia phib
      (E: f G n e = Some (e', n'))
      (MX: (nextbvar e) <= n)
      (WF: wfexp G e)
      (UN: forall m (BIG: n <= m),
             uniq (m :: vars2nat (bvars e) ++ vars2nat (purge_lbl G)))
      (ST: stable lw),
  <<MAX: max (nextbvar e') n <= n'>> /\
  <<SIM: rtc (mm_psim lw G ren_id phia phib) e' e>>.

Record annT :=
  { _annT :> Type
  ; perform: _annT -> exptf
  ; perform_wf: forall a, exptf_wf (perform a)
  ; perform_sim: forall a, exptf_sim (perform a)
  }.
Arguments perform [a] _ _ _ _.

(** ** Sum lifting for annotation types. *)

Definition annT_sum (A B: annT) : annT.
  refine {| _annT := A + B
          ; perform ann := match ann with
                             | inl a => perform a
                             | inr b => perform b
                           end
         |}.
  - i; destruct a; eauto using perform_wf.
  - i; destruct a; eauto using perform_sim.
Defined.


(** ** List lifting for annotation types.

    Transformations will be [perform]ed from right to left until one
    fails, in which case its input is returned.

    (We may want to also have list liftings with a different
    semantics.)
*)


(** *** Raw definition. *)

Fixpoint perform_list' (A: annT) (revanns: list A) : exptf :=
  fun G n e => match revanns with
             | nil => Some (e, n)
             | a::revanns' => match perform a G n e with
                              | None => Some (e, n)
                              | Some (e', n')
                                => perform_list' A revanns' G n' e'
                              end
  end.

Definition perform_list (A: annT) (anns: list A) : exptf :=
  perform_list' A (rev anns).

Lemma list_between_not_included: forall n m m' (GEQ : m <= m'),
  list_mem eqdec m' (list_between n m) = false.
Proof.
  unfold list_between.
  ii; apply minus_le_compat_r with (p:=n) in GEQ.
  revert GEQ; generalize (m - n); clear; intro m; revert n.
  induction m; ins; desf; try apply IHm; omega.
Qed.

Lemma gt_ge_n_Sm: forall n m
    (GT: n > m),
  n >= S m.
Proof.
 ii; omega.
Qed.

Lemma ge_eq_or_gt: forall n m
    (GE: n >= m),
  n = m \/ n > m.
Proof.
  ii; omega.
Qed.

Lemma lt_minus_plus_eq: forall n m
    (LT: n < m),
  m - n + n = m.
Proof.
  ii; omega. 
Qed.
Hint Resolve lt_minus_plus_eq.

Lemma ge_minus_0: forall n m
    (GE: n >= m),
  m - n = 0.
Proof.
  ii; omega.
Qed.
Hint Resolve ge_minus_0.

Lemma max_r_p: forall n m p q
    (MX: max n m = p)
    (LE: q <= m),
  q <= p.
Proof.
  ii; apply NPeano.Nat.eq_le_incl in MX.
  apply Max.max_lub_r in MX.
  eauto using le_trans.
Qed.

Lemma list_between_in1: forall x n n'
    (IN: In x (list_between n n')),
  n <= x < n'.
Proof.
  unfold list_between; ii; replace n' with (n'-n+n).
  - revert IN; generalize (n' - n); clear; intro m; revert n x.
    induction m; ins; desf; try (specialize (IHm _ _ IN)); omega.
  - apply lt_minus_plus_eq; apply not_gt; ii; rewrite ge_minus_0 in IN; auto
    ; omega.
Qed.

Lemma list_between_in2: forall x n n'
    (H: n <= x < n'),
  In x (list_between n n').
Proof.
  unfold list_between; ii; replace n' with (n'-n+n) in H; try omega.
  revert H; generalize (n' - n); clear; intro m; revert n x.
  induction m; ins; desf; try omega. 
  destruct (eq_nat_dec n x); eauto.
  right; eapply IHm; omega.
Qed.

Lemma list_between_in x n n' :
  In x (list_between n n') <-> n <= x < n'.
Proof.
  split; eauto using list_between_in1, list_between_in2.
Qed.

Lemma list_between_sub: forall n n' n''
    (R: n <= n' <= n''),
  list_between n n' ++ list_between n' n'' = list_between n n''.
Proof.
  ii; unfold list_between; des.
  pattern n' at 2; replace n' with (n+(n'-n)); [|omega].
  replace (n''-n) with ((n''-n')+(n'-n)) by omega; rhs rewrite plus_comm.
  generalize (n''-n'); generalize (n'-n); clear; intros m1 m2; revert n.
  induction m1; ins; [do !rewrite plus_0_r; auto |].
  rewrite <-(IHm1 (S n)); try omega; rewrite plus_Snm_nSm; auto.
Qed.

Lemma seq_nein: forall n n' m (LEQ: n' <= n),
  ~ In n' (seq (S n) m).
Proof.
  intros n n' m; revert n' n.
  induction m; ins; desf; try omega.
  eapply IHm; try (exact FH)
  ; apply le_trans with (m:=n); auto.
Qed.

Lemma seq_uniq: forall n m,
  uniq(seq n m).
Proof.
  i; revert n.
  induction m; ins; desf; splits; try apply IHm.
  apply seq_nein; auto.
Qed.

Lemma list_between_uniq: forall n m,
  uniq(list_between n m).
Proof.
  unfold list_between; ii; apply seq_uniq.
Qed.

Lemma lb_nil: forall n n'
    (LE: n' <= n),
  list_between n n' = nil.
Proof.
  ii; unfold list_between; rewrite ge_minus_0; [|by omega]; vauto.
Qed.

Lemma list_between_subuniq: forall xs ys n m p
    (SP: subuniq xs (list_between n m ++ ys))
    (LE: n <= m <= p),
  subuniq xs (list_between n p ++ ys).
Proof.
  i. eapply subuniq_trans; eauto.
  apply subuniq_sublist; apply sublist_app; auto.
  rhs rewrite <-(@list_between_sub n m) by omega.
  apply sublist_r_app_l; auto.
Qed.

Lemma list_between_subuniq_lub: forall n n' m
    (LE: n <= n'),
  subuniq (list_between n' m) (list_between n m).
Proof.
  ii; destruct (le_dec n' m) as [LE'|NLE'].
  - rewrite <-(@list_between_sub n n' m) in UN; [| by omega].
    eapply (uniq_sublist UN); eauto using sublist_app, sublist_r_app_r.
  - apply not_le in NLE'; unfold list_between; rewrite ->ge_minus_0
    ; try (by omega).
    rewrite (app_nil_r _); eauto using uniq_sublist.
Qed.

Lemma list_between_subuniq_up: forall n m m'
    (LE: m <= m'),
  subuniq (list_between n m) (list_between n m').
Proof.
  ii; destruct (le_dec n m) as [LE'|NLE'].
  - rewrite <-(@list_between_sub n m m') in UN; [| by omega].
    eapply (uniq_sublist UN); eauto using sublist_app, sublist_r_app_r.
  - apply not_le in NLE'; unfold list_between; rewrite ->ge_minus_0
    ; try (by omega).
    rewrite (app_nil_r _); eauto using uniq_sublist.
Qed.

Lemma uniq_with_list_between: forall n n'' xs ys
    (UN: forall m (BIG: n <= m), uniq (m :: xs ++ ys))
    (NINC: n <= n''),
  forall m (BIG: n'' <= m),
    uniq (list_between n n'' ++ m :: xs ++ ys).
Proof.
  ii; apply uniq_app
  ; [ apply list_between_uniq
    | apply (uniq_sublist (UN _ (le_trans _ _ _ NINC BIG)))
      ; auto|].
  
  intros x INLB INR; apply list_between_in1 in INLB; des.
  specialize (UN _ INLB); simpls.
  des; by [omega | auto].
Qed.


Lemma uniq_with_list_between_2: forall n n'' xs ys
    (UN: forall m (BIG: n <= m), uniq (m :: xs ++ ys))
    (NINC: n <= n''),
  uniq (list_between n n'' ++ xs ++ ys).
Proof.
  ii; apply uniq_app
  ; [ apply list_between_uniq
    | apply (uniq_sublist (UN _ NINC))
      ; auto|].
  
  intros x INLB INR; apply list_between_in1 in INLB; des.
  specialize (UN _ INLB); simpls.
  des; by [omega | auto].
Qed.


Ltac solve_subuniq :=
  first [ eassumption
        | match goal with
            | [H: subuniq (?x) ?y |- subuniq (?x ++ _) (?y ++ _)]
              => apply subuniq_app; [ by assumption |]; solve_subuniq
            | [H: subuniq (?x) ?y |- subuniq (?x) (?z)]
              => eapply subuniq_trans; [solve_subuniq|]; []
                 ; solve [ auto | apply subuniq_perm; crush_perm]
            | |- subuniq (?x ++ _) ?y =>
                first [ is_evar y; apply subuniq_app; solve_subuniq
                      | eapply subuniq_trans; [solve_subuniq|]; []
                        ; solve [ auto | apply subuniq_perm; crush_perm]
                      ]
          end
        | apply subuniq_refl].

Ltac crush_subuniq_uniq :=
  eapply subuniq_uniq_app; match goal with
                           | |- subuniq _ _
                             => try (by solve_subuniq)
                           | |- _ => idtac
                           end.

Lemma uniq_vars: forall n n' n'' xs ys zs
    (UN: forall m (BIG: n <= m), uniq(m :: xs ++ zs))
    (BV: subuniq (ys) (list_between n n' ++ xs))
    (NINC: n <= n'')
    (LE: n' <= n''),
  forall m (BIG: n'' <= m), uniq(m :: ys ++ zs).
Proof.
  i.
  assert (SP: subuniq (ys) (list_between n n'' ++ xs)).
  {
    eapply subuniq_trans; [|apply subuniq_app]
    ; [| apply list_between_subuniq_up with (m:=n') |]
    ; [exact BV| |]; auto.
  }
  clear BV.
  hexploit uniq_with_list_between; eauto.
  intros UN2; rewrite cons_app; rewrite app_assoc.
  crush_subuniq_uniq.
  apply (uniq_perm UN2); crush_perm.
Qed.

Lemma uniq_vars2: forall n n' n'' xs xs' ys ys' zs
    (UN: forall m (BIG: n <= m), uniq (m :: (xs ++ xs') ++ zs))
    (BV1: subuniq (ys) (list_between n n' ++ xs))
    (BV2: subuniq (ys') (list_between n' n'' ++ xs'))
    (EQ: n <= n' <= n''),
  forall m (BIG: n'' <= m), uniq (m :: (ys ++ ys') ++ zs).
Proof.
  ii; apply uniq_vars with (n:=n)(n':=n'')(n'':=n'')(xs:=(xs++xs'))
  ; try by first [omega | assumption].
  rewrite <-(@list_between_sub n n'); [|by omega].
  eapply subuniq_trans; [apply subuniq_app|]
  ; [exact BV1 | exact BV2|].
  apply subuniq_perm; crush_perm.
Qed.

Lemma subuniq_nbvar: forall e ys n
    (SP: subuniq (vars2nat (bvars e)) (ys)),
  subuniq (vars2nat (bvars e)) (list_between n (nextbvar e) ++ ys).
Proof.
  i; rewrite ->nextbvar_equiv in *.
  destruct (le_lt_dec n (nextbvar' e)) as [LE|NLE].
  - eapply subuniq_trans; [exact SP|]; eauto using subuniq_sublist.
  - rewrite lb_nil; vauto; []; omega.
Qed.

Lemma subuniq_lb_weaken_r: forall xs ys n n' n''
    (SP: subuniq xs (list_between n n' ++ ys))
    (LE: n' <= n''),
  subuniq xs (list_between n n'' ++ ys).
Proof.
  i; hexploit (@list_between_subuniq_up n); eauto; i.
  eapply subuniq_trans; [exact SP|].
  apply subuniq_app; try exact X; auto.
Qed.

Lemma subuniq_lb_weaken_l: forall xs ys n n' n''
    (SP: subuniq xs (list_between n' n'' ++ ys))
    (LE: n <= n'),
  subuniq xs (list_between n n'' ++ ys).
Proof.
  i; hexploit (@list_between_subuniq_up n); eauto; i.
  eapply subuniq_trans; [exact SP|].
  apply subuniq_app; try exact X; auto.
  apply list_between_subuniq_lub; auto.
Qed.

Lemma subuniq_sng: forall xs xs' n m p
    (SP1: subuniq (xs) (list_between n m ++ xs'))
    (LE1: p <= n),
  subuniq (xs) (list_between n (max m p) ++ xs').
Proof.
  i; destruct (le_lt_dec n (max m p)) as [NLE|NNLE].
  - destruct (Max.max_dec m p) as [LH|RH].
    + rewrite ->LH in *; apply NPeano.Nat.max_l_iff in LH.
      eapply subuniq_trans; [exact SP1|]; auto.
    + rewrite ->RH in *; apply NPeano.Nat.max_r_iff in RH.
      rewrite ->lb_nil in *; try omega; ss.
  - dup NNLE; rewrite ->NPeano.Nat.max_lub_lt_iff in NNLE; des.
    rewrite ->lb_nil in *; try omega; ss. 
Qed.


Lemma subuniq_dbl: forall xs xs' ys ys' nxs nys n n'
    (SP1: subuniq (xs) (list_between n nxs ++ xs'))
    (SP2: subuniq (ys) (list_between n' nys ++ ys'))
    (LE1: nxs <= n')
    (LE2: n <= n'),
  subuniq (xs ++ ys) (list_between n (max nxs nys) ++ xs' ++ ys').
Proof.
  i; destruct (le_lt_dec n (max nxs nys)) as [NLE|NNLE].
  - destruct (Max.max_dec nxs nys) as [LH|RH].
    + rewrite ->LH in *; apply NPeano.Nat.max_l_iff in LH
      ; rewrite lb_nil in SP2; [|by omega]; ss.
      eapply subuniq_trans; [apply subuniq_app|]; [exact SP1| exact SP2 |]
      ; svbv; auto.
    + rewrite ->RH in *; apply NPeano.Nat.max_r_iff in RH.
      destruct (le_lt_dec n' nys) as [N'LE|N'NLE].
      * rewrite <-(@list_between_sub n n' _); [|by omega]; svbv.
        eapply subuniq_lb_weaken_r with (n'':=n') in SP1; [| by omega].
        eapply subuniq_trans; [apply subuniq_app|]; [exact SP1| exact SP2|].
        eapply subuniq_perm; crush_perm.
      * rewrite lb_nil in SP2; [| by omega]; ss.
        eapply subuniq_lb_weaken_r with (n'':=nys) in SP1
        ; [| by omega].
        eapply subuniq_trans; [apply subuniq_app|]; [exact SP1| exact SP2|].
        eapply subuniq_perm; crush_perm.
  - dup NNLE; rewrite ->NPeano.Nat.max_lub_lt_iff in NNLE; des.
    rewrite lb_nil; [|by omega].
    rewrite lb_nil in SP1; [|by omega]; ss.
    rewrite lb_nil in SP2; [|by omega]; ss.
    eapply subuniq_app; auto.
Qed.

Lemma subuniq_dbl': forall xs xs' ys ys' nxs nys n n' m p
    (SP1: subuniq (xs) (list_between n nxs ++ xs'))
    (SP2: subuniq (ys) (list_between n' nys ++ ys'))
    (LE1: nxs <= n')
    (LE2: n <= n')
    (LE3: m <= p <= n),
  subuniq (m :: xs ++ ys) (list_between n (max p (max nxs nys)) ++ m :: xs'
                                        ++ ys').
Proof.
  i; destruct (Max.max_dec p (max nxs nys)) as [LH|RH].
  - rewrite ->LH in *; apply NPeano.Nat.max_l_iff in LH.
    rewrite ->NPeano.Nat.max_lub_iff in LH; des.
    rewrite ->lb_nil in *; try omega; ss.
    rewrite cons_app; rhs3 rewrite cons_app; eauto using subuniq_app.
  - rewrite ->RH in *; hexploit (subuniq_dbl _ _ _ SP1 SP2); eauto.
    i; rewrite cons_app; rhs rewrite cons_app.
    eapply subuniq_trans; [apply subuniq_app |]; [auto|exact H|].
    eapply subuniq_perm; crush_perm.
Qed.

Lemma subuniq_bvar_bd: forall e xs n n'
    (SP: subuniq (vars2nat (bvars e)) (list_between n n' ++ xs))
    (UN: forall m (BIG: n <= m), uniq (m :: xs))
    (GE: n <= n') 
    (GT: nextbvar e < n'),
  subuniq (vars2nat (bvars e)) (list_between n (nextbvar e) ++ xs).
Proof.
  ii.
  assert (UN2: uniq(list_between n n' ++ xs)).
  {
    apply uniq_with_list_between_2 with (xs:=nil); ins.
  }
  apply uniq_app.
  - apply uniq_appD in UN0; desf.
  - apply (SP nil); ins.
  - ii; destruct (lt_dec x n).
    + exploit (SP (x :: nil)); try (by ins; desf).
      apply uniq_appD in UN0; des.
      specialize (UN3 _ INa); ins; splits; auto.
      ii; apply UN3.
      apply in_app_or in FH; des.
      * apply list_between_in1 in FH; omega.
      * apply in_or_app; right; auto.
    + apply nextvar_nin_exp in INb; ins.
      apply NNPP; ii.
      apply uniq_appD in UN0; des.
      apply UN3 in INa; ins.
      apply in_or_app; left; unfold nextbvar.
      apply list_between_in2; omega.
Qed.

Lemma subuniq_bvar_trp: forall t nxs n nys m p xs xs' ys ys' e
    (UN: forall m (BIG: n <= m), uniq (m :: t :: xs' ++ ys'))
    (SPX: subuniq xs (list_between n nxs ++ xs'))
    (SPY: subuniq ys (list_between m nys ++ ys'))
    (SPZ: subuniq (map varid (bvars e))
                  (list_between p (nextbvar e) ++ t :: xs ++ ys))
    (LT: t < n)
    (MEQX: max nxs n <= m)
    (MEQY: max nys m <= p),
 subuniq (map varid (bvars e))
         (list_between n (nextbvar e) ++ t :: xs' ++ ys').
Proof.
  intros.
  hexploit subuniq_dbl; first exact SPX; first exact SPY
  ; rewrite ->!NPeano.Nat.max_lub_iff in *; des; vauto.
  i; destruct (le_lt_dec (max nxs nys) (nextbvar e)).
  - destruct (le_lt_dec (nextbvar e) p) as [LEP|NLP].
    + rewrite ->lb_nil in SPZ; [|by omega]; s in SPZ.
      apply subuniq_lb_weaken_r with (n'':=(nextbvar e)) in H; auto.
      rewrite ->cons_app in *.
      eapply subuniq_trans; [eapply subuniq_trans|]; [exact SPZ| |]
      ; [apply subuniq_app|]; [auto | exact H|].
      eapply subuniq_perm; crush_perm.
    + apply subuniq_lb_weaken_r with (n'':=p) in H
      ; eauto using Max.max_lub, le_trans.
      rewrite <-(@list_between_sub n p _); splits; eauto using le_trans.
      rewrite ->cons_app in *.
      eapply subuniq_trans; [eapply subuniq_trans|]; [exact SPZ| |]
      ; [apply subuniq_app|]; [auto | apply subuniq_app|]
      ; [auto | exact H|].
      eapply subuniq_perm; crush_perm.
  - destruct (le_lt_dec n (max nxs nys)).
    + rewrite ->lb_nil in SPZ
      ; [| rewrite ->!NPeano.Nat.max_lt_iff in l; des; eauto using le_trans]
      ; s in SPZ.
      assert (SP: subuniq (map varid (bvars e))
                          (list_between n (max nxs nys) ++ t :: xs' ++ ys')).
      {
        rewrite ->cons_app in *.
        eapply subuniq_trans; [eapply subuniq_trans|]; [exact SPZ| |]
        ; [apply subuniq_app|]; [auto | exact H|].
        eapply subuniq_perm; crush_perm.
      }
      eapply subuniq_bvar_bd; first exact SP; auto.
    + dup l0; rewrite ->NPeano.Nat.max_lub_lt_iff in l0; des.
      rewrite ->lb_nil; rewrite ->lb_nil in SPZ,H; try omega; ss.
      rewrite ->cons_app in *.
      eapply subuniq_trans; [eapply subuniq_trans|]; [exact SPZ| |]
      ; [apply subuniq_app|]; [auto | exact H|].
      eapply subuniq_perm; crush_perm.
Qed.

(** *** Syntactic correctness. *)

Lemma perform_list_wf: forall A anns,
  exptf_wf (perform_list A anns).
Proof.
  i; unfold perform_list; generalize (rev anns); clear anns.
  induction l; ii; s in E; clarify
  ; [ by splits; auto using subuniq_sublist, NPeano.Nat.max_lub|].
  destruct (perform a G n e) eqn:E'
  ; try (by clarify; splits
         ; auto using subuniq_sublist, NPeano.Nat.max_lub)
  ; destruct p as (e'',n'').
  - exploit (perform_wf A); eauto; []; i; des.
    dup MAX; apply NPeano.Nat.max_lub_iff in MAX; des.
    exploit IHl; eauto using uniq_vars, NPeano.Nat.max_lub_r.
    i; des; dup MAX2; apply NPeano.Nat.max_lub_iff in MAX2; des
    ; esplits; eauto using Max.max_lub, le_trans.
    destruct (Max.max_dec (nextbvar e') n'') as [LH|RH].
    + apply NPeano.Nat.max_l_iff in LH.
      eapply subuniq_lb_weaken_r in BV; eauto.
      rewrite <-(@list_between_sub n n'' (nextbvar e')); auto.
      eapply subuniq_trans; [exact BV0 | eapply subuniq_trans].
      { eapply subuniq_app; [apply subuniq_refl | exact BV]. }
      { apply subuniq_perm; crush_perm. }
    + apply NPeano.Nat.max_r_iff in RH.
      rewrite ->lb_nil in BV0; s in BV0; auto.
      eapply subuniq_trans in BV; [| exact BV0].
      destruct (Max.max_dec (nextbvar e'') n) as [LH|RH'].
      {
        
        apply NPeano.Nat.max_l_iff in LH.
        destruct (le_lt_dec (nextbvar e'') (nextbvar e'))
        ; [by eapply subuniq_lb_weaken_r in BV; eauto|].
        eapply subuniq_bvar_bd; eauto.
        ii; specialize (UN _ BIG); eauto using uniq_sublist.
      }
      {
        apply NPeano.Nat.max_r_iff in RH'.
        rewrite ->lb_nil in BV; s in BV; auto.
        eapply subuniq_trans, subuniq_sublist; eauto.
      }
Qed.

(** *** Semantic correctness. *)

Lemma perform_list_sim: forall A anns,
  exptf_sim (perform_list A anns).
Proof.
  i; unfold perform_list; generalize (rev anns); clear anns.
  induction l; ii; s in E; clarify
  ; [by splits; eauto using NPeano.Nat.max_lub|].
  destruct (perform a G n e) eqn:E'
  ; try (by clarify; splits; auto using NPeano.Nat.max_lub)
  ; destruct p as (e'',n'').
  - exploit (perform_wf A); eauto; []; i; des.
    exploit (perform_sim A); eauto; []; i; des.
    apply NPeano.Nat.max_lub_iff in MAX; des.
    exploit IHl; eauto using NPeano.Nat.max_lub_r, uniq_vars; i; des.
    esplits.
    + ssbvar; splits; eauto using le_trans.
    + econs 3; eauto.
Qed.

(** *** Final definition. *)

Definition annT_list (A: annT) : annT :=
  {| _annT := list A
   ; perform := perform_list _
   ; perform_wf := perform_list_wf _
   ; perform_sim := perform_list_sim _
  |}.

(** ** Process all annotations of an expression. *)


(** *** Definition. *)


Definition transform_ {A: annT} : list var -> nat -> aexp A ->
                                 (nat * (exp * aexp A)) :=
  fix transform G n e :=
   let (e',n') :=                                
    match e with
      | IElet a ea x eb 
        => let bv := match x with
                     | inl bx => inl (inr bx)
                     | inr k => inr k
                     end in
           let '(n',(ea', _)) := transform G n ea in
           let '(n'',(eb', _)) := transform (bv :: G) n' eb in
           let e' := lang_mid.IElet ea' x eb' in
           withd (e',n'') (perform a G n'' e')

      | IEapp a x y k
        => let e' := lang_mid.IEapp x y k in
           withd (e',n) (perform a G n e')

      | IEinst a x k
        => let e' := lang_mid.IEinst x k in
           withd (e',n) (perform a G n e')

      | IEfix f x k ea
        => let G' := inr k :: inl (inr x) :: inl (inr f) :: purge G in
           let '(n',(ea',_)) := transform G' n ea in
           (lang_mid.IEfix f x k ea',n')
      | IEgen k ea
        => let '(n',(ea',_)) := transform (inr k :: purge G) n ea in
           (lang_mid.IEgen k ea',n')
      | IEcont x ea
        => let '(n',(ea',_)) := transform (inl (inr x) :: G) n ea in
           (lang_mid.IEcont x ea', n')
      | IEnew x y ea
        => let '(n',(ea',_)) := transform (inl (inr y) :: G) n ea in
           (lang_mid.IEnew x y ea', n')
      | IEread x y ea
        => let '(n',(ea',_)) := transform (inl (inr y) :: G) n ea in
           (lang_mid.IEread x y ea',n')
      | IEoutput x ea
        => let '(n',(ea',_)) := transform G n ea in
           (lang_mid.IEoutput x ea',n')
      | IEwrite x y ea
        => let '(n',(ea',_)) := transform G n ea in
           (lang_mid.IEwrite x y ea',n')
      | IEinput x ea
        => let '(n',(ea',_)) := transform (inl (inr x) :: G) n ea in
           (lang_mid.IEinput x ea',n')
      | IEifnz x ea eb
        => let '(n',(ea',_)) := transform G n ea in
           let '(n'',(eb',_)) := transform G n' eb in
           (lang_mid.IEifnz x ea' eb',n'')

      | IEcase x y ea eb
        => let '(n',(ea',_)) := transform (inl (inr y) :: G) n ea in
           let '(n'',(eb',_)) := transform (inl (inr y) :: G) n' eb in
           (lang_mid.IEcase x y ea' eb',n'')

      | IEunit
        => (lang_mid.IEunit,n)
      | IEnat n0
        => (lang_mid.IEnat n0,n)
      | IEpair x y
        => (lang_mid.IEpair x y,n)
      | IEret k x
        => (lang_mid.IEret k x,n)
      | IEinl x
        => (lang_mid.IEinl x,n)
      | IEinr x
        => (lang_mid.IEinr x,n)
      | IEfst x
        => (lang_mid.IEfst x,n)
      | IEsnd x
        => (lang_mid.IEsnd x,n)
      | IEbop b xn yn
        => (lang_mid.IEbop b xn yn,n)
      | IErefeq x y
        => (lang_mid.IErefeq x y,n)
    end in (n',(e',e)).
Hint Unfold transform_.

Definition transform {A: annT} G n (ae : aexp A) : exp * nat :=
  let '(n,(e,_)) := transform_ G n ae in (e,n).
Hint Unfold transform.


Ltac trauto :=
  match goal with
  | [H: context[transform_ ?G ?n ?ae] |- context[transform ?G' ?n' ?ae']]
    => let n' := fresh in let p := fresh in
       by unify G G'; unify n n'; unify ae ae'; unfold transform
       ; destruct (transform_ G n ae) as (n',p); destruct p; clarify
  | _ => idtac
  end.

Ltac tranify ae :=
  match goal with
  | [H: context[transform_ ?G ?n ae] |- context[transform ?G' ?n' ?ae']]
    => let n' := fresh in let p := fresh in
       by unify G G'; unify n n'; unify ae ae'; unfold transform
       ; destruct (transform_ G n ae) as (n',p); destruct p; clarify
  | _ => idtac
  end.

Fixpoint transmodmap_ {A: annT}
                      (f: list var -> nat -> exp -> (nat * (exp * aexp A)))
                      (G: list var) (n : nat) (m : module) :
                       module * list (aexp A) :=
  match m with
  | nil => (nil,nil)
  | (lbl,e) :: m' => let '(n',(e',ae)) := f G n e in
                     let (tm,l) := transmodmap_ f (G ++ [inl (inl lbl)]) n' m'
                     in ((lbl,e') :: tm, ae :: l)
  end.

Definition remove_annT {A : annT} (trp: nat * (exp * aexp A)) : exp * nat :=
  let '(n,(e,_)) := trp in (e,n).

Fixpoint transmodmap (f: list var -> nat -> exp -> exp * nat)
                     (G: list var) (n : nat) (m : module) : module :=
  match m with
  | nil => nil
  | (lbl,e) :: m' => let (e',n') := f G n e in
                       (lbl,e') :: transmodmap f (G ++ [inl (inl lbl)]) n' m'
  end.

Fixpoint transfmap (fs: list (list var -> nat -> exp -> exp * nat))
                   (G: list var) (n : nat) (m : module) : module :=
  match fs,m with
  | nil,_ => m
  | f :: fs',(lbl,e) :: m'
    => let (e',n') := f G n e in
       (lbl,e') :: transfmap fs' (G ++ [inl (inl lbl)]) n' m'
  | _,nil => nil
  end.

Lemma le_trans' n m p:
  n <= m -> m <= p -> n <= p.
Proof.
  eauto using le_trans.
Qed.


Lemma transmodmap_wfmod: forall f FI FO m nv
    (WF: mid_wfmod FI m FO)
    (GE: nextvar m <= nv)
    (UV: forall n' (BIG: (nextvar m) <= n'),
           uniq (n' :: vars2nat (bvars_mod m)))
    (FWF: forall e e' F xs n n'
                 (GT: is_good_thunk e)
                 (T: f (map (inl<*>inl) F) n e = (e',n'))
                 (MX: nextbvar e <= n)
                 (WF: wfatom (map (inl<*>inl) F) e)
                 (UN: forall m (BIG: n <= m),
                        uniq (m :: vars2nat (bvars e) ++ vars2nat xs)),
            <<NINC: max (nextbvar e') n <= n'>> /\
            <<WF: wfatom (map (inl<*>inl) F) e'>> /\
            <<BV: subuniq (vars2nat (bvars e'))
                          (list_between n (nextbvar e') ++
                                        vars2nat (bvars e))>> /\
            <<GT: is_good_thunk e'>>),
  <<WF: mid_wfmod FI (transmodmap f (map (inl<*>inl) FI) nv m) FO>> /\
  <<UV: uniqvar (transmodmap f (map (inl<*>inl) FI) nv m)>>.
Proof.
  ii.
  match goal with |- ?P =>
    cut (P /\ subuniq (vars2nat (bvars_mod
                                   (transmodmap f (map (inl<*>inl) FI) nv m)))
                      (list_between nv
                                 (nextvar
                                 (transmodmap f (map (inl<*>inl) FI) nv m)) ++
                       vars2nat (bvars_mod m)))
  end; [by i; des; auto|].
  revert UV; depgen nv; induction WF
  ; [splits; vauto; ii; ss; apply uniq_appD in UN; des
     ; rewrite (app_nil_r _)|]; vauto; i.
  unfold transmodmap; fold transmodmap; destruct (f _ _ _) as (e',n') eqn:TR.
  hexploit FWF; eauto.
  - rewrite ->nextvar_rewrite in GE; apply Max.max_lub_l in GE; vauto.
  - ii; specialize (UV _ (le_trans _ _ _ GE BIG)); unfold vars2nat in UV
    ; simpl bvars_mod in UV; rewrite ->!map_app in UV; eauto.
  - i; des.
    hexploit (IHWF n'); eauto.
    + rewrite nextvar_rewrite in GE; apply Max.max_lub_r in GE.
      apply Max.max_lub_r in NINC; eauto using le_trans.
    + ii; apply what; auto.
      
      unfold vars2nat in *; simpl bvars_mod in UV
      ; rewrite ->!map_app in *; specialize (UV _ GE)
      ; ss; des; apply uniq_appD in UV0; des; auto.
    + i; des; rewrite !map_app in WF1; simpl in WF1; splits; eauto.
      { 
        r; s; unfold vars2nat in UV; simpl bvars_mod in UV
        ; eapply uniq_mapD with (f:=varid); rewrite ->!map_app in *.
        eapply subuniq_uniq_app; first last; [exact H0 | exact BV |].
        assert (FOO: forall n p n'' (GENV: nv <= n) (BIG: p <= n')
                            (LE: n <= n'),
                     uniq (list_between n p ++ vars2nat (bvars e) ++
                           list_between n' n'' ++ vars2nat (bvars_mod m))).
        {
          i; destruct (le_dec n p) as [LEP|NLEP]
          ; destruct (le_dec n' n'') as [LE'|NLE'].
          - hexploit (@uniq_with_list_between_2 n n''); try (by omega)
            ; try (by ii
                   ; specialize (UV _ (le_trans' GE (le_trans' GENV BIG0)))
                   ; exact UV)
            ; eauto using le_trans.
            intros LBH; rewrite <-(@list_between_sub _ n') in LBH by omega.
            eapply subuniq_uniq_app
            ; [| apply list_between_subuniq_up with (m':=n'); try omega 
               | apply subuniq_refl].
            eapply (uniq_perm LBH); crush_perm.
          - unfold list_between at 2; simpl seq.
            apply not_le in NLE'; rewrite ge_minus_0; [|by omega]; s; ii.
            hexploit (@uniq_with_list_between_2 n p); try (by omega)
            ; try (by ii
                   ; specialize (UV _ (le_trans' GE (le_trans' GENV BIG0)))
                   ; exact UV)
            ; eauto.
          - unfold list_between at 1; simpl seq.
            apply not_le in NLEP; rewrite ge_minus_0; [|by omega]; s; ii.
            hexploit (@uniq_with_list_between_2 n' n''); try (by omega)
            ; try (by ii
                   ; specialize (UV _ (le_trans' GE
                                             (le_trans' GENV
                                                        (le_trans' LE BIG0))))
                   ; exact UV)
            ; eauto.
            intros LBH; eapply (uniq_perm LBH); crush_perm.
          - unfold list_between; simpl seq.
            apply not_le in NLEP; apply not_le in NLE'
            ; rewrite !ge_minus_0; try (by omega); s; ii.
            specialize (UV _ GE); eauto using uniq_sublist.
        }
        destruct (Max.max_dec nv (nextbvar e')); rewrite -app_assoc
        ; apply FOO; apply NPeano.Nat.max_lub_iff in NINC; des; vauto.
      }
      { 
        
        unfold vars2nat in *; simpl bvars_mod in *; rewrite ->!map_app in *.
        eapply subuniq_trans; [apply subuniq_app|]; [exact BV | exact H0|].
        rewrite !nextvar_rewrite; unfold nextvar; subst; rewrite -app_assoc.
        destruct (Max.max_dec (nextbvar e') nv).
        {
          destruct (Max.max_dec (nextbvar e')
                                (S (maxvar (transmodmap f
                                   (map (inl <*> inl) FI ++ [inl (inl f0)])
                                   n' m))))
          ; rewrite e1; rewrite ->e0 in *.
          - unfold list_between at 2.
            rewrite ge_minus_0; vauto; []; apply NPeano.Nat.max_l_iff in e1
            ; eauto using le_trans.
          - rhs3 rewrite <-(@list_between_sub nv (S (maxbvar e')) _)
            ; try (by splits
                   ; [apply NPeano.Nat.max_l_iff | apply NPeano.Nat.max_r_iff]
                   ; auto).
            rewrite -app_assoc; apply subuniq_app; auto.
            eapply subuniq_trans; [apply subuniq_app|]
            ; [apply subuniq_refl | apply subuniq_app|]
            ; try (by eapply list_between_subuniq_lub with (n:=S(maxbvar e'))
                   ; auto)
            ; [apply subuniq_refl|].
            apply subuniq_perm; crush_perm.
        }
        {
          destruct (Max.max_dec (nextbvar e')
                                (S (maxvar (transmodmap f
                                   (map (inl <*> inl) FI ++ [inl (inl f0)])
                                   n' m))))
          ; rewrite e1; rewrite ->e0 in *.
          - unfold list_between at 1; rewrite ge_minus_0; s
            ; [|by apply NPeano.Nat.max_r_iff; vauto].
            unfold list_between at 2; rewrite ge_minus_0; s
            ; [|by apply NPeano.Nat.max_r_iff; vauto].
            unfold list_between; rewrite ge_minus_0; s; vauto
            ; try (by rewrite -> NPeano.Nat.max_r_iff in e0
                   ; rewrite -> NPeano.Nat.max_l_iff in e1; omega).
          - unfold list_between at 1; rewrite ge_minus_0; s
            ; [|by apply NPeano.Nat.max_r_iff; vauto].
            eapply subuniq_trans; [apply subuniq_app|]
            ; [apply subuniq_refl | apply subuniq_app|]
            ; try (by eapply list_between_subuniq_lub with (n:=nv); auto)
            ; [apply subuniq_refl | apply subuniq_perm; crush_perm].
        }
      }
Qed.

(** Special case for contification second phase. *)
Lemma transfmap_wfmod: forall fs FI FO m nv
    (WF: mid_wfmod FI m FO)
    (GE: nextvar m <= nv)
    (UV: forall n' (BIG: (nextvar m) <= n'),
           uniq (n' :: vars2nat (bvars_mod m)))
    (FWF: forall e e' f F xs n n'
                 (IN: In f fs)
                 (GT: is_good_thunk e)
                 (T: f (map (inl<*>inl) F) n e = (e',n'))
                 (MX: nextbvar e <= n)
                 (WF: wfatom (map (inl<*>inl) F) e)
                 (UN: forall m (BIG: n <= m),
                        uniq (m :: vars2nat (bvars e) ++ vars2nat xs)),
            <<NINC: max (nextbvar e') n <= n'>> /\
            <<WF: wfatom (map (inl<*>inl) F) e'>> /\
            <<BV: subuniq (vars2nat (bvars e'))
                          (list_between n (nextbvar e') ++
                                        vars2nat (bvars e))>> /\
            <<GT: is_good_thunk e'>>),
  <<WF: mid_wfmod FI (transfmap fs (map (inl<*>inl) FI) nv m) FO>> /\
  <<UV: uniqvar (transfmap fs (map (inl<*>inl) FI) nv m)>>.
Proof.
  ii.
  match goal with |- ?P =>
    cut (P /\ subuniq (vars2nat (bvars_mod
                                  (transfmap fs (map (inl<*>inl) FI) nv m)))
                      (list_between nv
                                (nextvar
                                (transfmap fs (map (inl<*>inl) FI) nv m)) ++
                       vars2nat (bvars_mod m)))
  end; [by i; des; auto|].
  revert UV; revert GE; revert FWF; depgen nv; depgen fs
  ; induction WF; destruct fs as [|f' fs']
  ; try (by splits; vauto; ii; ss; apply uniq_appD in UN; des
         ; rewrite (app_nil_r _); vauto); i.
  {
    s; splits; vauto.
    - unfold uniqvar; s in UV; s; specialize (UV _ GE); des
      ; unfold vars2nat in UV0; eapply uniq_mapD with (f:=varid); eauto.
    - rewrite ->lb_nil; auto.
  }
  unfold transfmap; fold transfmap; destruct (f' _ _ _) as (e',n') eqn:TR.
  hexploit FWF; eauto; vauto.
  - rewrite ->nextvar_rewrite in GE; apply Max.max_lub_l in GE; vauto.
  - ii; specialize (UV _ (le_trans _ _ _ GE BIG)); unfold vars2nat in UV
    ; simpl bvars_mod in UV; rewrite ->!map_app in UV; eauto.
  - i; des.
    exploit (IHWF fs' n'); try instantiate (1 := fs'); eauto.
    + i; eapply FWF; eauto; []; by s; right; auto.
    + rewrite nextvar_rewrite in GE; apply Max.max_lub_r in GE.
      apply Max.max_lub_r in NINC; eauto using le_trans.
    + ii; apply what; auto.
      
      unfold vars2nat in *; simpl bvars_mod in UV
      ; rewrite ->!map_app in *; specialize (UV _ GE)
      ; ss; des; apply uniq_appD in UV0; des; auto.
    + i; des; rewrite !map_app in WF1; simpl in WF1; splits; eauto.
      { 
        r; s; unfold vars2nat in UV; simpl bvars_mod in UV
        ; eapply uniq_mapD with (f:=varid); rewrite ->!map_app in *.
        eapply subuniq_uniq_app; first last; [exact x1 | exact BV |].
        assert (FOO: forall n p n'' (GENV: nv <= n) (BIG: p <= n')
                            (LE: n <= n'),
                     uniq (list_between n p ++ vars2nat (bvars e) ++
                           list_between n' n'' ++ vars2nat (bvars_mod m))).
        {
          i; destruct (le_dec n p) as [LEP|NLEP]
          ; destruct (le_dec n' n'') as [LE'|NLE'].
          - hexploit (@uniq_with_list_between_2 n n''); try (by omega)
            ; try (by ii
                   ; specialize (UV _ (le_trans' GE (le_trans' GENV BIG0)))
                   ; exact UV)
            ; eauto using le_trans.
            intros LBH; rewrite <-(@list_between_sub _ n') in LBH by omega.
            eapply subuniq_uniq_app
            ; [| apply list_between_subuniq_up with (m':=n'); try omega 
               | apply subuniq_refl].
            eapply (uniq_perm LBH); crush_perm.
          - unfold list_between at 2; simpl seq.
            apply not_le in NLE'; rewrite ge_minus_0; [|by omega]; s; ii.
            hexploit (@uniq_with_list_between_2 n p); try (by omega)
            ; try (by ii
                   ; specialize (UV _ (le_trans' GE (le_trans' GENV BIG0)))
                   ; exact UV)
            ; eauto.
          - unfold list_between at 1; simpl seq.
            apply not_le in NLEP; rewrite ge_minus_0; [|by omega]; s; ii.
            hexploit (@uniq_with_list_between_2 n' n''); try (by omega)
            ; try (by ii
                   ; specialize (UV _ (le_trans' GE
                                             (le_trans' GENV
                                                        (le_trans' LE BIG0))))
                   ; exact UV)
            ; eauto.
            intros LBH; eapply (uniq_perm LBH); crush_perm.
          - unfold list_between; simpl seq.
            apply not_le in NLEP; apply not_le in NLE'
            ; rewrite !ge_minus_0; try (by omega); s; ii.
            specialize (UV _ GE); eauto using uniq_sublist.
        }
        destruct (Max.max_dec nv (nextbvar e')); rewrite -app_assoc
        ; apply FOO; apply NPeano.Nat.max_lub_iff in NINC; des; vauto.
      }
      { 
        
        unfold vars2nat in *; simpl bvars_mod in *; rewrite ->!map_app in *.
        eapply subuniq_trans; [apply subuniq_app|]; [exact BV | exact x1|].
        rewrite !nextvar_rewrite; unfold nextvar; subst; rewrite -app_assoc.
        destruct (Max.max_dec (nextbvar e') nv).
        {
          destruct (Max.max_dec (nextbvar e')
                                (S (maxvar (transfmap fs'
                                   (map (inl <*> inl) FI ++ [inl (inl f)])
                                   n' m))))
          ; rewrite e1; rewrite ->e0 in *.
          - unfold list_between at 2.
            rewrite ge_minus_0; vauto; []; apply NPeano.Nat.max_l_iff in e1
            ; eauto using le_trans.
          - rhs3 rewrite <-(@list_between_sub nv (S (maxbvar e')) _)
            ; try (by splits
                   ; [apply NPeano.Nat.max_l_iff | apply NPeano.Nat.max_r_iff]
                   ; auto).
            rewrite -app_assoc; apply subuniq_app; auto.
            eapply subuniq_trans; [apply subuniq_app|]
            ; [apply subuniq_refl | apply subuniq_app|]
            ; try (by eapply list_between_subuniq_lub with (n:=S(maxbvar e'))
                   ; auto)
            ; [apply subuniq_refl|].
            apply subuniq_perm; crush_perm.
        }
        {
          destruct (Max.max_dec (nextbvar e')
                                (S (maxvar (transfmap fs'
                                   (map (inl <*> inl) FI ++ [inl (inl f)])
                                   n' m))))
          ; rewrite e1; rewrite ->e0 in *.
          - unfold list_between at 1; rewrite ge_minus_0; s
            ; [|by apply NPeano.Nat.max_r_iff; vauto].
            unfold list_between at 2; rewrite ge_minus_0; s
            ; [|by apply NPeano.Nat.max_r_iff; vauto].
            unfold list_between; rewrite ge_minus_0; s; vauto
            ; try (by rewrite -> NPeano.Nat.max_r_iff in e0
                   ; rewrite -> NPeano.Nat.max_l_iff in e1; omega).
          - unfold list_between at 1; rewrite ge_minus_0; s
            ; [|by apply NPeano.Nat.max_r_iff; vauto].
            eapply subuniq_trans; [apply subuniq_app|]
            ; [apply subuniq_refl | apply subuniq_app|]
            ; try (by eapply list_between_subuniq_lub with (n:=nv); auto)
            ; [apply subuniq_refl | apply subuniq_perm; crush_perm].
        }
      }
Qed.

Lemma transmodmap_fst_equiv: forall A f G n m,
    transmodmap (fun G n => @remove_annT A <*> f G n) G n m =
    fst (transmodmap_ f G n m).
Proof.
  ii; revert n G; induction m as [|(lbl,e) m'].
  - unfold transmodmap_,compose,transmodmap; vauto.
  - simpl transmodmap; unfold transmodmap_; fold (@transmodmap_ A).
    ii; destruct (f G n e) eqn:TR; destruct p as (e',ae).
    destruct (remove_annT _) eqn:RA; unfold remove_annT in RA; clarify.
    rewrite (IHm' n1 (G ++ [inl (inl lbl)])).
    destruct (transmodmap_ _ _ _ _) eqn:TM; simpl; auto.
Qed.

Lemma transmodmap_wfmod_: forall (A: annT) f FI FO m m' nv
                                 (anns: list (aexp A))
    (WF: mid_wfmod FI m FO)
    (GE: nextvar m <= nv)
    (UV: forall n' (BIG: (nextvar m) <= n'),
           uniq (n' :: vars2nat (bvars_mod m)))
    (FWF: forall e e' F xs n n'
                 (GT: is_good_thunk e)
                 (T: (remove_annT <*> f (map (inl<*>inl) F) n) e = (e',n'))
                 (MX: nextbvar e <= n)
                 (WF: wfatom (map (inl<*>inl) F) e)
                 (UN: forall m (BIG: n <= m),
                        uniq (m :: vars2nat (bvars e) ++ vars2nat xs)),
            <<NINC: max (nextbvar e') n <= n'>> /\
            <<WF: wfatom (map (inl<*>inl) F) e'>> /\
            <<BV: subuniq (vars2nat (bvars e'))
                          (list_between n (nextbvar e') ++
                                        vars2nat (bvars e))>> /\
            <<GT: is_good_thunk e'>>)
    (TR: transmodmap_ f (map (inl <*> inl) FI) nv m = (m',anns)),
  <<WF: mid_wfmod FI m' FO>> /\  <<UV: uniqvar m'>>.
Proof.
  ii; replace m' with (fst (transmodmap_ f (map (inl<*>inl) FI) nv m))
              by (rewrite TR; auto).
  rewrite <-transmodmap_fst_equiv.
  eapply transmodmap_wfmod; eauto.
Qed.

Definition transform_mod {A: annT} (analyze: exp -> aexp A)
                         : list var -> nat -> module -> module :=
  transmodmap (fun G n => @transform A G n <*> analyze).

Definition transform_mod_ {A: annT} (analyze: exp -> aexp A)
                          : list var -> nat -> module ->
                            module * list (aexp A) :=
  transmodmap_ (fun G n => @transform_ A G n <*> analyze).

(** *** Syntactic correctness. *)

Ltac rnbv :=
  try rewrite ->nextbvar_equiv in *.


Ltac mxbv_auto :=
  ssbvar; splits; eauto using le_trans.


Ltac crush_mxsp :=
 try
  match goal with
  | |- context[max ?a ?b] =>
      let H := fresh in
      destruct (Max.max_dec a b) as [H|H]; rewrite ->H in *
      ; match goal with
        | [MX: a <= ?n, LE: ?n <= a |- _] =>
          rewrite ->(le_antisym _ _ MX LE) in *
          ; apply NPeano.Nat.eq_le_incl in H
          ; rewrite ->!NPeano.Nat.max_lub_iff in H; des
          ; unfold list_between in *
          ; try (rewrite ->ge_minus_0 in *; first [omega | idtac]); ss
        | _ => idtac
        end; crush_mxsp
  end.

Ltac des_nbvar :=
  match goal with
  | |- context[list_between ?n ?n'] =>
      let LE := fresh "LE" in
      let NLE := fresh "NLE" in
      destruct (le_lt_dec n n')
               as [LE|NLE]
  end.

Lemma transform_wfatom_: forall (A: annT) e e' n n' (ae: aexp A) G
    (WF: wfatom G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: vars2nat (bvars e) ++ vars2nat (purge_lbl G)))
    (T: transform G n ae = (e',n'))
    (MX: (nextbvar (aerase ae)) <= n)
    (E: aerase ae = e)
    (IH: forall e', size e' < size e ->
                    forall n n' e'' ae' G, wfexp G e' ->
                    (nextbvar' (aerase ae')) <= n ->
           @aerase A ae' = e' ->
           transform G n ae' = (e'',n') ->
           (forall m (BIG: n <= m),
              uniq (m :: vars2nat (bvars e') ++
                      vars2nat (purge_lbl G))) ->
           <<NINC: max (nextbvar e'') n <= n'>> /\
           <<WF: wfexp G e''>> /\
           <<BV: subuniq (vars2nat (bvars e''))
                   (list_between n (nextbvar e'') ++ vars2nat (bvars e'))>> /\
           <<GT: good_exp e' -> good_exp e''>>),
  <<NINC: max (nextbvar e') n <= n'>> /\
  <<WF: wfatom G e'>> /\
  <<BV: subuniq (vars2nat (bvars e'))
                (list_between n (nextbvar e') ++ vars2nat (bvars e))>> /\
  <<GT: is_good_thunk e -> is_good_thunk e'>>.
Proof.
  i; destruct ae; try (by splits; unfold transform, transform_ in T; clarify
                       ; vauto; auto using subuniq_sublist,
                                NPeano.Nat.max_lub).
  - s in E; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; s in MX; ssbvar
    ; hexploit (IH (aerase ae)); vauto; trauto.
    { 
      
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG)
      ; rewrite <-!app_comm_cons in UN.
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
      - rewrite ->!app_comm_cons; apply sublist_app
        ; [auto | do !apply sublist_cons_keep]; []
        ; rewrite purge_purge_lbl; apply vars2nat_purge_sublist.
      - rewrite <-!app_comm_cons; crush_perm.
    }
    rnbv; i; des; splits; vauto; try (by mxbv_auto); repeat econs 2; eauto.
    + svbv; ssbvar; eapply subuniq_sng; eauto using NPeano.Nat.max_lub
      ; by rewrite cons_app3; rhs3 rewrite cons_app3
      ; eapply subuniq_trans; [apply subuniq_app|]
      ; [apply subuniq_refl | exact BV|]
      ; apply subuniq_perm; crush_perm.
    + s; intros; apply H; invc FH; rewrite -H4 in GT.
  - s in E; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; s in MX; ssbvar
    ; hexploit (IH (aerase ae)); vauto; trauto.
    { 
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG)
      ; rewrite <-!app_comm_cons in UN.
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
      - rewrite ->!app_comm_cons; apply sublist_app
        ; [auto | do !apply sublist_cons_keep]; []
        ; rewrite purge_purge_lbl; apply vars2nat_purge_sublist.
      - rewrite <-!app_comm_cons; crush_perm.
    }
    rnbv
    ; i; des; splits; vauto; try (by mxbv_auto); repeat econs 2; eauto.
    + svbv; ssbvar; eapply subuniq_sng; eauto using NPeano.Nat.max_lub
      ; by rewrite cons_app; rhs3 rewrite cons_app
      ; eapply subuniq_trans; [apply subuniq_app|]
      ; [apply subuniq_refl | exact BV|]
      ; apply subuniq_perm; crush_perm.
    + s; intros; apply H; invc FH; rewrite -H4 in GT.
  - s in E; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; s in MX
    ; rewrite ->!NPeano.Nat.max_lub_iff in MX; des
    ; hexploit (IH (aerase ae)); vauto; trauto.
    { 
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG)
      ; rewrite <-!app_comm_cons in UN.
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
      - rewrite ->!app_comm_cons; apply sublist_app
        ; [auto | do !apply sublist_cons_keep]; []
        ; apply vars2nat_sublist; apply sublist_refl.
      - rewrite <-!app_comm_cons; crush_perm.
    }
    rnbv
    ; i; des; splits; vauto; try (by mxbv_auto); repeat econs 2; eauto.
    + svbv; ssbvar; eapply subuniq_sng; eauto using NPeano.Nat.max_lub
      ; by rewrite cons_app; rhs3 rewrite cons_app
      ; eapply subuniq_trans; [apply subuniq_app|]
      ; [apply subuniq_refl | exact BV|]
      ; apply subuniq_perm; crush_perm.
Qed.

Lemma transform_wfexp: forall (A: annT) e e' n n' (ae: aexp A) G
    (E: aerase ae = e)
    (T: transform G n ae = (e',n'))
    (MX: nextbvar' e <= n)
    (WF: wfexp G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: vars2nat(bvars e) ++ vars2nat (purge_lbl G))),
  <<NINC: max (nextbvar' e') n <= n'>> /\
  <<WF: wfexp G e'>> /\
  <<BV: subuniq (vars2nat (bvars e'))
                (list_between n (nextbvar' e') ++ vars2nat (bvars e))>> /\
  <<GE: good_exp e -> good_exp e'>>.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  i; destruct ae; vauto; try (by s in WF; des; s; splits
                              ; unfold transform, transform_ in T; clarify
                              ; eauto using subuniq_sublist,
                                NPeano.Nat.max_lub).
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ _) eqn:TR; destruct p
    ; s in MX; rewrite ->!NPeano.Nat.max_lub_iff in MX; des; clarify.
    hexploit (IH (aerase ae)); trauto; vauto.
    { 
      
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG)
      ; rewrite <-!app_comm_cons in UN.
      eapply (uniq_perm UN); crush_perm.
    }
    i; des; splits; try (by mxbv_auto); vauto.
    + svbv; ssbvar; eapply subuniq_sng; eauto using NPeano.Nat.max_lub
      ; by rewrite cons_app; rhs3 rewrite cons_app
      ; eapply subuniq_trans; [apply subuniq_app|]
      ; [apply subuniq_refl | exact BV|]
      ; apply subuniq_perm; crush_perm.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ _) eqn:TR; destruct p
    ; s in MX; rewrite ->!NPeano.Nat.max_lub_iff in MX; des; clarify.
    hexploit (IH (aerase ae)); trauto; vauto.
    { 
      
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG)
      ; rewrite <-!app_comm_cons in UN.
      eapply (uniq_perm UN); crush_perm.
    }
    i; des; splits; try (by mxbv_auto); vauto.
    + svbv; ssbvar; eapply subuniq_sng; eauto using NPeano.Nat.max_lub
      ; by rewrite cons_app; rhs3 rewrite cons_app
      ; eapply subuniq_trans; [apply subuniq_app|]
      ; [apply subuniq_refl | exact BV|]
      ; apply subuniq_perm; crush_perm.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ _) eqn:TR; destruct p; s in MX.
    hexploit (IH (aerase ae)); trauto; vauto.
    i; des; splits; vauto.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ _) eqn:TR; destruct p
    ; s in MX; rewrite ->!NPeano.Nat.max_lub_iff in MX; des; clarify.
    hexploit (IH (aerase ae)); trauto; vauto.
    { 
      
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG)
      ; rewrite <-!app_comm_cons in UN.
      eapply (uniq_perm UN); crush_perm.
    }
    i; des; splits; try (by mxbv_auto); vauto.
    + svbv; ssbvar; eapply subuniq_sng; eauto using NPeano.Nat.max_lub
      ; by rewrite cons_app; rhs3 rewrite cons_app
      ; eapply subuniq_trans; [apply subuniq_app|]
      ; [apply subuniq_refl | exact BV|]
      ; apply subuniq_perm; crush_perm.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ _) eqn:TR; destruct p
    ; ss; clarify.
    hexploit (IH (aerase ae)); trauto; vauto.
    i; des; splits; vauto.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ ae1) eqn:TR1; destruct p
    ; destruct (transform_ _ _ ae2) eqn:TR2; destruct p
    ; s in MX; rewrite ->!NPeano.Nat.max_lub_iff in MX; des; clarify.
    hexploit (IH (aerase ae1)); first (by rr; s; omega)
    ; try (by trauto; vauto).
    {
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG).
      eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
    }
    i; des; rewrite ->!NPeano.Nat.max_lub_iff in *; des; clarify.
    hexploit (IH (aerase ae2)); first (by rr; s; omega)
    ; try (by trauto; vauto); eauto using le_trans.
    {
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ (le_trans _ _ _ NINC0 BIG)).
      eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
    }
    i; des; rewrite ->!NPeano.Nat.max_lub_iff in *; des; clarify
    ; splits; s; vauto; eauto using le_trans, NPeano.Nat.max_lub.
    svbv; eapply subuniq_dbl; eauto.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ ae1) eqn:TR1; destruct p
    ; destruct (transform_ _ _ ae2) eqn:TR2; destruct p
    ; s in MX; rewrite ->!NPeano.Nat.max_lub_iff in MX; des; clarify.
    hexploit (IH (aerase ae1)); first (by rr; s; omega)
    ; try (by trauto; vauto).
    {
      simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
      ; rewrite -app_comm_cons in UN; rewrite !map_app in UN
      ; simpl purge_lbl; simpl vars2nat; ii; specialize (UN _ BIG).
      eapply uniq_perm
      ; last by eapply Permutation_cons; eapply Permutation_cons_app; eauto.
      eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
    }
    i; des; rewrite ->!NPeano.Nat.max_lub_iff in *; des; clarify.
    hexploit (IH (aerase ae2)); first (by rr; s; omega)
    ; try (by trauto; vauto); eauto using le_trans.
    {
      simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
      ; rewrite -app_comm_cons in UN; rewrite !map_app in UN
      ; simpl purge_lbl; simpl vars2nat; ii
      ; specialize (UN _ (le_trans _ _ _ NINC0 BIG)).
      eapply uniq_perm
      ; last by eapply Permutation_cons; eapply Permutation_cons_app; eauto.
      eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
    }
    i; des; clarify; splits; s; rewrite ->!NPeano.Nat.max_lub_iff in *
    ; splits; des; s; vauto; eauto using le_trans, NPeano.Nat.max_lub.
    svbv; s; rewrite NPeano.Nat.max_comm; eapply subuniq_dbl'; eauto.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (withd _ _) eqn: W; destruct (perform _ _ _ _) eqn:E
    ; simpl withd
    
    ; try (by s in W; clarify; splits; s nextbvar'; vauto
           ; auto using NPeano.Nat.max_lub; []; rewrite ->lb_nil; auto)
    ; destruct p; s in W; clarify.
    + exploit (perform_wf _ a); rnbv; eauto; [by econs; eauto|]; i; des.
      splits; eauto.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (withd _ _) eqn: W; destruct (perform _ _ _ _) eqn:E
    ; simpl withd
    
    ; try (by s in W; clarify; splits; s nextbvar'; vauto
           ; auto using NPeano.Nat.max_lub; []; rewrite ->lb_nil; auto)
    ; destruct p; s in W; clarify.
    + exploit (perform_wf _ a); rnbv; eauto; [by econs; eauto|]; i; des.
      splits; eauto.
  - s in WF; destruct bx; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ ae1) eqn:TR1; destruct p
    ; destruct (transform_ _ _ ae2) eqn:TR2; destruct p
    ; s in MX; rewrite ->!NPeano.Nat.max_lub_iff in MX; des; clarify. 
    + hexploit transform_wfatom_; rnbv; try (by trauto; vauto).
      { 
        
        simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
        ; rewrite -app_comm_cons in UN; rewrite !map_app in UN
        ; rewrite -app_assoc in UN
        ; simpl purge_lbl; simpl vars2nat; ii; specialize (UN _ BIG).
        eapply uniq_sublist; eauto using sublist_app.
      }
      { i; rnbv; eapply IH; clarify; auto; rr in H; rr; s; try omega; eauto. }
      i; des; rewrite ->!NPeano.Nat.max_lub_iff in NINC; des.
      hexploit (IH (aerase ae2)); rnbv; first (by rr; s; omega)
      ; try (by trauto; vauto); eauto using le_trans.
      {
        simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
        ; rewrite -app_comm_cons in UN; rewrite !map_app in UN
        ; simpl purge_lbl; simpl vars2nat; ii
        ; specialize (UN _ (le_trans _ _ _ NINC0 BIG)).
        eapply uniq_perm
        ; last by eapply Permutation_cons; eapply Permutation_cons_app; eauto.
        eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
      }
      i; des; s; destruct (withd _ _) eqn:W; destruct (perform _ _ _ _) eqn:E
      ; s in W; clarify; ssbvar; last first.
      {
        splits; s; vauto.
        do !(apply NPeano.Nat.max_lub; eauto using le_trans).
        svbv; s; rewrite NPeano.Nat.max_comm; eauto using subuniq_dbl'.
      }
      exploit perform_wf; eauto; rnbv; ssbvar.
      * splits; eauto using le_trans.
      * s; eauto.
      * simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
        ; rewrite ->!map_app in *; rewrite ->app_comm_cons in *
        ; rewrite ->cons_app in *.
        apply (@uniq_vars2 _ n0 _ _ _ _ _ _ UN); try (by omega); auto.
        {
          eapply subuniq_lb_weaken_r with (n'':=n0) in BV; auto.
          eapply subuniq_trans; [eapply subuniq_app|]; [auto | exact BV|].
          eapply subuniq_perm; crush_perm.
        }
        { apply subuniq_lb_weaken_r with (n'':=n1) in BV0; auto. }
      * i; des; ssbvar; splits; eauto using NPeano.Nat.max_lub, le_trans
        ; [|by ii; rewrite H0 in GE0; eapply GE0; vauto].
        svbv; rewrite <-nextbvar_equiv; eapply subuniq_bvar_trp; rnbv
        ; eauto; instantiate
        ; try (by splits; apply NPeano.Nat.max_lub_iff; splits
             ; vauto).
        ii; specialize (UN _ BIG); eauto 6 using uniq_sublist, sublist_app.
    + 
      hexploit transform_wfatom_; rnbv; try (by trauto; vauto).
      { 
        
        simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
        ; rewrite -app_comm_cons in UN; rewrite !map_app in UN
        ; rewrite -app_assoc in UN
        ; simpl purge_lbl; simpl vars2nat; ii; specialize (UN _ BIG).
        eapply uniq_sublist; eauto using sublist_app.
      }
      { i; rnbv; eapply IH; clarify; auto; rr in H; rr; s; try omega; eauto. }
      i; des; rewrite ->!NPeano.Nat.max_lub_iff in NINC; des.
      hexploit (IH (aerase ae2)); rnbv; first (by rr; s; omega)
      ; try (by trauto; vauto); eauto using le_trans.
      {
        simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
        ; rewrite -app_comm_cons in UN; rewrite !map_app in UN
        ; simpl purge_lbl; simpl vars2nat; ii
        ; specialize (UN _ (le_trans _ _ _ NINC0 BIG)).
        eapply uniq_perm
        ; last by eapply Permutation_cons; eapply Permutation_cons_app; eauto.
        eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
      }
      i; des; s; destruct (withd _ _) eqn:W; destruct (perform _ _ _ _) eqn:E
      ; s in W; clarify; ssbvar; last first.
      {
        splits; s; vauto.
        do !(apply NPeano.Nat.max_lub; eauto using le_trans).
        svbv; s; rewrite NPeano.Nat.max_comm; eauto using subuniq_dbl'.
      }
      exploit perform_wf; eauto; rnbv; ssbvar.
      * splits; eauto using le_trans.
      * s; eauto.
      * simpl bvars in UN; simpl vars2nat in *; unfold vars2nat in *
        ; rewrite ->!map_app in *; rewrite ->app_comm_cons in *
        ; rewrite ->cons_app in *.
        apply (@uniq_vars2 _ n0 _ _ _ _ _ _ UN); try (by omega); auto.
        {
          eapply subuniq_lb_weaken_r with (n'':=n0) in BV; auto.
          eapply subuniq_trans; [eapply subuniq_app|]; [auto | exact BV|].
          eapply subuniq_perm; crush_perm.
        }
        { apply subuniq_lb_weaken_r with (n'':=n1) in BV0; auto. }
      * i; des; ssbvar; splits; eauto using NPeano.Nat.max_lub, le_trans
        ; [|by ii; rewrite H0 in GE0; eapply GE0; vauto].
        svbv; rewrite <-nextbvar_equiv; eapply subuniq_bvar_trp; rnbv; eauto
        ; instantiate
        ; try (by splits; apply NPeano.Nat.max_lub_iff; splits
               ; vauto).
        ii; specialize (UN _ BIG); eauto 6 using uniq_sublist, sublist_app.
Qed.

Lemma transform_wfatom: forall (A: annT) e e' n n' (ae: aexp A) G
    (E: aerase ae = e)
    (T: transform G n ae = (e',n'))
    (MX: nextbvar (aerase ae) <= n)
    (WF: wfatom G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: vars2nat (bvars e) ++ vars2nat(purge_lbl G))),
  <<NINC: max (nextbvar e') n <= n'>> /\
  <<WF: wfatom G e'>> /\
  <<BV: subuniq (vars2nat (bvars e'))
                (list_between n (nextbvar e') ++ vars2nat (bvars e))>> /\
  <<GT: is_good_thunk e -> is_good_thunk e'>>.
Proof.
  i; apply (@transform_wfatom_ A) with (ae:=ae); auto.
  i; rnbv; apply (@transform_wfexp A) with (ae:=ae'); vauto.
Qed.

Theorem transform_wfmod: forall (A: annT) (analyze: exp -> aexp A) FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: forall n' (BIG: (nextvar m) <= n'),
           uniq (n' :: vars2nat (bvars_mod m)))
    (AN: forall e, aerase (analyze e) = e),
  mid_wfmod FI (transform_mod analyze (map (inl<*> inl) FI) (nextvar m) m) FO
            /\
  uniqvar (transform_mod analyze (map (inl<*>inl) FI) (nextvar m) m).
Proof.
  i; apply transmodmap_wfmod; eauto; []; i.
  hexploit transform_wfatom; eauto.
  - rewrite (AN e); vauto.
  - ii; specialize (UN _ BIG); rewrite purge_lbl_map_inl_inl
    ; rewrite app_nil_r; eauto using uniq_sublist.
  - ii; des; splits; eauto.
Qed.

Lemma transform_fst_equiv: forall (A: annT) (analyze: exp -> aexp A) G n m,
    transform_mod analyze G n m = fst (transform_mod_ analyze G n m).
Proof.
 ii; unfold transform_mod,transform_mod_; eauto using transmodmap_fst_equiv.
Qed.

Theorem transform_wfmod_: forall (A: annT) (analyze: exp -> aexp A) FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: forall n' (BIG: (nextvar m) <= n'),
           uniq (n' :: vars2nat (bvars_mod m)))
    (AN: forall e, aerase (analyze e) = e),
  mid_wfmod FI
            (fst (transform_mod_ analyze (map (inl<*> inl) FI) (nextvar m) m))
            FO /\
  uniqvar (fst (transform_mod_ analyze (map (inl<*> inl) FI) (nextvar m) m)).
Proof.
  ii; rewrite <-transform_fst_equiv; auto using transform_wfmod.
Qed.

(** *** Semantic correctness. *)

Ltac uniq_auto H :=
  try (by simpl bvars in H;
       eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; inst;
       [rewrite ->cons_app, <-app_assoc; eapply Permutation_cons_app; eauto
       |rewrite -app_assoc; eauto using sublist_app]).


Lemma mmrtc_psim_phi_weaken : forall lw ea eb G phia phib,
    impl (rtc (mm.mm_psim lw G ren_id) ea eb)
  (rtc (mm_psim lw G ren_id phia phib) ea eb).
Proof.
  intros; intro SIM; induction SIM; vauto.
  econs; apply mm_psim_phi_weaken; auto.
Qed.

Lemma mmrtc_asim_phi_weaken : forall lw ea eb G phia phib,
    impl (rtc (mm.mm_asim lw G ren_id) ea eb)
  (rtc (mm_test.mm_asim lw G ren_id phia phib) ea eb).
Proof.
  intros; intro SIM; induction SIM; vauto.
  econs; apply mm_asim_phi_weaken; auto.
Qed.

Lemma mmrtc_ksim_phi_weaken : forall lw ea eb G phia phib,
    impl (rtc (mm.mm_ksim lw G ren_id) ea eb)
  (rtc (mm_test.mm_ksim lw G ren_id phia phib) ea eb).
Proof.
  intros; intro SIM; induction SIM; vauto.
  econs; apply mm_ksim_phi_weaken; auto.
Qed.

Lemma conj_app: forall (A B C: Prop), (B -> C) -> A /\ B -> A /\ C.
Proof. intuition. Qed.

Lemma boo: forall (P: Prop) lw ea eb G phia phib
    (CONJ: P /\ <<SIM:mm.mm_ksim lw G ren_id ea eb>>),
  P /\ <<SIM: mm_test.mm_ksim lw G ren_id phia phib ea eb>>.
Proof.
  ii; des; eapply conj_app; [eapply mm_ksim_phi_weaken|]; auto.
Qed.

Lemma transform_asim: forall (A: annT) (ae: aexp A) G e e' n (n':nat)
    (E: aerase ae = e)
    (T: transform G n ae = (e',n'))
    (MX: nextbvar (aerase ae) <= n)
    (WF: wfatom G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: vars2nat (bvars e) ++ vars2nat (purge_lbl G)))
    (IH: forall e', size e' < size e -> forall n n' e'' G ae',
           wfexp G e' ->
           (nextbvar (aerase ae')) <= n ->
           (forall m (BIG: n <= m),
              uniq (m :: vars2nat (bvars e') ++ vars2nat (purge_lbl G))) ->
           @aerase A ae' = e' ->
           transform G n ae' = (e'',n') ->
         <<MAX: max (nextbvar e'') n <= n'>> /\
         <<SIM: rtc (mm.mm_psim (lwemp mm_gw) G ren_id) e'' e'>>),
  <<MAX: max (nextbvar e') n <= n'>> /\
  <<SIM: rtc (mm.mm_asim (lwemp mm_gw) G ren_id) e' e>>.
Proof.
  i; destruct ae; try (by unfold transform, transform_ in T; clarify
                       ; vauto; auto using subuniq_sublist,
                                NPeano.Nat.max_lub).
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; clarify; rnbv; s in MX; s; ssbvar; s; s in WF; vauto.
    hexploit (IH (aerase ae)); trauto; rnbv; eauto; first last.
    ii; des; splits; ssbvar; splits; eauto using mmrtc_compat_fix, le_trans.
    
    simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
    ; simpl vars2nat; ii; specialize (UN _ BIG).
    eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
    + rewrite ->!app_comm_cons; apply sublist_app
      ; [auto | do !apply sublist_cons_keep]; []
      ; rewrite purge_purge_lbl; apply vars2nat_purge_sublist.
    + rewrite <-!app_comm_cons; crush_perm.
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; clarify; rnbv; s in MX; s; ssbvar; vauto.
    hexploit (IH (aerase ae)); trauto; rnbv; eauto; first last.
    ii; des; splits; ssbvar; splits; eauto using mmrtc_compat_gen, le_trans.
    simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
    ; simpl vars2nat; ii; specialize (UN _ BIG).
    eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
    + rewrite ->!app_comm_cons; apply sublist_app
      ; [auto | do !apply sublist_cons_keep]; []
      ; rewrite purge_purge_lbl; apply vars2nat_purge_sublist.
    + rewrite <-!app_comm_cons; crush_perm.
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; clarify; rnbv; s in MX; s; ssbvar; s; s in WF; vauto.
    hexploit (IH (aerase ae)); trauto; rnbv; eauto; first last.
    ii; des; splits; ssbvar; splits; vauto; eauto using le_trans.
    svbv; ii; specialize (UN _ BIG).
    eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
    + eauto using uniq_sublist.
    + crush_perm.
Qed.

Lemma transform_ksim: forall (A: annT) (ae: aexp A) G e e' n (n': nat)
    (E: aerase ae = e)
    (T: transform G n ae = (e',n'))
    (MX: (nextbvar (aerase ae) <= n))
    (WF: wfatom G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: vars2nat (bvars e) ++ vars2nat (purge_lbl G)))
    (IH: forall e', size e' < size e -> forall G ae' n n' e'',
           wfexp G e' ->
           (nextbvar' (aerase ae')) <= n ->
           (forall m (BIG: n <= m),
              uniq (m :: vars2nat (bvars e') ++ vars2nat (purge_lbl G))) ->
           @aerase A ae' = e' ->
           transform G n ae' = (e'',n') ->
           <<MAX: max (nextbvar e'') n <= n'>> /\
           <<SIM: rtc (mm.mm_psim (lwemp mm_gw) G ren_id) e'' e'>>),
  <<MAX: max (nextbvar e') n <= n'>> /\
  <<SIM: rtc (mm.mm_ksim (lwemp mm_gw) G ren_id) e' e>>.
Proof.
  i; destruct ae; try (by unfold transform,transform_ in T; vauto
                       ; auto using NPeano.Nat.max_lub).
    
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; clarify; rnbv; s in MX; s; ssbvar; s; s in WF; vauto.
    hexploit (IH (aerase ae)); trauto; rnbv; eauto; first last.
    ii; des; splits; ssbvar; splits; vauto; eauto using le_trans.
    simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
    ; simpl vars2nat; ii; specialize (UN _ BIG).
    eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
    + rewrite ->!app_comm_cons; apply sublist_app
      ; [auto | do !apply sublist_cons_keep]; []
      ; rewrite purge_purge_lbl; apply vars2nat_purge_sublist.
    + rewrite <-!app_comm_cons; crush_perm.
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; clarify; rnbv; s in MX; s; ssbvar; s; s in WF; vauto.
    hexploit (IH (aerase ae)); trauto; rnbv; eauto; first last.
    ii; des; splits; ssbvar; splits; vauto; eauto using le_trans.
    simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
    ; simpl vars2nat; ii; specialize (UN _ BIG).
    eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
    + rewrite ->!app_comm_cons; apply sublist_app
      ; [auto | do !apply sublist_cons_keep]; []
      ; rewrite purge_purge_lbl; apply vars2nat_purge_sublist.
    + rewrite <-!app_comm_cons; crush_perm.
  
  - unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; clarify; s in MX; ssbvar; s in WF.
    hexploit (IH (aerase ae)); trauto; rnbv; eauto.
    { 
      simpl bvars in UN; simpl vars2nat in UN; simpl purge_lbl
      ; simpl vars2nat; ii; specialize (UN _ BIG).
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; first last.
      - rewrite ->!app_comm_cons; apply sublist_app
        ; [auto | do !apply sublist_cons_keep]; []
        ; apply vars2nat_sublist; apply sublist_refl.
      - rewrite <-!app_comm_cons; crush_perm.
    }
    i; des; simpl aerase.
    splits; ssbvar; splits; eauto using mmrtc_compat_cont, le_trans.
Qed.

Ltac suniq_auto UN BIG :=
  apply UN in BIG; svbv; eauto 6 using uniq_sublist,sublist_app.


Ltac duniq_auto UN NINC BIG :=
  specialize (UN _ (le_trans' NINC BIG)); svbv; simpl purge_lbl
  ; eapply uniq_perm; last (by eapply Permutation_cons
                            ; eapply Permutation_cons_app; eauto)
  ; inst; eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.


Ltac spec_uniq n n0 n1 UN BV BV0 H:=
  hexploit (@uniq_with_list_between n n1); try (by omega)
  ; first (by intros m0 LRG; specialize (UN _ LRG); exact UN)
  ; try eassumption
  ; eapply subuniq_lb_weaken_r with (n'':=n0) in BV; [|by omega]
  ; eapply subuniq_lb_weaken_r with (n'':=n1) in BV0; [|by omega]
  ; intros H; rewrite <-(@list_between_sub n n0) in H; [|by omega]
  ; rewrite cons_app; eapply subuniq_uniq_app; [|auto|]
  ; [|rewrite -app_comm_cons; rewrite cons_app]
  ; [|apply subuniq_app]; [|auto| apply subuniq_app]
  ; [| exact BV| apply subuniq_app]; [|exact BV0|]; auto
  ; eapply (uniq_perm H); crush_perm.

Lemma mm_psim_psim: forall lw G e e'
    (SIM: mm_psim lw G ren_id nil nil e' e),
  mm.mm_psim lw G ren_id e' e.
Proof.
  i; rrdes SIM; rr; eauto.
Qed.

Lemma mm_rtcpsim_psim: forall lw G e e'
    (SIM: rtc (mm_psim lw G ren_id nil nil) e' e),
 rtc (mm.mm_psim lw G ren_id) e' e.
Proof.
  ii; induction SIM; vauto; []; econs; auto using mm_psim_psim.
Qed.

Lemma transform_psim: forall (A: annT) e (ae: aexp A) G e' n (n': nat)
    (E: aerase ae = e)
    (T: transform G n ae = (e',n'))
    (MX: (nextbvar (aerase ae) <= n))
    (WF: wfexp G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: vars2nat (bvars e) ++ vars2nat (purge_lbl G))),
  <<MAX: max (nextbvar e') n <= n'>> /\
  <<SIM: rtc (mm.mm_psim (lwemp mm_gw) G ren_id) e' e>>.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  i; destruct ae; clarify; try (by s in WF; des; splits
                                ; unfold transform, transform_ in T; clarify
                                ; eauto using subuniq_sublist,
                                  NPeano.Nat.max_lub).
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; clarify; s in MX.
    ssbvar; s; s in WF; des; hexploit (IH (aerase ae)); trauto; rnbv; vauto.
    svbv; ii; apply UN in BIG; eapply (uniq_perm BIG); crush_perm.
    ii; des; splits; ssbvar; eauto using mmrtc_compat_new, lwemp_stable,
                             le_trans.
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; clarify; s in MX.
    ssbvar; s; s in WF; des; hexploit (IH (aerase ae)); trauto; rnbv; vauto.
    svbv; ii; apply UN in BIG; eapply (uniq_perm BIG); crush_perm.
    ii; des; splits; ssbvar; eauto using mmrtc_compat_read, lwemp_stable,
                             le_trans.
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; clarify; s in MX.
    ssbvar; s; s in WF; des; hexploit (IH (aerase ae)); trauto; rnbv; vauto.
    ii; des; splits; ssbvar; eauto using mmrtc_compat_write, lwemp_stable,
                             le_trans.
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; clarify; s in MX.
    ssbvar; s; s in WF; des; hexploit (IH (aerase ae)); trauto; rnbv; vauto.
    svbv; ii; apply UN in BIG; eapply (uniq_perm BIG); crush_perm.
    ii; des; splits; ssbvar; eauto using mmrtc_compat_input, lwemp_stable,
                             le_trans.
  - svbv; unfold transform in T; s in T; destruct (transform_ _ _ _) eqn:TR
    ; destruct p; rnbv; clarify; s in MX.
    ssbvar; s; s in WF; des; hexploit (IH (aerase ae)); trauto; rnbv; vauto.
    ii; des; splits; ssbvar; eauto using mmrtc_compat_output, lwemp_stable,
                             le_trans.
  - svbv; unfold transform in T; s in T
    ; destruct (transform_ _ _ ae1) eqn:TR1; destruct p
    ; destruct (transform_ _ _ ae2) eqn:TR2; destruct p
    ; rnbv; clarify; s in MX.
    ssbvar; s; s in WF; des.
    hexploit (IH (aerase ae1)); rnbv; first (by rr; s; omega)
    ; try exact MX; trauto; vauto.
    { ii; apply UN in BIG; eauto using uniq_sublist, sublist_app. }
    ii; des; ssbvar.
    hexploit (IH (aerase ae2)); rnbv; first (by rr; s; omega)
    ; tranify ae2; vauto; eauto using le_trans.
    + ii; specialize (UN _ (le_trans' MAX0 BIG))
      ; eauto using uniq_sublist, sublist_app.
    + ii; rnbv; des; splits; ssbvar
      ; eauto using mmrtc_compat_ifnz, lwemp_stable, le_trans.
  - svbv; unfold transform in T; s in T
    ; destruct (transform_ _ _ ae1) eqn:TR1; destruct p
    ; destruct (transform_ _ _ ae2) eqn:TR2; destruct p
    ; rnbv; clarify; s in MX.
    ssbvar; s; s in WF; des.
    hexploit (IH (aerase ae1)); rnbv; first (by rr; s; omega)
    ; tranify ae1; vauto.
    {
      simpl purge_lbl; simpl map.
      ii; apply UN in BIG; eapply uniq_perm
      ; last by eapply Permutation_cons; eapply Permutation_cons_app; eauto.
      eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
    }
    ii; des; ssbvar.
    hexploit (IH (aerase ae2)); rnbv; first (by rr; s; omega)
    ; tranify ae2; vauto; eauto using le_trans.
    + ii; simpl purge_lbl; simpl map.
      specialize (UN _ (le_trans' MAX0 BIG)); eapply uniq_perm
      ; last by eapply Permutation_cons; eapply Permutation_cons_app; eauto.
      eapply uniq_sublist; eauto using vars2nat_sublist, sublist_app.
    + ii; rnbv; des; do !(splits; ssbvar)
      ; eauto using mmrtc_compat_case, lwemp_stable, le_trans.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (withd _ _) eqn: W; destruct (perform _ _ _ _) eqn:E
    ; simpl withd
    
    ; try (by s in W; clarify; splits; ssbvar; vauto)
    ; destruct p; rnbv; clarify; s in MX.
    
    rewrite <-nextbvar_equiv; ss; clarify
    ; eapply conj_app; [eapply mm_rtcpsim_psim|]; eapply perform_sim
    ; rnbv; vauto; eauto using lwemp_stable.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (withd _ _) eqn: W; destruct (perform _ _ _ _) eqn:E
    ; simpl withd
    
    ; try (by s in W; clarify; splits; ssbvar; vauto)
    ; destruct p; rnbv; clarify; s in MX.
    
    rewrite <-nextbvar_equiv; ss; clarify
    ;  eapply conj_app; [eapply mm_rtcpsim_psim|]; eapply perform_sim
    ; rnbv; vauto; eauto using lwemp_stable.
  - s in WF; des; unfold transform in T; s in T
    ; destruct (transform_ _ _ ae1) eqn:TR1; destruct p
    ; destruct (transform_ _ _ ae2) eqn:TR2; destruct p
    ; rnbv; clarify; s in MX; destruct (withd _ _) eqn:W
    ; destruct (perform _ _ _ _) eqn:E; simpl withd.
    + destruct p; rnbv; ssbvar; s in W; clarify; s; split.
      * destruct bx; ssbvar; s in WF; clarify; s
        ; hexploit transform_wfatom; rnbv; tranify ae1; vauto
        ; try (by ii; suniq_auto UN BIG)
        ; i; des; ssbvar; hexploit transform_wfexp; rnbv; tranify ae2; vauto
        ; eauto using le_trans; inst
        ; try (by intros m' LRG; duniq_auto UN NINC0 LRG)
        ; inst; i; des; try (by ssbvar; splits; eauto using le_trans).
        {
          assert (H: max (nextbvar e') n1 <= n').
          {
            eapply perform_wf; rnbv; ssbvar; eauto.
            - ssbvar; splits; eauto using le_trans.
            - s in WF2; s; splits; auto.
            - ii; svbv; ssbvar; rewrite ->app_comm_cons in *.
              spec_uniq n n0 n1 UN BV BV0 H.
          }
          unnw; rnbv; ssbvar; splits; eauto using le_trans.
        }
        {
          assert (H: max (nextbvar e') n1 <= n').
          {
            eapply perform_wf; rnbv; ssbvar; eauto.
            - ssbvar; splits; eauto using le_trans.
            - s in WF2; s; splits; auto.
            - ii; svbv; ssbvar; rewrite ->app_comm_cons in *.
              spec_uniq n n0 n1 UN BV BV0 H.
          }
          unnw; rnbv; ssbvar; splits; eauto using le_trans.
        }
      * {
        econs 3.
          - eapply conj_app; [eapply mm_rtcpsim_psim|]; eapply perform_sim; first exact E
            ; rnbv; vauto; eauto using lwemp_stable; destruct bx
            ; ssbvar; s in WF; clarify
            ; hexploit transform_wfatom; rnbv; tranify ae1; vauto
            ; try (by ii; suniq_auto UN BIG)
            ; i; des; ssbvar; hexploit transform_wfexp; rnbv; tranify ae2
            ; vauto; eauto using le_trans; inst
            ; try (by intros m' LRG; duniq_auto UN NINC0 LRG)
            ; inst; i; des; try (by ssbvar; splits; eauto using le_trans)
            ; svbv; ssbvar; rewrite ->app_comm_cons in *
            ; spec_uniq n n0 n1 UN BV BV0 H.
          - destruct bx; ssbvar; s in WF; clarify
            ; hexploit transform_wfatom; rnbv; tranify ae1; vauto
            ; try (by ii; suniq_auto UN BIG)
            ; i; des; ssbvar; hexploit transform_wfexp; rnbv; tranify ae2
            ; vauto; eauto using le_trans; inst
            ; try (by intros m' LRG; duniq_auto UN NINC0 LRG)
            ; inst; i; des; try (by do !(ssbvar; splits)
                                 ; eauto using le_trans).
            + eapply mmrtc_compat_let; eauto using lwemp_stable.
              * eapply transform_asim; rnbv; tranify ae1; vauto
                ; try (by ii; suniq_auto UN BIG).
                i; eapply IH; eauto; rr; s; omega.
              * eapply (IH (aerase ae2)); rnbv; tranify ae2
                ; eauto using le_trans; try (by ii; duniq_auto UN NINC0 BIG)
                ; []; rr; s; omega.
            + eapply mmrtc_compat_letcont; eauto using lwemp_stable.
              * eapply transform_ksim; rnbv; tranify ae1; vauto
                ; try (by ii; suniq_auto UN BIG).
                i; eapply IH; rnbv; eauto; rr; s; omega.
              * eapply (IH (aerase ae2)); rnbv; tranify ae2
                ; eauto using le_trans; try (by ii; duniq_auto UN NINC0 BIG)
                ; []; rr; s; omega.
        }
   + s in W; destruct bx; ssbvar; s in WF; clarify; s
     ; hexploit transform_wfatom; rnbv; tranify ae1; vauto
     ; try (by ii; suniq_auto UN BIG)
     ; i; des; ssbvar; hexploit transform_wfexp; rnbv; tranify ae2; vauto
     ; eauto using le_trans; inst
     ; try (by intros m' LRG; duniq_auto UN NINC0 LRG)
     ; inst; i; des; splits
     ; try (by do !(ssbvar; splits); eauto using le_trans).
     {
       eapply mmrtc_compat_let; eauto using lwemp_stable.
       - eapply transform_asim; rnbv; tranify ae1; vauto
         ; try (by ii; suniq_auto UN BIG).
         i; eapply IH; eauto; rr; s; omega.
       - eapply (IH (aerase ae2)); rnbv; tranify ae2
         ; eauto using le_trans; try (by ii; duniq_auto UN NINC0 BIG)
         ; []; rr; s; omega.
     }
     {
       eapply mmrtc_compat_letcont; eauto using lwemp_stable.
       - eapply transform_ksim; rnbv; tranify ae1; vauto
         ; try (by ii; suniq_auto UN BIG).
         i; eapply IH; rnbv; eauto; rr; s; omega.
       - eapply (IH (aerase ae2)); rnbv; tranify ae2
         ; eauto using le_trans; try (by ii; duniq_auto UN NINC0 BIG)
         ; []; rr; s; omega.
     }
Grab Existential Variables.
  auto. auto. auto.
Qed.

Lemma transform_psim_phi: forall (A: annT) e (ae: aexp A) G e' n (n': nat)
                                 phia phib
    (E: aerase ae = e)
    (T: transform G n ae = (e',n'))
    (MX: (nextbvar (aerase ae) <= n))
    (WF: wfexp G e)
    (UN: forall m (BIG: n <= m),
           uniq (m :: vars2nat (bvars e) ++ vars2nat (purge_lbl G))),
  <<MAX: max (nextbvar e') n <= n'>> /\
  <<SIM: rtc (mm_psim (lwemp mm_gw) G ren_id phia phib) e' e>>.
Proof.
  ii; eapply conj_app; [eapply mmrtc_psim_phi_weaken|].
  eapply transform_psim; eauto.
Qed.

Theorem transform_msim: forall nv (A: annT) (analyze: exp -> aexp A) FI FO m
    (WF: mid_wfmod FI m FO)
    (GE: nextvar m <= nv)
    (UV: forall n' (BIG: (nextvar m) <= n'),
           uniq (n' :: vars2nat (bvars_mod m)))
    (AN: forall e, aerase (analyze e) = e),
  rtc (msim mm_gw FI FO)
      (transform_mod analyze (map (inl<*>inl) FI) nv m) m.
Proof.
  i; ginduction WF; eauto.
  ii; unfold transform_mod; s; destruct (transform _ _ _) eqn:TR.
  i; hexploit (IHWF n); eauto using le_nextvar.
  {
    hexploit transform_wfatom; tranify (analyze e); eauto.
    - rewrite nextvar_rewrite in GE; unfold nextbvar; ssbvar.
      specialize (AN e); rewrite AN; auto.
    - ii; specialize (UV _ (le_trans' GE BIG)); svbv.
      rewrite purge_lbl_map_inl_inl; simpl map; rewrite app_nil_r.
      eauto using uniq_sublist.
    - ii; rewrite nextvar_rewrite in GE; des; ssbvar; eauto using le_trans.
  }
  {
    ii; apply what; auto.
    
    svbv; specialize (UV _ GE); ss; des; apply uniq_appD in UV0; des; auto.
  }
  ii; eapply mm_rtcmsim_cons; eauto using lwemp_stable.
  - hexploit (AN e); [].
    destruct e; vauto.
    + destruct (analyze _) eqn:E; vauto; simpl aerase; i; clarify; [].
      unfold transform in TR; destruct (transform_ _ _ _) eqn:TR'
      ; destruct p; clarify; s in TR'
      ; destruct (transform_ _ _ _) eqn:TR''; destruct p; clarify.
      s in WFE; rewrite map_comp purge_map_inl -map_comp in WFE,TR''.
      apply mm.mmrtc_tlsim_fix; eapply transform_psim; trauto; eauto
      ; try (by rewrite nextvar_rewrite in GE; rnbv; s in GE; ssbvar; auto).
      svbv; ii; specialize (UV _ (le_trans' GE BIG)).
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
      * apply Permutation_cons; rewrite cons_app3.
        eapply Permutation_trans; [apply Permutation_app|]
        ; [apply Permutation_rev| auto |].
        eapply three_way_perm.
      * rewrite purge_lbl_map_inl_inl; s; rhs3 rewrite cons_app3.
        eauto using sublist_app.
    + destruct (analyze _) eqn:E; vauto; simpl aerase; i; clarify; [].
      unfold transform in TR; destruct (transform_ _ _ _) eqn:TR'
      ; destruct p; clarify; s in TR'
      ; destruct (transform_ _ _ _) eqn:TR''; destruct p; clarify.
      s in WFE; rewrite map_comp purge_map_inl -map_comp in WFE,TR''.
      apply mmrtc_tlsim_gen; eapply transform_psim; trauto; eauto
      ; try (by rewrite nextvar_rewrite in GE; rnbv; s in GE; ssbvar; auto).
      svbv; ii; specialize (UV _ (le_trans' GE BIG)).
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
      * apply Permutation_cons; apply Permutation_cons_app; auto. 
      * rewrite purge_lbl_map_inl_inl; s; rhs3 rewrite (cons_app bk).
        eauto using sublist_app.
  - unfold transform_mod in H; rewrite !map_app in H; s in H; auto.
  - inst_pairs; i; rr; s; unfold mid_cemp; eauto.
Grab Existential Variables.
  exact tt.
Qed.

Theorem transform_msim_: forall (A: annT) (analyze: exp -> aexp A) FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: forall n' (BIG: (nextvar m) <= n'),
           uniq (n' :: vars2nat (bvars_mod m)))
    (AN: forall e, aerase (analyze e) = e),
  rtc (msim mm_gw FI FO)
      (fst (transform_mod_ analyze (map (inl<*>inl) FI) (nextvar m) m)) m.
Proof.
  ii; rewrite <-transform_fst_equiv; auto using transform_msim.
Qed.

