(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Types for our source language *)

Require Import Arith Bool List Omega.
Require Import vgtac basis countable lang_common.
Require Import ssr.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".

Definition typename := nat.
Definition typebvar := nat.


(** * Definitions. *)

(** Syntax (binding treated in locally nameless style). *)
Inductive typ : Set :=
 | Tname (aa:typename)
 | Tbvar (bb:typebvar)
 | Tunit
 | Tnat
 | Tprod (tya:typ) (tyb:typ)
 | Tsum (tya:typ) (tyb:typ)
 | Tarrow (tya:typ) (tyb:typ)
 | Tmu (ty:typ)
 | Tforall (ty:typ)
 | Texists (ty:typ)
 | Tref (ty:typ)
.


(** Free type names *)
Fixpoint ftn_typ (tty:typ) (aa: typename) : Prop :=
  match tty with
  | Tname a => aa = a
  | Tbvar _ => False
  | Tunit => False
  | Tnat => False
  | Tprod tya tyb => ftn_typ tya aa \/ ftn_typ tyb aa
  | Tsum tya tyb => ftn_typ tya aa \/ ftn_typ tyb aa
  | Tarrow tya tyb => ftn_typ tya aa \/ ftn_typ tyb aa
  | Tmu ty => ftn_typ ty aa
  | Tforall ty => ftn_typ ty aa
  | Texists ty => ftn_typ ty aa
  | Tref ty => ftn_typ ty aa
  end.

Fixpoint ftn_typ_l (tys: list typ) (aa: typename) :=
  match tys with
  | nil => False
  | ty :: tys' => ftn_typ ty aa \/ ftn_typ_l tys' aa
  end.


(** ** Substitutions. *)

Fixpoint tsubst_bvar (x: typebvar) (tyx ty: typ) :=
  match ty with
  | Tname aa    => Tname aa
  | Tbvar bb    =>
    match nat_compare bb x with
    | Lt => Tbvar bb
    | Eq => tyx
    | Gt => Tbvar (pred bb)
    end
  | Tunit       => Tunit 
  | Tnat        => Tnat 
  | Tprod tya tyb  => Tprod  (tsubst_bvar x tyx tya) (tsubst_bvar x tyx tyb)
  | Tsum tya tyb   => Tsum   (tsubst_bvar x tyx tya) (tsubst_bvar x tyx tyb)
  | Tarrow tya tyb => Tarrow (tsubst_bvar x tyx tya) (tsubst_bvar x tyx tyb)
  | Tmu ty     => Tmu     (tsubst_bvar (S x) tyx ty)
  | Tforall ty => Tforall (tsubst_bvar (S x) tyx ty)
  | Texists ty => Texists (tsubst_bvar (S x) tyx ty)
  | Tref ty    => Tref (tsubst_bvar x tyx ty)
  end.

Fixpoint tsubst_name (sub:typename->typ) (tty:typ) {struct tty} : typ :=
  match tty with
  | Tname aa => sub aa
  | Tbvar bb => Tbvar bb
  | Tunit => Tunit 
  | Tnat => Tnat 
  | Tprod tya tyb => Tprod (tsubst_name sub tya) (tsubst_name sub tyb)
  | Tsum tya tyb => Tsum (tsubst_name sub tya) (tsubst_name sub tyb)
  | Tarrow tya tyb => Tarrow (tsubst_name sub tya) (tsubst_name sub tyb)
  | Tmu ty => Tmu (tsubst_name sub ty)
  | Tforall ty => Tforall (tsubst_name sub ty)
  | Texists ty => Texists (tsubst_name sub ty)
  | Tref ty => Tref (tsubst_name sub ty)
  end.

Definition tid :=
  fun x => Tname x.




Definition termenv : Set := (list (termvar * typ)).

(** Being locally closed. *)
Fixpoint lc_typ (n : nat) (ty : typ) :=
  match ty with
  | Tbvar aa    => aa < n
  | Tname bb    => True
  | Tunit       => True
  | Tnat        => True
  | Tprod tya tyb  => lc_typ n tya /\ lc_typ n tyb
  | Tsum tya tyb   => lc_typ n tya /\ lc_typ n tyb
  | Tarrow tya tyb => lc_typ n tya /\ lc_typ n tyb
  | Tmu ty     => lc_typ (S n) ty
  | Tforall ty => lc_typ (S n) ty
  | Texists ty => lc_typ (S n) ty
  | Tref ty    => lc_typ n ty
  end.
Hint Unfold lc_typ.

Definition lc_env n (G: termenv) : Prop :=
  forall x ty (IN: In (x, ty) G), lc_typ n ty.

Definition is_thunktyp (ty: typ) : Prop :=
  match ty with
  | Tarrow _ _ => True
  | Tforall _ => True
  | _ => False
  end.
Hint Unfold is_thunktyp.

Definition lblenv := list (label * typ). 

Fixpoint is_thunkenv (G: lblenv) : Prop :=
  match G with
  | nil => True
  | (l,ty) :: G' => is_thunktyp ty /\ is_thunkenv G' 
  end.


(** * Type substitutions *)


Definition lc_tsub (tsub: typename -> typ) := 
  forall x, lc_typ 0 (tsub x).
Hint Unfold lc_tsub.


(** * Properties *)


Lemma is_thunkenv_app: forall GA GB
    (FLEA: is_thunkenv GA)
    (FLEB: is_thunkenv GB),
  is_thunkenv (GA++GB).
Proof.
  induction GA; eauto.
  destruct a; simpl in *; intuition.
Qed.

Lemma lc_mon :
  forall ty n m (LT: n <= m), lc_typ n ty -> lc_typ m ty.
Proof.
  induction ty; simpl in *; intros; intuition eauto using le_n_S; omega.
Qed.

Lemma lc_tsubname:
  forall d (CLO: lc_tsub d)
    ty n (H: lc_typ n ty),
  lc_typ n (tsubst_name d ty).
Proof.
  induction ty; ins; des; eauto using lc_mon, le_0_n.
Qed.

Lemma tsubname_lc: 
  forall s ty n (LC: lc_typ n (tsubst_name s ty)), lc_typ n ty.
Proof.
  by intros s ty; induction ty; intros; simpls; des; eauto.
Qed.

Lemma tsubbvar_id : 
  forall tyA ty n, lc_typ n ty -> tsubst_bvar n tyA ty = ty.
Proof.
  induction ty; ins; des; f_equal; eauto.
  destruct (nat_compare_spec bb n); try (cut False; [|omega]); try done.
Qed.

Lemma tsubst_name_tid: forall ty,
  tsubst_name tid ty = ty.
Proof.
  induction ty; unfold tid; auto; s
  ; try (by rewrite IHty; auto)
  ; try (by rewrite IHty1 IHty2; auto).
Qed.



Lemma tsubbvar_lctyp: forall ty ty' n
    (CLO: lc_typ (S n) ty)
    (CLO': lc_typ n ty'),
  lc_typ n (tsubst_bvar n ty' ty).
Proof.
  induction ty; s; i; des; eauto using lc_mon.
  ss; depdes CLO.
  - by rewrite (proj2 (nat_compare_eq_iff n n) refl_equal).
  - rewrite (proj1 (nat_compare_lt bb n)); s; omega.
Qed.

Lemma tsubbvar_lctyp_r: forall ty ty' n
    (CLO: lc_typ n (tsubst_bvar n ty' ty)),
  lc_typ (S n) ty.
Proof.
  induction ty; s; i; des; eauto using lc_mon.
  destruct (nat_compare bb n) eqn: X.
  - apply nat_compare_eq_iff in X; subst; omega.
  - apply nat_compare_lt in X; omega.
  - apply nat_compare_gt in X; ss; omega.
Qed.

Lemma tsub_name_bvar:
  forall t ty n dd (LC: forall x, lc_typ 0 (dd x)) , 
    tsubst_name dd (tsubst_bvar n t ty) =
    tsubst_bvar n (tsubst_name dd t) (tsubst_name dd ty).
Proof.
  induction ty; i; s; f_equal; eauto.
  2: by destruct nat_compare.
  rewrite tsubbvar_id; eauto using lc_mon with arith.
Qed.

Lemma tsubname_agree: forall (ty: typ) d d'
    (AGREE: forall x (IN: ftn_typ ty x), d x = d' x),
  tsubst_name d ty = tsubst_name d' ty.
Proof.
  induction ty; s; i; 
  try erewrite IHty; try erewrite IHty1; try erewrite IHty2; eauto.
Qed.

Fixpoint frnam (ty: typ) : typename :=
  match ty with
  | Tname aa => S aa
  | Tprod tya tyb => frnam tya + frnam tyb
  | Tsum tya tyb => frnam tya + frnam tyb
  | Tarrow tya tyb => frnam tya + frnam tyb
  | Tmu ty => frnam ty
  | Tforall ty => frnam ty
  | Texists ty => frnam ty
  | Tref ty => frnam ty
  | _ => 0
  end.

Fixpoint frnam_l (tys: list typ) : typename :=
  match tys with
    | nil => 0
    | ty :: tys' => frnam ty + frnam_l tys'
  end.
  
Lemma frnam_l_app: forall tysa tysb,
  frnam_l tysa + frnam_l tysb <= frnam_l (tysa ++ tysb).
Proof.
  induction tysa; s; [by eauto|].
  i; specialize (IHtysa tysb); omega.
Qed.

Lemma frnam_big: forall ty a (FR: ftn_typ ty a), a < frnam ty.
Proof.
  induction ty; s; i; eauto; try omega 
  ; (des; [apply IHty1 in FR|apply IHty2 in FR]; omega).
Qed.

Lemma frnam_l_big: forall tys a (FR: ftn_typ_l tys a), a < frnam_l tys.
Proof.
  induction tys; [done|].
  s; i; des; [apply frnam_big in FR | apply IHtys in FR]; omega.
Qed.

Lemma frnam_fresh: forall ty, ~ ftn_typ ty (frnam ty).
Proof. i; apply frnam_big in FH; omega. Qed.

Lemma frnam_l_fresh: forall tys, ~ ftn_typ_l tys (frnam_l tys).
Proof. i; apply frnam_l_big in FH; omega. Qed.

Lemma frnam_l_tsub: forall G (a aa: typename) tya
    (FN: ftn_typ_l (map (tsubst_name [tya//a]>>tid) G) aa),
  ftn_typ_l (tya :: G) aa.
Proof.
  i; induction G; ss; des.
  - clear IHG; induction a0; ss; des; intuition.
    unfold tmext in FN; destruct (eqdec _ _); subst; eauto.
  - intuition.
Qed.

Lemma tsubname_id: forall a tya ty
    (FR: ~ ftn_typ ty a),
  tsubst_name [tya//a]>>tid ty = ty.
Proof.
  i; ginduction ty; i; ss; try f_equal; eauto.
  by rewrite tmext_neq; eauto; intuition.
Qed.

Lemma tsubnameL_id: forall a tya (G: termenv)
    (FR: ~ ftn_typ_l (map snd G) a),
  map (fun x => (fst x, tsubst_name [tya//a]>>tid (snd x))) G = G.
Proof.
  i; ginduction G; eauto.
  i; ss; rewrite tsubname_id; eauto.
  destruct a; f_equal; eauto.
Qed.

Lemma lc_tmext_name: forall x a b,
  lc_typ 0 ([Tname b//a]>>tid x).
Proof.
  by ii; unfold tmext; destruct eq_nat_dec; s; eauto.
Qed.

Lemma lc_tmext_ty: forall a tya x,
  lc_typ 0 tya -> lc_typ 0 ([tya//a]>>tid x).
Proof.
  ii; unfold tmext; destruct eq_nat_dec; s; subst; eauto.
Qed.



(** * Countability *)

Local Arguments nat_pair: simpl never.

Lemma typename_cntinf: countably_infinite_set (fun _: typename => True).
Proof.
  exists (@proj1_sig _ _); exists (fun x => exist _ x I); split; auto.
  i; destruct a; destruct t; auto.
Qed.
Hint Resolve typename_cntinf.

Lemma typ_cntinf: countably_infinite typ.
Proof.
  set (tyenc := fix tyenc ty := 
    match ty with
      | Tname n => nat_pair (0,n)
      | Tbvar a => nat_pair (1,a)
      | Tunit => nat_pair (2,0)
      | Tnat => nat_pair (3,0)
      | Tprod ty1 ty2 => nat_pair (5, nat_pair (tyenc ty1, tyenc ty2))
      | Tsum ty1 ty2 => nat_pair (6, nat_pair (tyenc ty1, tyenc ty2))
      | Tarrow ty1 ty2 => nat_pair (7, nat_pair (tyenc ty1, tyenc ty2))
      | Tmu ty => nat_pair (8, tyenc ty)
      | Tforall ty => nat_pair (9, tyenc ty)
      | Texists ty => nat_pair (10, tyenc ty)
      | Tref ty => nat_pair (11, tyenc ty)
    end).
  set (natembed n := Tname n).

  eapply ci2_impl_ci.
  exists tyenc; exists natembed; split; [|by r; intros EQ; i; inv EQ].
  intro ty; induction ty; intro ty'; destruct ty'
  ; i; eapply nat_pair_inj in H; fold tyenc in H; inv H
  ; try (by erewrite IHty; eauto)
  ; try (by eapply nat_pair_inj in H; inv H; erewrite IHty1; eauto; 
            eapply nat_pair_inj in H1; inv H1; erewrite IHty2; eauto).
Qed.

Notation cid := constructive_indefinite_description.

Instance Enumerable_typ: Enumerable typ :=
  {| enum := proj1_sig (cid _ (ci_impl_ci2 typ_cntinf)) |}.
Proof.
  destruct (cid  _ _); des; auto.
Qed.
