(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import vgtac basis.
Require Import types lang_common lang_src_lsi lang_mid lang_tgt lang_spec.
Require Import model_common model.
Require Export gw_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



(** * References. *)

Record refst :=
  { refst_st :> typ -> location -> location -> Prop
  ; refst_bij: forall ty ty' va va' vb vb'
        (IN: refst_st ty va vb)
        (IN': refst_st ty' va' vb'),
      (va' = va -> ty' = ty /\ vb' = vb) /\
      (vb' = vb -> ty' = ty /\ va' = va)
  ; refst_fin: exists l,
      <<FIN: forall ty la lb (IN: refst_st ty la lb), In (ty, la, lb) l>>
  }.
Arguments refst_bij [r ty ty' va va' vb vb'] _ _.

Program Definition refst_emp: refst :=
  Build_refst (fun _ _ _ => False) _ _.
Next Obligation. vauto. Qed.
Next Obligation. ex @nil; vauto. Qed.

Program Definition ref_sts : transys :=
  {| state := refst
   ; ext s' s := s <r= s'
   ; extpub s' s := s <r= s'
  |}.


Definition refst_alloc (s: typ -> location -> location -> Prop) ty la lb
  : typ -> location -> location -> Prop :=
  s \3/ (fun ty' la' lb' => ty' = ty /\ la' = la /\ lb' = lb).

Program Definition refst_allocP (s: refst) ty la lb
    (NEWa: forall ty' lb', ~ s ty' la lb')
    (NEWb: forall ty' la', ~ s ty' la' lb) : refst :=
  Build_refst (refst_alloc s ty la lb) _ _.
Next Obligation.
  unfold refst_alloc in *; des; vauto.
  - hexploit (refst_bij IN IN'); eauto.
  - splits; i; subst; [hexploit NEWa|hexploit NEWb]; vauto.
  - splits; i; subst; [hexploit NEWa|hexploit NEWb]; vauto.
Qed.
Next Obligation.
  hexploit (refst_fin s); []; i; des.
  exists ((ty, la, lb) :: l).
  ii; destruct IN; des; subst; s; eauto.
Qed.

Lemma refst_alloc_ext: forall ty la lb (s s': state ref_sts)
    (RST: refst_alloc s ty la lb = s'),
  ext ref_sts s' s.
Proof.
  destruct s'; ii; s.
  unfold refst_alloc in RST; ss; subst; eauto.
Qed.


(** Subtraction for [refst].  Used in proof of vertical compositionality. *)
Program Definition refst_minus (sr sr': refst) : refst :=
  {| refst_st ty la lb := sr ty la lb /\ ~ sr' ty la lb |}.
Next Obligation.
  hexploit (refst_bij H1 H); eauto.
Qed.
Next Obligation.
  hexploit (refst_fin sr); []; i; des.
  exists l; ii; eapply FIN; des; eauto.
Qed.


(** Extensionality principle for [refst]. *)
Lemma refst_extensionality: forall (sr sr': refst)
    (E: forall ty la lb, refst_st sr ty la lb <-> refst_st sr' ty la lb),
  sr = sr'.
Proof.
  i; destruct sr, sr'; s in E.
  assert (refst_st0 = refst_st1).
  { extensionalities; apply prop_ext; auto. }
  subst; f_equal; auto using proof_irrelevance.
Qed.



(** * Mid <-> Src.  *)

Definition ms_gwsts :=
  tsprod (tsprod mid_gw_sts src_gw_sts) ref_sts.

Definition ms_rquery (s: ms_gwsts.(state)) ty va (vb: lang_src.val) :=
  exists la lb, <<Ea: va = IVloc la>> /\
                <<Eb: e_of vb = Eloc lb>> /\
                <<REG: snd s ty la lb>>.

Lemma ms_rquery_mon: forall (s: ms_gwsts.(state)) ty va vb s'
    (IN: ms_rquery s ty va vb)
    (EXT: ext _ s' s),
  ms_rquery s' ty va vb.
Proof.
  i; destruct s as [[]], s' as [[]], EXT.
  rrdes IN; ss; des; subst; rr; esplits; ss; eauto.
Qed.








Program Definition ms_gwquery : method_query mid_lsi src_lsi ms_gwsts :=
  {| vquerya s := mid_vquery (fst (fst s))
   ; vqueryb s := src_vquery
   ; cquerya s := mid_cquery (fst (fst s))
   ; cqueryb s := src_cquery
   ; rquery := ms_rquery
  |}.
Next Obligation. inv WV; vauto. Qed.
Next Obligation. inv CQ; eauto. Qed.
Next Obligation. inv CQ; eauto. Qed.
Next Obligation. eapply ms_rquery_mon; s; eauto. Qed.

Definition ms_consconfrel (s: ms_gwsts.(state)) (ha: mid_lsi.(conf)) (hb: src_lsi.(conf)) :=
  <<CPa: mid_confpred (fst (fst s)) ha>> /\
  <<CPb: src_confpred (snd (fst s)) hb>>.


(** ** Configuration relation for references. *)


Definition ms_refconfrel' (sr: refst) g (s: ms_gwsts.(state))
                          (ha: mid_lsi.(conf)) (hb: src_lsi.(conf)) :=
  exists ha' hb',
    <<Ea: ha = (ha', Nothing)>> /\
    <<Eb: hb = (hb', Nothing, Nothing)>> /\
    <<VLDa: pmvalid ha'>> /\
    <<VLDb: pmvalid hb'>> /\
    <<DOMa: forall la v (ALC: pmread ha' la v), exists ty lb, sr ty la lb>> /\
    <<DOMb: forall lb v (ALC: pmread hb' lb v), exists ty la, sr ty la lb>> /\
    <<RELV: forall ty la lb (REG: sr ty la lb), exists va vb,
              <<RDa: pmread ha' la va>> /\
              <<RDb: pmread hb' lb (vb : lang_src.val)>> /\
              <<RVL: vclos ms_gwquery s (g s) ty va vb>> >>.

Definition ms_refconfrel g s :=
  ms_refconfrel' (snd s) g s.


Lemma ms_refconfrel'_mon: forall sr g s ha hb
    (RCR: ms_refconfrel' sr g s ha hb)
    g' (LE: g <g= g'),
  ms_refconfrel' sr g' s ha hb.
Proof.
  i; rrdes RCR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.

Lemma ms_refconfrel_mon : forall g s ha hb
    (CR: ms_refconfrel g s ha hb)
    g' (LE: g <g= g'),
   ms_refconfrel g' s ha hb.
Proof.
  i; eapply ms_refconfrel'_mon; eauto.
Qed.

Lemma ms_refconfrel_fin: forall g s ha hb
    (RCR: ms_refconfrel g s ha hb),
  <<FINa: pmfinite (fst ha)>> /\ <<FINb: pmfinite (fst (fst hb))>>.
Proof.
  i; hexploit (refst_fin (snd s)); []; i; des.
  rrdes RCR; subst; split.
  - exists (map (snd<*>fst) l); i.
    apply DOMa in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
  - exists (map snd l); i.
    apply DOMb in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
Qed.


Program Definition ms_gwconf: method_conf mid_lsi src_lsi ms_gwsts :=
  {| confrel g s := crelplus (ms_consconfrel s) (ms_refconfrel g s) |}.
Next Obligation.
  rrdes PR; rr; esplits; eauto using ms_refconfrel_mon.
Qed.

Definition ms_gw: world_glob mid_lsi src_lsi :=
  {| gwsts := ms_gwsts
   ; gwquery := ms_gwquery
   ; gwconf := ms_gwconf
  |}.


(** Subtraction for [refst] part.  Used in proof of vertical compositionality. *)

Definition ms_minus (sg sg': ms_gw.(state)) : ms_gw.(state) :=
  let '(sm, ss, sr) := sg in
  let '(_, _, sr') := sg' in
  (sm, ss, refst_minus sr sr').

Lemma ms_minus_ext: forall s s',
  ext ms_gw s (ms_minus s s').
Proof.
  unfold ms_minus; i; destruct s as [[]], s' as [[]].
  s; splits; auto; i; des; auto.
Qed.


(** Misc. *)

Lemma ms_refconfrel'_smon: forall sr g s ha hb
    (RCR: ms_refconfrel' sr g s ha hb)
    s' (EXT: ext _ s' s)
    (MON: g s <r= g s'),
  ms_refconfrel' sr g s' ha hb.
Proof.
  i; rrdes RCR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.

Lemma ms_refconfrel'_ext: forall sr sr' g sg ca cb
    (RCR: ms_refconfrel' sr' g sg ca cb)
    (EXT: ext ref_sts sr' sr),
  crelplus (ms_refconfrel' sr g sg)
           (ms_refconfrel' (refst_minus sr' sr) g sg) ca cb.
Proof.
  i; rrdes RCR; subst.
  hexploit (pmplus_from_pred ha' (fun la => exists ty lb, sr ty la lb)); []; i; des; subst.
  hexploit (pmplus_from_pred hb' (fun lb => exists ty la, sr ty la lb)); []; i; des; subst.
  dup VLDa; split_pmvalid VLDa; des.
  dup VLDb; split_pmvalid VLDb; des.
  exists ((hP, Nothing) : conf mid_lsi).
  exists ((hNP, Nothing) : conf mid_lsi).
  exists ((hP0, Nothing, Nothing) : conf src_lsi).
  exists ((hNP0, Nothing, Nothing) : conf src_lsi).
  - rr; esplits; s; noemp; eauto.
    + rr; esplits; eauto; i.
      * apply RDP in ALC; des; eauto.
      * apply RDP0 in ALC; des; eauto.
      * dup REG; apply EXT, RELV in REG; des; esplits; eauto.
        { eapply (proj2 (RDP _ _)); eauto. }
        { eapply (proj2 (RDP0 _ _)); eauto. }
    + rr; esplits; eauto; i.
      * apply RDNP in ALC; des.
        apply DOMa in ALC; des; econs; econs; rr; eauto.
      * apply RDNP0 in ALC; des.
        apply DOMb in ALC; des; econs; econs; rr; eauto.
      * rrdes REG; dup REG; apply RELV in REG; des; esplits; eauto.
        { apply RDNP; split; eauto.
          i; rrdes FH; dup FH; apply EXT in FH.
          hexploit (refst_bij FH REG1); eauto; []; i; des.
          specialize (X eq_refl); des; subst; eauto.
        }
        { apply RDNP0; split; eauto.
          i; rrdes FH; dup FH; apply EXT in FH.
          hexploit (refst_bij FH REG1); eauto; []; i; des.
          specialize (X0 eq_refl); des; subst; eauto.
        }
Qed.

Lemma ms_refconfrel_emp: forall g (sg: state ms_gw)
    (NOREF: snd sg <3= bot3) ha hb,
  ms_refconfrel g sg ha hb <-> ha = cemp /\ hb = cemp.
Proof.
  simpl cemp; unfold mid_cemp, src_cemp; split; intro RCR.
  - rrdes RCR; subst; s.
    split; repeat f_equal; apply pmemp_from_read; eauto; i.
    + apply DOMa in FH; des; apply NOREF in FH; vauto.
    + apply DOMb in FH; des; apply NOREF in FH; vauto.
  - des; subst.
    destruct sg; rr; esplits; s; eauto; i.
    + apply pmread_empD in ALC; vauto.
    + apply pmread_empD in ALC; vauto.
    + exfalso; eapply NOREF; rr; eauto.
Qed.



(** * Tgt <-> Src. *)

Definition ts_gwsts := tsprod (tsprod tgt_gw_sts src_gw_sts) ref_sts.

Definition ts_rquery (s: ts_gwsts.(state)) ty va vb :=
  exists lb, <<IRb: vb = Eloc lb>> /\ <<REG: snd s ty va lb>>.

Lemma ts_rquery_mon: forall (s: ts_gwsts.(state)) ty va vb s'
    (IN: ts_rquery s ty va vb)
    (EXT: ext _ s' s),
  ts_rquery s' ty va vb.
Proof.
  i; destruct s as [[]], s' as [[]], EXT.
  rrdes IN; ss; des; subst; rr; esplits; ss; eauto.
Qed.

Program Definition ts_gwquery: method_query tgt_lsi src_lsi ts_gwsts :=
  {| vquerya s := tgt_vquery (fst (fst s))
   ; vqueryb s := src_vquery
   ; cquerya s := tgt_cquery (fst (fst s))
   ; cqueryb s := src_cquery
   ; rquery := ts_rquery
  |}.
Next Obligation. eapply tgt_vquery_mon; ss; eauto. Qed.
Next Obligation. inv CQ; r; s; eauto. Qed.
Next Obligation. inv CQ; eauto. Qed.
Next Obligation. eapply ts_rquery_mon; ss; eauto. Qed.

Definition ts_consconfrel (s: ts_gwsts.(state)) (ha: tgt_lsi.(conf)) (hb: src_lsi.(conf)) :=
  <<CPa: tgt_confpred (fst (fst s)) ha>> /\
  <<CPb: src_confpred (snd (fst s)) hb>>.


(** ** Configuration relation for references. *)

Definition ts_refconfrel' (sr: refst) g (s: ts_gwsts.(state))
                          (ha: tgt_lsi.(conf)) (hb: src_lsi.(conf)) :=
  exists ha' hb',
    <<Ea: ha = {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := ha' |}>> /\
    <<Eb: hb = (hb', Nothing, Nothing)>> /\
    <<VLDa: pmvalid ha'>> /\
    <<VLDb: pmvalid hb'>> /\
    <<DOMa: forall la v (ALC: pmread ha' la v), exists ty lb, sr ty la lb>> /\
    <<DOMb: forall lb v (ALC: pmread hb' lb v), exists ty la, sr ty la lb>> /\
    <<RELV: forall ty la lb (REG: sr ty la lb), exists va vb,
              <<RDa: pmread ha' la va>> /\
              <<RDb: pmread hb' lb (vb: lang_src.val)>> /\
              <<RVL: vclos ts_gwquery s (g s) ty va vb>> >>.

Definition ts_refconfrel g s :=
  ts_refconfrel' (snd s) g s.

Lemma ts_refconfrel'_mon: forall sr g s ha hb
    (RCR: ts_refconfrel' sr g s ha hb)
    g' (LE: g <g= g'),
  ts_refconfrel' sr g' s ha hb.
Proof.
  i; rrdes RCR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.

Lemma ts_refconfrel_mon : forall g s ha hb
    (CR: ts_refconfrel g s ha hb)
    g' (LE: g <g= g'),
   ts_refconfrel g' s ha hb.
Proof.
  i; eapply ts_refconfrel'_mon; eauto.
Qed.

Lemma ts_refconfrel'_smon: forall sr g s ha hb
    (RCR: ts_refconfrel' sr g s ha hb)
    s' (EXT: ext _ s' s)
    (MON: g s <r= g s'),
  ts_refconfrel' sr g s' ha hb.
Proof.
  i; rrdes RCR; rr; esplits; eauto.
  i; hexploit RELV; eauto; []; i; des.
  esplits; eauto using vclos_mon.
Qed.

Lemma ts_refconfrel_shape: forall s r ca cb
    (CR: ts_refconfrel s r ca cb),
  exists refha refhb,
    ca = {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := refha |} /\
    cb = (refhb, Nothing, Nothing).
Proof.
  i; destruct ca, cb; rrdes CR; ss; subst; eauto.
Qed.

Program Definition ts_gwconf: method_conf tgt_lsi src_lsi ts_gwsts :=
  {| confrel g s := crelplus (ts_consconfrel s) (ts_refconfrel g s) |}.
Next Obligation.
  rrdes PR; rr; esplits; eauto using ts_refconfrel_mon.
Qed.

Definition ts_gw: world_glob tgt_lsi src_lsi :=
  {| gwsts := ts_gwsts
   ; gwquery := ts_gwquery
   ; gwconf := ts_gwconf
  |}.


Lemma ts_refconfrel_fin: forall g s ha hb
    (RCR: ts_refconfrel g s ha hb),
  <<FINa: pmfinite (chp ha)>> /\ <<FINb: pmfinite (fst (fst hb))>>.
Proof.
  i; hexploit (refst_fin (snd s)); []; i; des.
  rrdes RCR; subst; split.
  - exists (map (snd<*>fst) l); i.
    apply DOMa in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
  - exists (map snd l); i.
    apply DOMb in RD; des; apply FIN in RD.
    apply in_map_iff; econs; inst_pairs; s; eauto.
Qed.

Lemma ts_refconfrel_emp: forall g (sg: state ts_gw)
    (NOREF: snd sg <3= bot3) ha hb,
  ts_refconfrel g sg ha hb <-> ha = cemp /\ hb = cemp.
Proof.
  simpl cemp; unfold tgt_cemp, src_cemp; split; intro RCR.
  - rrdes RCR; subst; s.
    split; repeat f_equal; apply pmemp_from_read; eauto; i.
    + apply DOMa in FH; des; apply NOREF in FH; vauto.
    + apply DOMb in FH; des; apply NOREF in FH; vauto.
  - des; subst.
    destruct sg; rr; esplits; s; eauto; i.
    + apply pmread_empD in ALC; vauto.
    + apply pmread_empD in ALC; vauto.
    + exfalso; eapply NOREF; rr; eauto.
Qed.
